
/******************************************************************************
* MODULE     : matrix_integer.hpp
* DESCRIPTION: Multi-modular multiplication of integer matrices
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_MATRIX_INTEGER_HPP
#define __MMX_MATRIX_INTEGER_HPP
#include <numerix/integer.hpp>
#include <algebramix/matrix.hpp>
#include <algebramix/matrix_block.hpp>
namespace mmx {

/******************************************************************************
* Low level Multi-modular multiplication
******************************************************************************/

void
multi_modular_mul (integer* d, nat d_rs, nat d_cs,
		   const integer* s1, nat s1_rs, nat s1_cs,
		   const integer* s2, nat s2_rs, nat s2_cs,
		   nat r, nat l, nat c);

/******************************************************************************
* Kronecker multiplication
******************************************************************************/

template<nat thr, typename V>
struct matrix_integer_multiplication: public V {

template<typename C>
static inline void
mul (C* d, const C* s1, const C* s2,
     nat r, nat rr, nat l, nat ll, nat c, nat cc) {
  V::mul (d, s1, s2, r, rr, l, ll, c, cc); }

static inline void
mul (integer* d, const integer* s1, const integer* s2,
     nat r, nat rr, nat l, nat ll, nat c, nat cc)
{
  if (r <= thr || l <= thr || c <= thr)
    V::Naive::template mul<mul_op> (d, s1, s2, r, rr, l, ll, c, cc);
  else multi_modular_mul
	 (d , V::index (1, 0, rr, cc), V::index (0, 1, rr, cc),
	  s1, V::index (1, 0, rr, ll), V::index (0, 1, rr, ll),
	  s2, V::index (1, 0, ll, cc), V::index (0, 1, ll, cc),
	  r, l, c);
}

template<typename C>
static inline void
mul (C* d, const C* s1, const C* s2,
     nat r, nat l, nat c){
  V::mul (d, s1, s2, r, l, c); }

static inline void
mul (integer* d, const integer* s1, const integer* s2,
     nat r, nat l, nat c)
{
  if (r <= thr || l <= thr || c <= thr)
    V::Naive::template mul<mul_op> (d, s1, s2, r, r, l, l, c, c);
  else multi_modular_mul
	 (d , V::index (1, 0, r, c), V::index (0, 1, r, c),
	  s1, V::index (1, 0, r, l), V::index (0, 1, r, l),
	  s2, V::index (1, 0, l, c), V::index (0, 1, l, c),
	  r, l, c);
}

}; // matrix_integer_multiplication<thr,V>

/******************************************************************************
* Use multi-modular multiplication by default for integer matrices
******************************************************************************/

struct matrix_integer:
    public matrix_naive_application<
	     matrix_integer_multiplication<7,
               matrix_enrich<
		 matrix_naive_multiplication<
		   matrix_naive_row_column<
		     matrix_vectorial<
		       matrix_naive_abstractions<
			 vector_naive> > > > > > >
{
  static const nat def_rows = 0;
  static const nat def_cols = 0;
  static const nat init_rows= 1;
  static const nat init_cols= 1;
};

STMPL
struct matrix_variant_helper<integer> {
  typedef matrix_integer MV;
};

}
#endif // __MMX_MATRIX_INTEGER_HPP
