
/******************************************************************************
* MODULE     : modular_chinese.hpp
* DESCRIPTION: Subroutines for efficient Chinese remaindering
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_MODULAR_CHINESE_HPP
#define __MMX_MODULAR_CHINESE_HPP
#include <numerix/integer.hpp>

namespace mmx {

//#define BITS_IN_MODULUS 24
//#define BITS_IN_CHINESE 23 /* BITS_IN_MODULUS - 1 */
//#define FREE_BITS_IN_DOUBLE  5 /* 53 - 2 * BITS_IN_MODULUS */
//#define FREE_BITS_IN_INT64  16 /* 64 - 2 * BITS_IN_MODULUS */
#define BITS_IN_MODULUS 23
#define BITS_IN_CHINESE 22 /* BITS_IN_MODULUS - 1 */
#define FREE_BITS_IN_DOUBLE  7 /* 53 - 2 * BITS_IN_MODULUS */
#define FREE_BITS_IN_INT64  18 /* 64 - 2 * BITS_IN_MODULUS */

typedef mp_limb_t mod_t;

vector<mod_t> chinese_build_moduli (nat cap);
  // Returns a vector of moduli p[i], whose product has at least cap+2 bits
vector<mod_t> chinese_build_converter (const vector<mod_t>& mods);
  // Build auxiliary conversion information for the moduli p[i]
void chinese_encode (mod_t* dest, const integer& src,
		     const vector<mod_t>& mods, const vector<mod_t>& cv);
  // Compute src mod p[i] for each of the moduli p[i]
integer chinese_decode (const mod_t* src,
			const vector<mod_t>& mods, const vector<mod_t>& cv);
  // Reconstruct the integer from its remainders w.r.t. p[i] for each i

}
#endif // __MMX_MODULAR_CHINESE_HPP
