
/******************************************************************************
* MODULE     : modulus_naive_polynomial.hpp
* DESCRIPTION: Modular arithmetic for polynomials
* COPYRIGHT  : (C) 2008  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_MODULUS_NAIVE_POLYNOMIAL_HPP
#define __MMX_MODULUS_NAIVE_POLYNOMIAL_HPP
#include <numerix/modulus_naive.hpp>
#include <algebramix/polynomial.hpp>

namespace mmx {

#define TMPL template<typename C, typename M>

/******************************************************************************
* Variant
******************************************************************************/

template<typename V>
struct modulus_inv_naive_polynomial : public V {

  TMPL static inline void
  inv_mod (C& a, const M& m) {
    if (m.p == 0) {
      if (deg (a) == 0)
        a= 1 / a [0];
      else
        ERROR ("inv_mod: argument is not invertible");
    }
    a = invert_modulo (a, m.p);
    if (a == 0)
      ERROR ("inv_mod: argument is not invertible"); }

  TMPL static inline void
  inv_mod (C& dest, const C& s, const M& m) {
    dest = s;
    inv_mod (dest, m); }
};

struct modulus_naive_polynomial:
  public modulus_encoding_naive<
         modulus_div_naive<
         modulus_inv_naive_polynomial<
         modulus_mul_naive<
         modulus_add_naive<
         modulus_reduction_naive<
         modulus_normalization_naive> > > > > > {};

/******************************************************************************
* Helper specialization
******************************************************************************/

template<typename C, typename V>
struct modulus_variant_helper<polynomial<C,V> > {
  typedef modulus_naive_polynomial MV; };

#undef TMPL

} // namespace mmx
#endif // __MMX_MODULUS_NAIVE_POLYNOMIAL_HPP
