
/******************************************************************************
* MODULE     : polynomial_tft.hpp
* DESCRIPTION: FFT based algorithms
* COPYRIGHT  : (C) 2008  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__POLYNOMIAL_TFT__HPP
#define __MMX__POLYNOMIAL_TFT__HPP
#include <algebramix/polynomial_dicho.hpp>
#include <algebramix/polynomial_balanced.hpp>
#include <algebramix/fft_truncated.hpp>

namespace mmx {
#define TMPL template<typename C>

struct tft_threshold {};

/******************************************************************************
* Multiplication
******************************************************************************/

template<typename V, typename Th= tft_threshold>
struct polynomial_tft_multiplication: public V {

TMPL static inline void
mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  typedef fft_truncated_transformer<C> Tfter;
  if (n1 < Threshold(C,Th) || n2 < Threshold(C,Th))
    V::mul (dest, s1, s2, n1, n2);
  else {
    nat n= n1 + n2 - 1;
    nat m= next_power_of_two (n);
    nat l= aligned_size<C,V> (m);
    C* temp0= mmx_new<C> (l);
    C* temp1= mmx_new<C> (l);
    C* tempd= mmx_new<C> (l);
    Tfter tfter (n);
    V::copy (temp0, s1, n1); V::clear (temp0+n1, n-n1);
    V::copy (temp1, s2, n2); V::clear (temp1+n2, n-n2);
    tfter.direct_transform (temp0);
    tfter.direct_transform (temp1);
    V::Vec::mul (tempd, temp0, temp1, n);
    tfter.inverse_transform (tempd);
    V::copy (dest, tempd, n);
    mmx_delete<C> (temp0, l);
    mmx_delete<C> (temp1, l);
    mmx_delete<C> (tempd, l);
  }
}

}; // polynomial_tft_multiplication<V>

/******************************************************************************
* TFT based variant polynomials
******************************************************************************/

template<typename V>
struct polynomial_multiply_tft:
  public polynomial_balanced_multiplication<
           polynomial_tft_multiplication<V> >
{};

struct polynomial_tft:
  public polynomial_algos_dicho<
           polynomial_multiply_tft<
	     polynomial_multiply_dicho<
	       polynomial_multiply_naive<
		 vector_naive> > > >
{};

#undef TMPL
} // namespace mmx
#endif //__MMX__POLYNOMIAL_TFT__HPP
