
/******************************************************************************
* MODULE     : quotient_polynomial.hpp
* DESCRIPTION: Rational functions
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_QUOTIENT_POLYNOMIAL__HPP
#define __MMX_QUOTIENT_POLYNOMIAL__HPP
#include <algebramix/quotient.hpp>
#include <algebramix/polynomial.hpp>
namespace mmx {
#define TMPL template<typename C, typename V>
#define Polynomial polynomial<C,V>
#define Quotient quotient<Polynomial,Polynomial>

TMPL inline Quotient
operator + (const Quotient& x1, const C& x2) {
  return Quotient (numerator (x1) + x2 * denominator (x1),
		   denominator (x1));
}

TMPL inline Quotient
operator + (const C& x1, const Quotient& x2) {
  return Quotient (x1 * denominator (x2) + numerator (x2),
		   denominator (x2));
}

TMPL inline Quotient
operator - (const Quotient& x1, const C& x2) {
  return Quotient (numerator (x1) - x2 * denominator (x1),
		   denominator (x1));
}

TMPL inline Quotient
operator - (const C& x1, const Quotient& x2) {
  return Quotient (x1 * denominator (x2) - numerator (x2),
		   denominator (x2));
}

TMPL inline Quotient
operator * (const C& c, const Quotient& x) {
  return Quotient (c * numerator (x), denominator (x));
}

TMPL inline Quotient
operator * (const Quotient& x, const C& c) {
  return Quotient (numerator (x) * c, denominator (x));
}

TMPL inline Quotient
operator / (const C& c, const Quotient& x) {
  ASSERT (numerator (x) != 0, "division by zero");
  return Quotient (c * denominator (x), numerator (x));
}

TMPL inline Quotient
operator / (const Quotient& x, const C& c) {
  ASSERT (c != 0, "division by zero");
  return Quotient (numerator (x) / c, denominator (x));
}

#undef TMPL
#undef Polynomial
#undef Quotient
} // namespace mmx
#endif // __MMX_QUOTIENT_POLYNOMIAL__HPP
