
/******************************************************************************
* MODULE     : vector_simd.hpp
* DESCRIPTION: vectors using SIMD instructions
* COPYRIGHT  : (C) 2008  Joris van der Hoeven and Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__VECTOR_SIMD__HPP
#define __MMX__VECTOR_SIMD__HPP
#include <numerix/simd.hpp>
#include <algebramix/vector_unroll.hpp>
#include <algebramix/vector_aligned.hpp>
#include <algebramix/vector_sse.hpp>

namespace mmx {

/******************************************************************************
* vector_simd variant
******************************************************************************/

template<nat len1, nat len2>
struct vector_simd:
    public vector_enrich<vector_aligned_abstractions<
			   vector_unroll<len1>,
			   vector_unroll<len2> > >
{
  static const nat def_len = 0;
  static const nat init_len= 1;
};

/******************************************************************************
* Suggested specilizations
******************************************************************************/

#define Vector_simd_variant(C) vector_simd_variant_helper<C>::VV

template<typename C>
struct vector_simd_variant_helper {
  typedef vector_simd<4,4> VV; };

STMPL
struct vector_simd_variant_helper<double> {
  typedef vector_simd<8,4> VV; };

STMPL
struct vector_simd_variant_helper<char> {
  typedef vector_simd<32,4> VV; };

STMPL
struct vector_simd_variant_helper<signed char> {
  typedef vector_simd<32,4> VV; };

STMPL
struct vector_simd_variant_helper<unsigned char> {
  typedef vector_simd<32,4> VV; };

STMPL
struct vector_simd_variant_helper<short int> {
  typedef vector_simd<32,4 > VV; };

STMPL
struct vector_simd_variant_helper<short unsigned int> {
  typedef vector_simd<32, 4> VV; };

STMPL
struct vector_simd_variant_helper<int> {
  typedef vector_simd<16, 4> VV; };

STMPL
struct vector_simd_variant_helper<unsigned int> {
  typedef vector_simd<16, 4> VV; };

STMPL
struct vector_simd_variant_helper<long int> {
  typedef vector_simd<16, 4> VV; };

STMPL
struct vector_simd_variant_helper<long unsigned int> {
  typedef vector_simd<16, 4> VV; };

STMPL
struct vector_simd_variant_helper<long long int> {
  typedef vector_simd<8,4> VV; };

STMPL
struct vector_simd_variant_helper<long long unsigned int> {
  typedef vector_simd<8,4> VV;
};

} // namespace mmx

#endif // __MMX__VECTOR_SIMD__HPP
