
/******************************************************************************
* MODULE     : vector_threads.hpp
* DESCRIPTION: multi-threaded vector operations
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__VECTOR_THREADS__HPP
#define __MMX__VECTOR_THREADS__HPP
#include <algebramix/vector_unroll.hpp>
#include <basix/threads.hpp>

namespace mmx {
#define TMPL  template<typename C>
#define TMPLX template<typename C, typename X>

/******************************************************************************
* Abstract low level vector routines on vectors in memory
******************************************************************************/

template<typename V>
struct vector_threads_abstractions: public V {

template<typename Op, typename C>
struct unary_task_rep: public task_rep {
  C* dest; const C* s; nat n;
public:
  inline unary_task_rep (C* dest2, const C* s2, nat n2):
    dest (dest2), s (s2), n (n2) {}
  void execute () { V::template unary<Op> (dest, s, n); }
};

template<typename Op, typename C> static inline void
unary (C* dest, const C* s, nat n) {
  if (n < 256) V::template unary<Op> (dest, s, n);
  else {
    task ts[threads_number];
    for (nat id=0; id<threads_number; id++) {
      nat i1= (id*n)     / threads_number;
      nat i2= ((id+1)*n) / threads_number;
      ts[id]= new unary_task_rep<Op,C> (dest + i1, s + i1, i2 - i1);
    }
    threads_execute (ts, threads_number);
    for (nat id=0; id<threads_number; id++)
      delete ts[id];
  }
}

template<typename Op, typename C>
struct binary_task_rep: public task_rep {
  C* dest; const C* s1; const C* s2; nat n;
public:
  inline binary_task_rep (C* dest2, const C* s1b, const C* s2b, nat n2):
    dest (dest2), s1 (s1b), s2 (s2b), n (n2) {}
  void execute () { V::template binary<Op> (dest, s1, s2, n); }
};

template<typename Op, typename C> static inline void
binary (C* dest, const C* s1, const C* s2, nat n) {
  if (n < 256) V::template binary<Op> (dest, s1, s2, n);
  else {
    task ts[threads_number];
    for (nat id=0; id<threads_number; id++) {
      nat i1= (id*n)     / threads_number;
      nat i2= ((id+1)*n) / threads_number;
      ts[id]= new binary_task_rep<Op,C> (dest + i1, s1 + i1, s2 + i1, i2 - i1);
    }
    threads_execute (ts, threads_number);
  }
}

}; // vector_threads_abstractions

template<typename V>
struct vector_threads:
  public vector_enrich<vector_threads_abstractions<V> >
{
  static const nat def_len = 0;
  static const nat init_len= 1;
};

#undef TMPL
#undef TMPLX
} // namespace mmx
#endif //__MMX__VECTOR_THREADS__HPP
