
/******************************************************************************
* MODULE     : analytic_linear.hpp
* DESCRIPTION: Vectors and matrices of analytic functions
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_ANALYTIC_LINEAR_HPP
#define __MMX_ANALYTIC_LINEAR_HPP
#include <analyziz/analytic.hpp>
#include <algebramix/series_linear.hpp>
namespace mmx {
#define TMPL_DEF template<typename C, typename V, typename W>
#define TMPL template<typename C, typename V, typename W>
#define Analytic analytic<C,V>
#define Analytic_rep analytic_rep<C,V>
#define Assume assume_bound<typename unvectorize<Error >::val>
#define Scalar typename unvectorize<C>::val
#define Series series<C>
#define Error Abs_type(C)

/******************************************************************************
* Prototypes
******************************************************************************/

TMPL Analytic access (const analytic<vector<C,W>,V>& f, nat i);
TMPL vector<Analytic,W> as_vector (const analytic<vector<C,W>,V>& f);
TMPL analytic<vector<C,W>,V> as_analytic (const vector<Analytic,W>& v);

TMPL analytic<vector<C,W>,V> trig (const analytic<vector<C,W>,V>& f);
TMPL vector<Analytic> cos_sin (const Analytic& f);

TMPL analytic<vector<C,W>,V>
solve_vector_lde_init (const analytic<vector<C,W>,V>& f, const vector<C,W>& c);
TMPL vector<Analytic,W>
solve_lde_init (const vector<Analytic,W>& f, const vector<C,W>& c);

TMPL Analytic access (const analytic<matrix<C,W>,V>& f, nat i, nat j);
TMPL matrix<Analytic,W> as_matrix (const analytic<matrix<C,W>,V>& f);
TMPL analytic<matrix<C,W>,V> as_analytic (const matrix<Analytic,W>& m);

TMPL analytic<matrix<C,W>,V>
solve_matrix_lde_init (const analytic<matrix<C,W>,V>& f, const matrix<C,W>& c);
TMPL matrix<Analytic,W>
solve_lde (const matrix<Analytic,W>& f);
TMPL matrix<Analytic,W>
solve_lde_init (const matrix<Analytic,W>& f, const matrix<C,W>& c);

/******************************************************************************
* Vectors of analytic functions
******************************************************************************/

#define Analytic_vector analytic<vector<C,W>,V>
#define Vector_analytic vector<analytic<C,V>,W>
STYPE_TO_TYPE(TMPL,as_vector_type,Analytic_vector,Vector_analytic);
#undef Vector_analytic
#undef Analytic_vector

template<typename C, typename W>
struct unvectorize<vector<C,W> > {
  // FIXME: recursion, etc.
  typedef C val;
  static inline vector<val> encode (const vector<C,W>& x) {
    return x; }
  static inline vector<C,W>
  decode (const vector<val>& v, const vector<C,W>& gauge) {
    (void) gauge;
    return v; }
};

TMPL
class vector_access_analytic_rep: public Analytic_rep {
protected:
  const analytic<vector<C,W>,V> f;
  nat i;
public:
  vector_access_analytic_rep (const analytic<vector<C,W>,V>& f2, nat i2):
    f (f2), i (i2) {}
  Series Expand () const {
    return access (expand (f), i); }
  Error Tail_bound (nat order, Assume& a) const {
    return read (tail_bound (this->f, order, a), i); }
  Analytic Focus (const Scalar& z) const {
    return access (focus (f, z), i); }
  Analytic Move (const Scalar& z) const {
    return access (move (f, z), i); }
  Analytic Derive () const {
    return access (derive (f), i); }
};

TMPL Analytic
access (const analytic<vector<C,W>,V>& f, nat i) {
  return (Analytic_rep*) new vector_access_analytic_rep<C,V,W> (f, i);
}

TMPL vector<Analytic,W>
as_vector (const analytic<vector<C,W>,V>& f) {
  nat n= N(f[0]);
  vector<Analytic,W> v (Analytic (0), n);
  for (nat i=0; i<n; i++)
    v[i]= access (f, i);
  return v;
}

TMPL
class vector_analytic_rep: public analytic_rep<vector<C,W>,V> {
  const vector<Analytic,W> v;
public:
  vector_analytic_rep (const vector<Analytic,W>& v2): v (v2) {}
  series<vector<C,W> > Expand () const {
    nat i, n= N(v);
    vector<Series,W> r (Series (0), n);
    for (i=0; i<n; i++)
      r[i]= expand (v[i]);
    return as_series (r); }
  vector<Error,W> Tail_bound (nat order, Assume& a) const {
    nat i, n= N(v);
    vector<Error,W> r (Error (0), n);
    for (i=0; i<n; i++)
      r[i]= tail_bound (v[i], order, a);
    return r; }
  analytic<vector<C,W>,V> Focus (const Scalar& z) const {
    nat i, n= N(v);
    vector<Analytic,W> r (Analytic (0), n);
    for (i=0; i<n; i++)
      r[i]= focus (v[i], z);
    return as_analytic (r); }
  analytic<vector<C,W>,V> Move (const Scalar& z) const {
    nat i, n= N(v);
    vector<Analytic,W> r (Analytic (0), n);
    for (i=0; i<n; i++)
      r[i]= move (v[i], z);
    return as_analytic (r); }
  analytic<vector<C,W>,V> Derive () const {
    return as_analytic (derive (v)); }
};

TMPL analytic<vector<C,W>,V>
as_analytic (const vector<Analytic,W>& v) {
  return (analytic_rep<vector<C,W>,V>*) new vector_analytic_rep<C,V,W> (v);
}

TMPL analytic<vector<C,W>,V>
from_vector (const vector<Analytic,W>& v) {
  return (analytic_rep<vector<C,W>,V>*) new vector_analytic_rep<C,V,W> (v);
}

/******************************************************************************
* Trigonometric functions and linear differential equations
******************************************************************************/

TMPL analytic<vector<C,W>,V>
trig (const analytic<vector<C,W>,V>& f) {
  return unary_implicit_analytic<trig_op> (f);
}

template<typename C,typename V> vector<Analytic >
cos_sin (const Analytic& f) {
  return as_vector (trig (as_analytic (vec (f))));
}

template<typename C,typename V> Analytic
cos (const Analytic& f) {
  return unary_implicit_analytic<cos_op> (f);
}

template<typename C,typename V> Analytic
sin (const Analytic& f) {
  return unary_implicit_analytic<sin_op> (f);
}

template<typename C,typename V> Analytic
tan (const Analytic& f) {
  return unary_implicit_analytic<tan_op> (f);
}

TMPL analytic<vector<C,W>,V>
solve_vector_lde_init (const analytic<vector<C,W>,V>& f, const vector<C,W>& c) {
  return unary_implicit_analytic<solve_vector_lde_op> (f, c);
}

TMPL vector<Analytic,W>
solve_lde_init (const vector<Analytic,W>& f, const vector<C,W>& c) {
  return as_vector (solve_vector_lde_init (as_analytic (f), c));
}

/******************************************************************************
* Matrices of analytic functions
******************************************************************************/

#define Analytic_matrix analytic<matrix<C,W>,V>
#define Matrix_analytic matrix<analytic<C,V>,W>
STYPE_TO_TYPE(TMPL,as_matrix_type,Analytic_matrix,Matrix_analytic);
#undef Matrix_analytic
#undef Analytic_matrix

template<typename C, typename W>
struct unvectorize<matrix<C,W> > {
  // FIXME: recursion, etc.
  typedef C val;
  static inline vector<val> encode (const matrix<C,W>& x) {
    nat nr= rows (x), nc= cols (x);
    vector<val> r (C(0), nr*nc);
    for (nat i=0; i<nr; i++)
      for (nat j=0; j<nc; j++)
	r[i*nc+j]= x(i,j);
    return r; }
  static inline matrix<C,W>
  decode (const vector<val>& v, const matrix<C,W>& gauge) {
    nat nr= rows (gauge), nc= cols (gauge);
    matrix<C,W> r (C(0), nr, nc);
    for (nat i=0; i<nr; i++)
      for (nat j=0; j<nc; j++)
	r (i, j)= v[i*nc+j];
    return r; }
};

TMPL
class matrix_access_analytic_rep: public Analytic_rep {
protected:
  const analytic<matrix<C,W>,V> f;
  nat i, j;
public:
  matrix_access_analytic_rep (const analytic<matrix<C,W>,V>& f2, nat i2, nat j2):
    f (f2), i (i2), j (j2) {}
  Series Expand () const {
    return access (expand (f), i, j); }
  Error Tail_bound (nat order, Assume& a) const {
    return read (tail_bound (this->f, order, a), i, j); }
  Analytic Focus (const Scalar& z) const {
    return access (focus (f, z), i, j); }
  Analytic Move (const Scalar& z) const {
    return access (move (f, z), i, j); }
  Analytic Derive () const {
    return access (derive (f), i, j); }
};

TMPL Analytic
access (const analytic<matrix<C,W>,V>& f, nat i, nat j) {
  return (Analytic_rep*) new matrix_access_analytic_rep<C,V,W> (f, i, j);
}

TMPL matrix<Analytic,W>
as_matrix (const analytic<matrix<C,W>,V>& f) {
  nat nr= rows (f[0]), nc= cols (f[0]);
  matrix<Analytic,W> m (Analytic (0), nr, nc);
  for (nat i=0; i<nr; i++)
    for (nat j=0; j<nc; j++)
      m (i, j)= access (f, i, j);
  return m;
}

TMPL
class matrix_analytic_rep: public analytic_rep<matrix<C,W>,V> {
  const matrix<Analytic,W> m;
public:
  matrix_analytic_rep (const matrix<Analytic,W>& m2): m (m2) {}
  series<matrix<C,W> > Expand () const {
    nat i, j, nr= rows (m), nc= cols (m);
    matrix<Series,W> r (Series (0), nr, nc);
    for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	r (i, j)= expand (m (i, j));
    return as_series (r); }
  matrix<Error,W> Tail_bound (nat order, Assume& a) const {
    nat i, j, nr= rows (m), nc= cols (m);
    matrix<Error,W> r (Error (0), nr, nc);
    for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	r (i, j)= tail_bound (m (i, j), order, a);
    return r; }
  analytic<matrix<C,W>,V> Focus (const Scalar& z) const {
    nat i, j, nr= rows (m), nc= cols (m);
    matrix<Analytic,W> r (Analytic (0), nr, nc);
    for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	r (i, j)= focus (m (i, j), z);
    return as_analytic (r); }
  analytic<matrix<C,W>,V> Move (const Scalar& z) const {
    nat i, j, nr= rows (m), nc= cols (m);
    matrix<Analytic,W> r (Analytic (0), nr, nc);
    for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	r (i,j)= move (m (i, j), z);
    return as_analytic (r); }
  analytic<matrix<C,W>,V> Derive () const {
    return as_analytic (derive (m)); }
};

TMPL analytic<matrix<C,W>,V>
as_analytic (const matrix<Analytic,W>& m) {
  return (analytic_rep<matrix<C,W>,V>*) new matrix_analytic_rep<C,V,W> (m);
}

TMPL analytic<matrix<C,W>,V>
from_matrix (const matrix<Analytic,W>& m) {
  return (analytic_rep<matrix<C,W>,V>*) new matrix_analytic_rep<C,V,W> (m);
}

TMPL matrix<Analytic,W>
move (const matrix<Analytic,W>& m, const C& z) {
  nat nr= rows (m), nc= cols (m);
  matrix<Analytic,W> r (Analytic (0), nr, nc);
  for (nat i=0; i<nr; i++)
    for (nat j=0; j<nc; j++)
      r (i, j)= move (m (i, j), z);
  return r;  
}

TMPL matrix<C,W>
eval (const matrix<Analytic,W>& m, const C& z) {
  nat nr= rows (m), nc= cols (m);
  matrix<C,W> r (C (0), nr, nc);
  for (nat i=0; i<nr; i++)
    for (nat j=0; j<nc; j++)
      r (i, j)= eval (m (i, j), z);
  return r;  
}

/******************************************************************************
* Linear differential matrix equations
******************************************************************************/

TMPL analytic<matrix<C,W>,V>
solve_matrix_lde_init (const analytic<matrix<C,W>,V>& f, const matrix<C,W>& c) {
  return unary_implicit_analytic<solve_matrix_lde_op> (f, c);
}

TMPL matrix<Analytic,W>
solve_lde (const matrix<Analytic,W>& f) {
  ASSERT (is_square_matrix (f), "square matrix expected");
  matrix<C,W> c (C(1), rows (f), cols (f));
  return as_matrix (solve_matrix_lde_init (as_analytic (f), c));
}

TMPL matrix<Analytic,W>
solve_lde_init (const matrix<Analytic,W>& f, const matrix<C,W>& c) {
  ASSERT (is_square_matrix (f), "square matrix expected");
  ASSERT (rows (f) == rows (c), "dimensions don't match");
  return as_matrix (solve_matrix_lde_init (as_analytic (f), c));
}

/******************************************************************************
* Connection matrices
******************************************************************************/

template<typename C> list<C>
straight_path (const C& start, const C& end, nat steps= 5) {
  list<C> r;
  for (nat i=steps; i>0; i--)
    r= cons (start + (i*(end-start)) / steps, r);
  return r;
}

template<typename C> list<C>
circle_path (const C& start, const C& center, int orient= 1, nat steps= 17) {
  list<C> r;
  for (int i=0; ((nat) i)<steps; i++) {
    C z= exp ((C(-orient*2*i) * Pi(C) * Imaginary(C)) / C(steps));
    r= cons (center + z * (start - center), r);
  }
  return r;
}

template<typename C> list<C>
delta_to_absolute (const list<C>& l, const C& pos= C(0)) {
  if (is_nil (l)) return l;
  else return cons (pos + car (l), delta_to_absolute (cdr (l), pos + car (l)));
}

template<typename C> list<C>
absolute_to_delta (const list<C>& l, const C& pos= C(0)) {
  if (is_nil (l)) return l;
  else return cons (car (l) - pos, absolute_to_delta (cdr (l), car (l)));
}

template<typename T, typename C> T
move (const T& x, const list<C>& p) {
  if (is_nil (p)) return x;
  else return move (move (x, car (p)), cdr (p));
}

TMPL matrix<C,W>
connection_matrix (const matrix<Analytic,W>& m, const list<C>& p) {
  nat n= N(p);
  if (n <= 1) {
    ASSERT (is_square_matrix (m), "square matrix expected");
    if (n == 0) return matrix<C,W> (C(1), rows (m), cols (m));
    else return eval (solve_lde (m), car (p));
  }
  else {
    list<C> p1= range (p, 0, n>>1);
    list<C> p2= range (p, n>>1, n);
    matrix<C,W> r1= connection_matrix (m, p1);
    matrix<C,W> r2= connection_matrix (move (m, p1), p2);
    return r2 * r1;
  }
}

#undef TMPL_DEF
#undef TMPL
#undef Analytic
#undef Analytic_rep
#undef Assume
#undef Scalar
#undef Series
#undef Error
} // namespace mmx
#endif // __MMX_ANALYTIC_LINEAR_HPP
