
/******************************************************************************
* MODULE     : riemann_point.hpp
* DESCRIPTION: points on Riemann surfaces
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_RIEMANN_POINT_HPP
#define __MMX_RIEMANN_POINT_HPP
#include <analyziz/riemann_square.hpp>

namespace mmx {
#define TMPL_DEF template<typename R,typename V= std_riemann>
#define TMPL template<typename R,typename V>
#define C complex<R>
#define Square riemann_square<R,V>
#define Point riemann_point<R,V>

/******************************************************************************
* Points on Riemann surfaces
******************************************************************************/

TMPL_DEF
class riemann_point {
MMX_ALLOCATORS;
public:
  Square s;
  C      z;
  inline riemann_point (): z (0) {}
  inline riemann_point (const Square& s2, const C& z2): s (s2), z (z2) {}
};

TMPL inline Square owner (const Point& p) { return p.s; }
TMPL inline C project (const Point& p) { return p.z; }

/******************************************************************************
* Required routines
******************************************************************************/

template<typename Op, typename R, typename V> nat
unary_hash (const Point& p) {
  nat h= Op::op (p.s);
  return (h<<1) ^ (h<<3) ^ (h<<9) ^ (h>>29) ^ Op::op (p.z);
}

template<typename Op, typename R, typename V> bool
binary_test (const Point& p1, const Point& p2) {
  return Op::op (p1.s, p2.s) && Op::op (p1.z, p2.z);
}

TMPL inline syntactic
flatten (const Point& p) {
  return apply ("point", flatten (p.z), flatten (p.s));
}

TRUE_IDENTITY_OP_SUGAR(TMPL,Point)
SOFT_IDENTITY_OP_SUGAR(TMPL,Point)
HARD_IDENTITY_OP_SUGAR(TMPL,Point)
PRINT_SUGAR(TMPL,Point)

#undef TMPL_DEF
#undef TMPL
#undef C
#undef Square
#undef Point
} // namespace mmx
#endif // __MMX_RIEMANN_POINT_HPP
