
/******************************************************************************
* MODULE     : compound.hpp
* DESCRIPTION: Compound generic expressions
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __COMPOUND_HPP
#define __COMPOUND_HPP
#include <basix/vector.hpp>
#include <basix/symbol.hpp>

//! \file

namespace mmx {

/******************************************************************************
* Compound expressions
******************************************************************************/

class compound {
MMX_ALLOCATORS
  vector<generic> v;
  symbol<vector<generic> > sym;
public:
  inline vector<generic> operator * () const { return v; }
  inline friend vector<generic> as_vector (const compound& v) { return v.v; }
  inline friend symbol<vector<generic> > as_symbol (const compound& v) {
    return v.sym; }
  inline compound (): v (), sym (v) {}
  inline compound (const vector<generic>& v2): v (v2), sym (v) {}
  inline compound (const compound& c): v (c.v), sym (c.sym) {}
  inline compound (const symbol<vector<generic> >& s2): v (*s2), sym (s2) {}
};

inline nat hash (const compound& c) { return hash (as_symbol (c)); }
inline nat soft_hash (const compound& c) { return soft_hash (as_symbol (c)); }
inline nat hard_hash (const compound& c) { return hard_hash (as_vector (c)); }
inline bool operator == (const compound& c1, const compound& c2) {
  return as_symbol (c1) == as_symbol (c2); }
inline bool operator != (const compound& c1, const compound& c2) {
  return as_symbol (c1) != as_symbol (c2); }
inline bool eq (const compound& c1, const compound& c2) {
  return eq (as_symbol (c1), as_symbol (c2)); }
inline bool neq (const compound& c1, const compound& c2) {
  return neq (as_symbol (c1), as_symbol (c2)); }
inline bool hard_eq (const compound& c1, const compound& c2) {
  return hard_eq (as_vector (c1), as_vector (c2)); }
inline bool hard_neq (const compound& c1, const compound& c2) {
  return hard_neq (as_vector (c1), as_vector (c2)); }

vector<generic> compound_to_vector (const generic& g);
generic vector_to_compound (const vector<generic>& v);
syntactic flatten (const compound& c);
PRINT_SUGAR(,compound)

STMPL
struct inspector<compound> {
  static inline nat length (const compound& v) {
    return mmx::N (as_vector (v)); }
  static inline generic access (const compound& v, nat i) {
    VERIFY (i < N (as_vector (v)), "out of range");
    return read (as_vector (v), i); }
};

} // namespace mmx
#endif // __COMPOUND_HPP
