
/******************************************************************************
* MODULE     : defaults.hpp
* DESCRIPTION: Templates and default implementations of several operations
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_DEFAULTS_HPP
#define __MMX_DEFAULTS_HPP
#include <basix/basix.hpp>

//! \file

namespace mmx {

/******************************************************************************
* Hash functions and equality testing
******************************************************************************/

typedef void* void_ptr;
typedef char* char_ptr;

inline nat hash (char c) { return (nat) c; }
inline nat hash (signed char c) { return (nat) c; }
inline nat hash (unsigned char c) { return (nat) c; }
inline nat hash (short int c) { return (nat) c; }
inline nat hash (short unsigned int c) { return (nat) c; }
inline nat hash (int c) { return (nat) c; }
inline nat hash (unsigned int c) { return (nat) c; }
inline nat hash (long int c) { return (nat) c; }
inline nat hash (long unsigned int c) { return (nat) c; }
inline nat hash (long long int c) {
  nat h = (nat) ((c >> (4 * sizeof (long int))) >> (4 * sizeof (long int)));
  return (h<<1) ^ (h<<5) ^ (h>>27) ^ ((nat) c); }
inline nat hash (long long unsigned int c) {
  nat h = (nat) ((c >> (4 * sizeof (long int))) >> (4 * sizeof (long int)));
  return (h<<1) ^ (h<<5) ^ (h>>27) ^ ((nat) c); }
inline nat hash (const float& x) {
  return (*((nat*) ((void*) &x))) & 0xffffffff; }
inline nat hash (const double& x) { return *((nat*) ((void*) &x)); }
template<typename C> inline nat hash (C* p) { return as_hash (p); }

inline nat soft_hash (char c) { return (nat) c; }
inline nat soft_hash (signed char c) { return (nat) c; }
inline nat soft_hash (unsigned char c) { return (nat) c; }
inline nat soft_hash (short int c) { return (nat) c; }
inline nat soft_hash (short unsigned int c) { return (nat) c; }
inline nat soft_hash (int c) { return (nat) c; }
inline nat soft_hash (unsigned int c) { return (nat) c; }
inline nat soft_hash (long int c) { return (nat) c; }
inline nat soft_hash (long unsigned int c) { return (nat) c; }
inline nat soft_hash (long long int c) {
  nat h = (nat) ((c >> (4 * sizeof (long int))) >> (4 * sizeof (long int)));
  return (h<<1) ^ (h<<5) ^ (h>>27) ^ ((nat) c); }
inline nat soft_hash (long long unsigned int c) {
  nat h = (nat) ((c >> (4 * sizeof (long int))) >> (4 * sizeof (long int)));
  return (h<<1) ^ (h<<5) ^ (h>>27) ^ ((nat) c); }

inline nat soft_hash (const float& x) {
  return (*((nat*) ((void*) &x))) & 0xffffffff; }
inline nat soft_hash (const double& x) { return *((nat*) ((void*) &x)); }
template<typename C> inline nat soft_hash (C* p) { return as_hash (p); }

inline bool eq (char c1, char c2) { return c1 == c2; }
inline bool eq (signed char c1, signed char c2) { return c1 == c2; }
inline bool eq (unsigned char c1, unsigned char c2) { return c1 == c2; }
inline bool eq (short int c1, short int c2) { return c1 == c2; }
inline bool eq (short unsigned int c1, short unsigned int c2) {
  return c1 == c2; }
inline bool eq (int c1, int c2) { return c1 == c2; }
inline bool eq (unsigned int c1, unsigned int c2) { return c1 == c2; }
inline bool eq (long int c1, long int c2) { return c1 == c2; }
inline bool eq (long unsigned int c1, long unsigned int c2) {
  return c1 == c2; }
inline bool eq (long long int c1, long long int c2) { return c1 == c2; }
inline bool eq (long long unsigned int c1, long long unsigned int c2) {
  return c1 == c2; }

inline bool eq (float x, float y) { return x == y; }
inline bool eq (double x, double y) { return x == y; }
template<typename C> inline bool eq (C* p1, C* p2) { return p1 == p2; }


inline bool neq (char c1, char c2) { return c1 != c2; }
inline bool neq (signed char c1, signed char c2) { return c1 != c2; }
inline bool neq (unsigned char c1, unsigned char c2) { return c1 != c2; }
inline bool neq (short int c1, short int c2) { return c1 != c2; }
inline bool neq (short unsigned int c1, short unsigned int c2) {
  return c1 != c2; }
inline bool neq (int c1, int c2) { return c1 != c2; }
inline bool neq (unsigned int c1, unsigned int c2) { return c1 != c2; }
inline bool neq (long int c1, long int c2) { return c1 != c2; }
inline bool neq (long unsigned int c1, long unsigned int c2) {
  return c1 != c2; }
inline bool neq (long long int c1, long long int c2) { return c1 != c2; }
inline bool neq (long long unsigned int c1, long long unsigned int c2) {
  return c1 != c2; }

inline bool neq (float x, float y) { return x != y; }
inline bool neq (double x, double y) { return x != y; }
template<typename C> inline bool neq (C* p1, C* p2) { return p1 != p2; }

#define TRUE_TO_SOFT_IDENTITY_SUGAR(TMPL,T)	\
TMPL inline nat soft_hash (const T& x) {	\
  return hash (x); }				\
TMPL inline bool eq (const T& x, const T& y) {	\
  return x == y; }				\
TMPL inline bool neq (const T& x, const T& y) {	\
  return x != y; }

#define HARD_TO_SOFT_IDENTITY_SUGAR(TMPL,T)	\
TMPL inline nat soft_hash (const T& x) {	\
  return hard_hash (x); }			\
TMPL inline bool eq (const T& x, const T& y) {	\
  return hard_eq (x, y); }			\
TMPL inline bool neq (const T& x, const T& y) {	\
  return hard_neq (x, y); }

#define HARD_TO_TRUE_IDENTITY_SUGAR(TMPL,T)     	\
TMPL inline nat hash (const T& x) {             	\
  return hard_hash (x); }       			\
TMPL inline bool operator == (const T& x, const T& y) {	\
  return hard_eq (x, y); }      			\
TMPL inline bool operator != (const T& x, const T& y) {	\
  return hard_neq (x, y); }

#define TRUE_IDENTITY_OP_SUGAR(TMPL,T)			\
TMPL inline nat hash (const T& x) {			\
  return unary_hash<hash_op> (x); }			\
TMPL inline bool operator == (const T& x, const T& y) {	\
  return binary_test<equal_op> (x, y); }		\
TMPL inline bool operator != (const T& x, const T& y) {	\
  return !binary_test<equal_op> (x, y);	}

#define SOFT_IDENTITY_OP_SUGAR(TMPL,T)		\
TMPL inline nat soft_hash (const T& x) {	\
  return unary_hash<soft_hash_op> (x); }	\
TMPL inline bool eq (const T& x, const T& y) {	\
  return binary_test<eq_op> (x, y); }		\
TMPL inline bool neq (const T& x, const T& y) {	\
  return !binary_test<eq_op> (x, y); }

#define HARD_IDENTITY_OP_SUGAR(TMPL,T)			\
TMPL inline nat hard_hash (const T& x) {		\
  return unary_hash<hard_hash_op> (x); }		\
TMPL inline bool hard_eq (const T& x, const T& y) {	\
  return binary_test<hard_eq_op> (x, y); }		\
TMPL inline bool hard_neq (const T& x, const T& y) {	\
  return !binary_test<hard_eq_op> (x, y); }

#define EQUAL_INT_SUGAR(TMPL,T,C)					      \
TMPL inline bool operator == (const T& x, const int& y) { return x == C(y); } \
TMPL inline bool operator != (const T& x, const int& y) { return x != C(y); } \
TMPL inline bool eq (const T& x, const int& y) { return eq (x, C(y)); }	      \
TMPL inline bool neq (const T& x, const int& y) { return neq (x, C(y)); }

/******************************************************************************
* Printing
******************************************************************************/

#define PRINT_SUGAR(TMPL,T)				        \
TMPL inline ostream& operator << (ostream& out, const T& x) {	\
  out << as_generic (flatten (x));                              \
  return out;                                                   \
}

/******************************************************************************
* Copy
******************************************************************************/

inline char copy (char c) { return c; }
inline signed char copy (signed char c) { return c; }
inline unsigned char copy (unsigned char c) { return c; }
inline short int copy (short int c) { return c; }
inline short unsigned int copy (short unsigned int c) { return c; }
inline int copy (int c) { return c; }
inline unsigned int copy (unsigned int c) { return c; }
inline long int copy (long int c) { return c; }
inline long unsigned int copy (long unsigned int c) { return c; }
inline long long int copy (long long int c) { return c; }
inline long long unsigned int copy (long long unsigned int c) { return c; }

inline float copy (float c) { return c; }
inline double copy (double c) { return c; }

template<typename C> inline void* copy (C* p) { return p; }
template<typename C> inline C duplicate (const C& x) { return x; }

#define COPY_OP_SUGAR(TMPL,T)			\
TMPL inline T copy (const T& x) {		\
  return unary_map<id_op> (x); }  	        \
TMPL inline T duplicate (const T& x) {		\
  return unary_map<duplicate_op> (x); }

/******************************************************************************
* Arithmetic
******************************************************************************/

#define ARITH_INT_SUGAR(TMPL,T,C)					 \
TMPL inline T operator + (const T& x, const int& y) { return x + T(y); } \
TMPL inline T operator + (const int& x, const T& y) { return T(x) + y; } \
TMPL inline T operator - (const T& x, const int& y) { return x - T(y); } \
TMPL inline T operator - (const int& x, const T& y) { return T(x) - y; } \
TMPL inline T operator * (const T& x, const int& y) { return x * C(y); } \
TMPL inline T operator * (const int& x, const T& y) { return C(x) * y; } \
TMPL inline T operator / (const T& x, const int& y) { return x / C(y); } \
TMPL inline T operator / (const int& x, const T& y) { return T(x) / y; }

#define VAR_ARITH_INT_SUGAR(TMPL,T)					 \
TMPL inline T operator + (const T& x, const int& y) { return x + T(y); } \
TMPL inline T operator + (const int& x, const T& y) { return T(x) + y; } \
TMPL inline T operator - (const T& x, const int& y) { return x - T(y); } \
TMPL inline T operator - (const int& x, const T& y) { return T(x) - y; }

#define SHIFT_INT_SUGAR(TMPL,T) 					\
TMPL inline T& operator <<= (T& x, const int& y) { return x= x << y; }	\
TMPL inline T& operator >>= (T& x, const int& y) { return x= x >> y; }

#define ARITH_SCALAR_SUGAR(TMPL,T,C)				        \
TMPL inline T operator - (const T& x) { return T(0) - x; }		\
TMPL inline T operator * (const T& x, const C& y) { return x * T(y); }	\
TMPL inline T operator * (const C& x, const T& y) { return T(x) * y; }	\
TMPL inline T operator / (const T& x, const C& y) { return x / T(y); }

#define ARITH_SWAP_TIMES_SUGAR(TMPL,T,C)				\
TMPL inline T operator * (const C& x, const T& y) { return y * x; }

template<typename C, typename D> inline C&
operator += (C& x, const D& y) { return x= x+y; }
template<typename C, typename D> inline C&
operator -= (C& x, const D& y) { return x= x-y; }
template<typename C, typename D> inline C&
operator *= (C& x, const D& y) { return x= x*y; }
template<typename C, typename D> inline C&
operator /= (C& x, const D& y) { return x= x/y; }

template<typename C, typename D> inline C
quo (const C& x, const D& y) { // NOTE: by default we divide on the right
  if (y == D(0)) return C(0); else return x/y; }
template<typename C, typename D> inline C
rem (const C& x, const D& y) { // NOTE: by default we divide on the right
  if (y == D(0)) return x; else return C(0); }
template<typename C> inline C
skew_div (const C& x, const C& y, bool left= false) {
  (void) left; return x / y; }
template<typename C> inline C
skew_quo (const C& x, const C& y, bool left= false) {
  (void) left; return quo (x, y); }
template<typename C> inline C
skew_rem (const C& x, const C& y, bool left= false) {
  (void) left; return rem (x, y); }

/******************************************************************************
* Shifting
******************************************************************************/

template<typename C, typename S> inline C
lshift2 (const C& x, const S& y) { return x << y; }
template<typename C> inline C
lshift2 (const C& x) { return lshift2<C,int> (x, 1); }
template<typename C, typename S> inline void
lshift2_assign (C& x, const S& y) { x <<= y; }
template<typename C, typename S> inline void
lshift2 (C& x, const C& y, const S& z) { x= y << z; }
template<typename C, typename S> inline C
rshift2 (const C& x, const S& y) { return x >> y; }
template<typename C> inline C
rshift2 (const C& x) { return rshift2<C,int> (x, 1); }
template<typename C, typename S> inline void
rshift2_assign (C& x, const S& y) { x >>= y; }
template<typename C, typename S> inline void
rshift2 (C& x, const C& y, const S& z) { x= y >> z; }

template<typename C, typename S> inline C
lshiftz (const C& x, const S& y) { return x << y; }
template<typename C> inline C
lshiftz (const C& x) { return lshiftz<C,int> (x, 1); }
template<typename C, typename S> inline void
lshiftz_assign (C& x, const S& y) { x <<= y; }
template<typename C, typename S> inline void
lshiftz (C& x, const C& y, const S& z) { x= y << z; }
template<typename C, typename S> inline C
rshiftz (const C& x, const S& y) { return lshiftz (x, -y); }
template<typename C> inline C
rshiftz (const C& x) { return rshiftz<C,int> (x, 1); }
template<typename C, typename S> inline void
rshiftz_assign (C& x, const S& y= 1) { lshiftz_assign (x, -y); }
template<typename C, typename S> inline void
rshiftz (C& x, const C& y, const S& z) { lshiftz (x, y, -z); }

template<typename C, typename S> inline C
incexp2 (const C& x, const S& y) { return x << y; }
template<typename C> inline C
incexp2 (const C& x) { return incexp2<C,int> (x, 1); }
template<typename C, typename S> inline void
incexp2_assign (C& x, const S& y) { x <<= y; }
template<typename C, typename S> inline void
incexp2 (C& x, const C& y, const S& z) { x= y << z; }
template<typename C, typename S> inline C
decexp2 (const C& x, const S& y) { return x >> y; }
template<typename C> inline C
decexp2 (const C& x) { return decexp2<C,int> (x, 1); }
template<typename C, typename S> inline void
decexp2_assign (C& x, const S& y) { x >>= y; }
template<typename C, typename S> inline void
decexp2 (C& x, const C& y, const S& z) { x= y >> z; }

/******************************************************************************
* Greatest common divisors and least common multiples
******************************************************************************/

template<typename C>
struct xgcd_matrix {
MMX_ALLOCATORS
  C a, b, c, d;
  // NOTE: the xgcd matrix is an invertible matrix with
  // a x + b y = gcd (x, y), c x + d y = 0 and d y = lcm (x, y).

  inline xgcd_matrix<C> () {}
  inline xgcd_matrix<C> (const C& a2, const C& b2, const C& c2, const C& d2):
    a (a2), b (b2), c (c2), d (d2) {}
  inline xgcd_matrix<C> (const xgcd_matrix& m):
    a (m.a), b (m.b), c (m.c), d (m.d) {}
  inline xgcd_matrix<C>& operator = (const xgcd_matrix& m) {
    a= m.a; b= m.b; c= m.c; d= m.d; return m; }
};

#define GCD_SUGAR(TMPL,C)					\
TMPL inline C gcd (const C& x, const C& y) {			\
  return C ((x == 0 && y == 0)? 0: 1); }			\
TMPL inline C lcm (const C& x, const C& y) {			\
  return C ((x == 0 || y == 0)? 0: x*y); }			\
TMPL inline xgcd_matrix<C> xgcd (const C& x, const C& y) {	\
  if (x == 0) {							\
    if (y == 0) return xgcd_matrix<C> (C(1), C(0), C(0), C(1));	\
    else return xgcd_matrix<C> (C(0), 1/y, C(1), C(0));		\
  }								\
  else {							\
    if (y == 0) return xgcd_matrix<C> (1/x, C(0), C(0), C(1));	\
    else return xgcd_matrix<C> (1/x, C(0), -y, x);		\
  }								\
}

/******************************************************************************
* Comparisons
******************************************************************************/

#ifndef MMX_SIGN_FCT
#define MMX_SIGN_FCT
template<typename C> inline int
sign (const C& x) {
  if (x>0) return 1;
  if (x<0) return -1;
  return 0;
}

template<> inline int 	 
sign (const unsigned char& x) { return (x == 0) ? 0 : 1; } 	 
  	 
template<> inline int 	 
sign (const short unsigned int& x) { return (x == 0) ? 0 : 1; } 	 
  	 
template<> inline int 	 
sign (const unsigned int& x) { return (x == 0) ? 0 : 1; } 	 
  	 
template<> inline int 	 
sign (const long unsigned int& x) { return (x == 0) ? 0 : 1; } 	 
  	 
template<> inline int 	 
sign (const long long unsigned int& x) { return (x == 0) ? 0 : 1; }
#endif

template<typename C> inline int
var_sign (const C& x) {
  return sign (x);
}

template<typename C> inline int
compare (const C& x, const C& y) {
  return sign (x-y);
}

#ifndef MMX_ABS_FCT
#define MMX_ABS_FCT
template<typename C> inline C
abs (const C& x) { return x>=0? x: -x; }

template<> inline unsigned char 	 
abs (const unsigned char& x) { return x; } 	 
  	 
template<> inline short unsigned int 	 
abs (const short unsigned int& x) { return x; } 	 
  	 
template<> inline unsigned int 	 
abs (const unsigned int& x) { return x; } 	 
  	 
template<> inline long unsigned int 	 
abs (const long unsigned int& x) { return x; } 	 
  	 
template<> inline long long unsigned int 	 
abs (const long long unsigned int& x) { return x; }
#endif

template<typename C> C min (const C& x, const C& y) { return x<y? x: y; }
template<typename C> C max (const C& x, const C& y) { return x>y? x: y; }
template<typename C> C inf (const C& x, const C& y) { return min (x, y); }
template<typename C> C sup (const C& x, const C& y) { return max (x, y); }

#define COMPARE_SUGAR(TMPL,T)				\
TMPL inline bool operator <  (const T& x, const T& y) {	\
  return compare (x,y) <  0; }				\
TMPL inline bool operator <= (const T& x, const T& y) {	\
  return compare (x,y) <= 0; }				\
TMPL inline bool operator >  (const T& x, const T& y) {	\
  return compare (x,y) >  0; }				\
TMPL inline bool operator >= (const T& x, const T& y) {	\
  return compare (x,y) >= 0; }

#define STRICT_COMPARE_SUGAR(TMPL,T)			\
TMPL inline bool operator <  (const T& x, const T& y) {	\
  return x<=y && !(x==y); }				\
TMPL inline bool operator >  (const T& x, const T& y) {	\
  return x>=y && !(x==y); }

#define COMPARE_INT_SUGAR(TMPL,T, C)					      \
TMPL inline bool operator <  (const T& x, const int& y) { return x <  C(y); } \
TMPL inline bool operator <= (const T& x, const int& y) { return x <= C(y); } \
TMPL inline bool operator >  (const T& x, const int& y) { return x >  C(y); } \
TMPL inline bool operator >= (const T& x, const int& y) { return x >= C(y); }

/******************************************************************************
* Elementary functions
******************************************************************************/

template<typename C> inline C
square (const C& x) {
  return x*x;
}

template<typename C> C
binpow (const C& i, const nat& n) {
  // FIXME: how to merge with pow?
  if (n <= 1) return n==0? C(1): i;
  C j= square (binpow (i, n >> 1));
  if ((n&1) == 0) return j;
  else return i * j;
}

#define POOR_MAN_SQRT_SUGAR(TMPL,C)					\
TMPL inline C sqrt (const C& x) {					\
  ASSERT (x == (C)(0) || x == (C)(1), "zero or one expected"); return x; }

#define POOR_MAN_ELEMENTARY_SUGAR(TMPL,C)		\
TMPL inline C exp (const C& x) {			\
  ASSERT (x == (C(0)), "zero expected"); return 1; }	\
TMPL inline C log (const C& x) {			\
  ASSERT (x == (C(1)), "one expected"); return 0; }	\
TMPL inline C cos (const C& x) {			\
  ASSERT (x == (C(0)), "zero expected"); return 1; }	\
TMPL inline C sin (const C& x) {			\
  ASSERT (x == (C(0)), "zero expected"); return 0; }	\
TMPL inline C tan (const C& x) {			\
  ASSERT (x == (C(0)), "zero expected"); return 0; }	\
TMPL inline C acos (const C& x) {			\
  ASSERT (x == (C(1)), "one expected"); return 0; }	\
TMPL inline C asin (const C& x) {			\
  ASSERT (x == (C(0)), "zero expected"); return 0; }	\
TMPL inline C atan (const C& x) {			\
  ASSERT (x == (C(0)), "zero expected"); return 0; }

#define HYPOT_SUGAR(TMPL,C)			\
TMPL inline C hypot (const C& x, const C& y) {	\
  return sqrt (square (x) + square (y)); }

#define POW_SUGAR(TMPL,C)			\
TMPL inline C pow (const C& x, const C& y) {	\
  return exp (y * log (x)); }

#define ATAN2_SUGAR(TMPL,C)						\
TMPL inline C atan2 (const C& y, const C& x) {				\
  if (x > C(0)) return atan (y/x);					\
  else if (y == C(0)) return x == C(0)? C(0): C(-4) * atan (C(1));	\
  else if (x == C(0)) return C (2 * sign (y)) * atan (C(1));		\
  else return sign (y) * (C(4) * atan (C(1)) - atan (abs (y/x))); }

#define INV_TRIGO_SUGAR(TMPL,C)			\
TMPL inline C sec (const C& x) {		\
  return C(1) / cos (x); }			\
TMPL inline C csc (const C& x) {		\
  return C(1) / sin (x); }			\
TMPL inline C cot (const C& x) {		\
  return cos (x) / sin (x); }

#define HYPER_SUGAR(TMPL,C)				\
TMPL inline C cosh (const C& x) {			\
  return (exp (x) + exp (-x)) / C (2); }		\
TMPL inline C sinh (const C& x) {			\
  return (exp (x) - exp (-x)) / C (2); }		\
TMPL inline C tanh (const C& x) {			\
  return (exp (x) - exp (-x)) / (exp (x) + exp (-x)); }

#define INV_HYPER_SUGAR(TMPL,C)				\
TMPL inline C sech (const C& x) {			\
  return C(2) / (exp (x) + exp (-x)); }			\
TMPL inline C csch (const C& x) {			\
  return C(2) / (exp (x) - exp (-x)); }			\
TMPL inline C coth (const C& x) {			\
  return (exp (x) + exp (-x)) / (exp (x) - exp (-x)); }

#define ARG_HYPER_SUGAR(TMPL,C)				\
TMPL inline C acosh (const C& x) {			\
  return log (x + sqrt (square (x) - C (1))); }		\
TMPL inline C asinh (const C& x) {			\
  return log (x + sqrt (square (x) + C (1))); }		\
TMPL inline C atanh (const C& x) {			\
  return log ((C (1) + x) / (C (1) - x)) / C (2); }

/******************************************************************************
* Approximate arithmetic
******************************************************************************/

template<typename C> bool is_exact_zero (const C& x) { return x == C(0); }
template<typename C> bool is_number (const C& x) { (void) x; return true; }
template<typename C> bool is_infinity (const C& x) { (void) x; return false; }
template<typename C> bool is_fuzz (const C& x) { (void) x; return false; }
template<typename C> bool is_nan (const C& x) { (void) x; return false; }

template<typename T> inline double magnitude (const T& x) {
  return (double) exponent (x); }
template<typename T> inline double double_exponent (const T& x) {
  return (double) exponent (x); }

/******************************************************************************
* Calculus
******************************************************************************/

template<typename T, typename V> T
derive (const T& f, const V& v, nat n) {
  if (n == 0) return f;
  else return derive (derive (f, v, n-1), v);
}

/******************************************************************************
* Utilities
******************************************************************************/

template<typename C> inline bool
always_false (const C& x) {
  (void) x; return false;
}

template<typename C> inline bool
always_true (const C& x) {
  (void) x; return true;
}

template<typename C> inline void
swap (C& x, C& y) {
  C temp= x;
  x= y;
  y= temp;
}

template<typename C> inline const C&
read_only (const C& c) {
  return c;
}

template<typename C> inline void
force_error (const C* ptr) {
  exp (ptr) + ptr;
}

} // namespace mmx
#endif // __MMX_DEFAULTS_HPP
