
/******************************************************************************
* MODULE     : dynamic.hpp
* DESCRIPTION: Dynamic objects, capable of updating themselves automatically
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __DYNAMIC_HPP
#define __DYNAMIC_HPP
#include <basix/routine.hpp>

//! \file

namespace mmx {
class observer_rep;
class observer;
class dynamic_rep;
class dynamic;

/******************************************************************************
* Observers of dynamic objects
******************************************************************************/

class observer_rep: public rep_struct {
public:
  dynamic_rep* ref;

  inline observer_rep (dynamic_rep* ref2): ref (ref2) {}
  inline virtual ~observer_rep () {}

  virtual void destroy () const = 0;
  virtual void modify () const = 0;
  virtual inline void message (const generic& msg) const { (void) msg; }

  friend class observer;
  friend class dynamic;
};

class observer {
INDIRECT_PROTO (observer, observer_rep)
public:
  inline observer (): rep (NULL) {}
  inline friend observer_rep* inspect (const observer& g) {
    return g.rep; }
  inline friend syntactic flatten (const observer& g) {
    return "observer"; }
  friend class observer_rep;
};
INDIRECT_NULL_IMPL (observer, observer_rep)

HARD_TO_SOFT_IDENTITY_SUGAR(,observer);
HARD_TO_TRUE_IDENTITY_SUGAR(,observer);
PRINT_SUGAR(,observer);

/******************************************************************************
* The dynamic class
******************************************************************************/

class dynamic_rep: public rep_struct {
public:
  generic val;
  vector<observer> obs;

  dynamic_rep (const generic& val2);
  ~dynamic_rep ();

  friend class observer;
  friend class dynamic;
};

class dynamic {
INDIRECT_PROTO (dynamic, dynamic_rep)
public:
  inline dynamic (): rep (NULL) {}
  inline dynamic (const generic& val): rep (new dynamic_rep (val)) {}
  dynamic (const routine& fun, const vector<dynamic>& args);
  inline friend dynamic_rep* inspect (const dynamic& d) { return d.rep; }
  friend class dynamic_rep;
};
INDIRECT_NULL_IMPL (dynamic, dynamic_rep)

syntactic flatten (const dynamic& g);
inline generic contents (const dynamic& d) { return inspect(d)->val; }
void assign (const dynamic& d, const generic& val);
void message (const dynamic& d, const generic& msg);
void attach (const dynamic& d, const observer& o);

HARD_TO_SOFT_IDENTITY_SUGAR(,dynamic);
HARD_TO_TRUE_IDENTITY_SUGAR(,dynamic);
PRINT_SUGAR(,dynamic);

} // namespace mmx
#endif // __DYNAMIC_HPP
