
/******************************************************************************
* MODULE     : glue.hpp
* DESCRIPTION: Routines for defining glue
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __GLUE_HPP
#define __GLUE_HPP
#include <basix/evaluator.hpp>
#include <basix/routine.hpp>
#include <basix/primitive.hpp>
#include <basix/tuple.hpp>
#include <basix/alias.hpp>

//! \file

namespace mmx {

template<typename C> inline generic type_name ();
void define_prerequisites ();

/******************************************************************************
* Main functions for gluing routines to the current evaluator
******************************************************************************/

template<typename D> inline void
define_constant (const generic& name, const D& x) {
  current_ev->set (name, as<generic> (x));
}

template<typename D> inline void
define_constructor (generic (*fun) (const D&)) {
  current_ev->overload (GEN_NEW, as<generic> (unary_routine (GEN_NEW, fun)));
}

template<typename D> inline void
define (const generic& name, D (*fun) ()) {
  current_ev->overload (name, as<generic> (nullary_routine (name, fun)));
}

template<typename D, typename S1> inline void
define (const generic& name, D (*fun) (const S1&)) {
  current_ev->overload (name, as<generic> (unary_routine (name, fun)));
}

template<typename D, typename S1, typename S2> inline void
define (const generic& name, D (*fun) (const S1&, const S2&)) {
  current_ev->overload (name, as<generic> (binary_routine (name, fun)));
}

template<typename D, typename S1, typename S2, typename S3>
inline void
define (const generic& name, D (*fun) (const S1&, const S2&, const S3&)) {
  current_ev->overload (name, as<generic> (ternary_routine (name, fun)));
}

template<typename D, typename S1> inline void
define_converter (const generic& name, D (*f) (const S1&), nat p) {
  generic con= gen (GEN_INTO, type_name<S1> (), type_name<D> ());
  routine fun= unary_routine (con, f);
  current_ev->overload (name, as<generic> (fun), p);
}

inline void
define_primitive (const generic& name, generic (*fun) (const generic&)) {
  current_ev->set (name, as<generic> (primitive (name, fun)));
}

/******************************************************************************
* Gluing new data types
******************************************************************************/

vector<generic> all_type_names ();
void define_type_sub (const generic& name, nat id);
nat type_id (const generic& name);
generic type_name (nat id);

template<typename C> inline nat type_id () {
  return type_information<C>::id; }
template<typename C> inline generic type_name () {
  return type_name (type_id<C> ()); }

template<typename C>
struct define_type_helper {
  static bool helper_equal (const C& x, const C& y) { return x==y; }
  static bool helper_unequal (const C& x, const C& y) { return x!=y; }
  static syntactic helper_flatten (const C& x) { return flatten (x); }
  static inline void def_type (const generic& name);
  static inline void def_default ();
};

template<typename C> inline void
define_type_helper<C>::def_type (const generic& name) {
  define_type_sub (name, type_id<C> ());
  define_type_sub (gen (GEN_TUPLE_TYPE, name), type_id<tuple<C> > ());
  define_type_sub (gen (GEN_ALIAS_TYPE, name), type_id<alias<C> > ());
  define_type_sub (gen (GEN_GENERIC_ALIAS_TYPE, name),
		   type_id<generic_alias<C> > ());

  define (GEN_ALIAS, new_alias<C>);
  routine specializer= unary_routine (GEN_SPECIALIZE, new_genalias<C>);
  alias_specializer (type_id<C> (), specializer);
  routine getter= unary_routine (GEN_UNALIAS, get_genalias<C>);
  alias_getter (type_id<generic_alias<C> > (), getter);  
  getter= unary_routine (GEN_UNALIAS, get_alias<C>);
  alias_getter (type_id<alias<C> > (), getter);  
  routine setter= binary_routine (GEN_UNALIAS, set_genalias<C>);
  alias_setter (type_id<generic_alias<C> > (), setter);  
  setter= binary_routine (GEN_UNALIAS, set_alias<C>);
  alias_setter (type_id<alias<C> > (), setter);

  define_converter<C,alias<C> >
    (GEN_REWRITE, get_alias<C>, PENALTY_AUTOMATIC);
  define_converter<C,generic_alias<C> >
    (GEN_REWRITE, get_genalias<C>, PENALTY_AUTOMATIC);
  define_converter<alias<generic>,generic_alias<C> >
    (GEN_REWRITE, generalize_genalias<C>, PENALTY_AUTOMATIC);
  define_converter<alias<C>,generic_alias<C> >
    (GEN_REWRITE, incarnate_genalias<C>, PENALTY_AUTOMATIC);
}

template<typename C> inline void
define_type_helper<C>::def_default () {
  define<bool,C,C> (GEN_EQUAL, helper_equal);
  define<bool,C,C> (GEN_UNEQUAL, helper_unequal);
  define<syntactic,C> (GEN_FLATTEN, helper_flatten);
}

template<> inline void
define_type_helper<generic>::def_default () {
}

template<typename C> inline void
define_type (const generic& name) {
  define_type_helper<C>::def_type (name);
  define_type_helper<C>::def_default ();
}

} // namespace mmx
#endif // __GLUE_HPP
