
/******************************************************************************
* MODULE     : literal.hpp
* DESCRIPTION: Atomic literals
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __LITERAL_HPP
#define __LITERAL_HPP
#include <basix/string.hpp>
#include <basix/symbol.hpp>

//! \file

namespace mmx {

/******************************************************************************
* Literals
******************************************************************************/

class literal {
MMX_ALLOCATORS
  string s;
  symbol<string> sym;
public:
  inline string operator * () const { return s; }
  inline friend string as_string (const literal& l) { return l.s; }
  inline friend symbol<string> as_symbol (const literal& s) { return s.sym; }
  inline literal (): s (""), sym (s) {}
  inline literal (const char *s2): s (string (s2)), sym (s) {}
  inline literal (const string& s2): s (s2), sym (s) {}
  inline literal (const literal& l): s (l.s), sym (l.sym) {}
  inline literal (const symbol<string>& sym2): s (*sym2), sym (sym2) {}
};

inline nat hash (const literal& c) { return hash (as_symbol (c)); }
inline nat soft_hash (const literal& c) { return soft_hash (as_symbol (c)); }
inline nat hard_hash (const literal& c) { return hard_hash (as_string (c)); }
inline bool operator == (const literal& c1, const literal& c2) {
  return as_symbol (c1) == as_symbol (c2); }
inline bool operator != (const literal& c1, const literal& c2) {
  return as_symbol (c1) != as_symbol (c2); }
inline bool eq (const literal& c1, const literal& c2) {
  return eq (as_symbol (c1), as_symbol (c2)); }
inline bool neq (const literal& c1, const literal& c2) {
  return neq (as_symbol (c1), as_symbol (c2)); }
inline bool hard_eq (const literal& c1, const literal& c2) {
  return hard_eq (as_string (c1), as_string (c2)); }
inline bool hard_neq (const literal& c1, const literal& c2) {
  return hard_neq (as_string (c1), as_string (c2)); }

string literal_to_string (const generic& g);
syntactic flatten (const literal& s);
PRINT_SUGAR(,literal)

template<typename C> C
make_literal (const literal& lit) {
  return C (as_string (lit));
}

} // namespace mmx
#endif // __LITERAL_HPP
