
/******************************************************************************
* MODULE     : pair.hpp
* DESCRIPTION: Pairs
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_PAIR_HPP
#define __MMX_PAIR_HPP
#include <basix/syntactic.hpp>
#include <basix/operators.hpp>

//! \file

namespace mmx {
#define TMPL template<typename C1, typename C2>
#define Pair pair<C1,C2>

TMPL
class pair {
MMX_ALLOCATORS;
public:
  C1 x1;
  C2 x2;
  inline pair () {}
  inline pair (const C1& x1b, const C2& x2b): x1 (x1b), x2 (x2b) {}
};

template<typename Op, typename C1, typename C2> nat
unary_hash (const Pair& p) {
  nat h= Op::op (p.x1);
  return (h<<1) ^ (h<<3) ^ (h<<9) ^ (h>>29) ^ Op::op (p.x2);
}

template<typename Op, typename C1, typename C2> bool
binary_test (const Pair& p1, const Pair& p2) {
  return Op::op (p1.x1, p2.x1) && Op::op (p1.x2, p2.x2);
}

TMPL inline syntactic
flatten (const Pair& p) {
  return apply (GEN_SQTUPLE, flatten (p.x1), flatten (p.x2));
}

TRUE_IDENTITY_OP_SUGAR(TMPL,Pair)
SOFT_IDENTITY_OP_SUGAR(TMPL,Pair)
HARD_IDENTITY_OP_SUGAR(TMPL,Pair)
PRINT_SUGAR(TMPL,Pair)

TMPL inline Pair
operator * (const Pair& p1, const Pair& p2) {
  return Pair (p1.x1 * p2.x1, p1.x2 * p2.x2);
}

#undef TMPL
#undef Pair
} // namespace mmx
#endif // __MMX_PAIR_HPP
