
/******************************************************************************
* MODULE     : primitive.hpp
* DESCRIPTION: Abstract primitives for evaluators
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __PRIMITIVE_HPP
#define __PRIMITIVE_HPP
#include <basix/syntactic.hpp>
#include <basix/identifiers.hpp>
#include <basix/wrap.hpp>

//! \file

namespace mmx {

class routine;
class primitive;
class primitive_rep: public rep_struct {
public:
  generic name;
public:
  inline primitive_rep (const generic& name2): name (name2) {}
  virtual inline ~primitive_rep () {}
  inline virtual generic apply (const generic& x) const { return x; }
};

class primitive {
INDIRECT_PROTO (primitive, primitive_rep)
public:
  inline const primitive_rep* operator * () const { return rep; }
  inline primitive (): rep (NULL) {}
  primitive (const generic& name, generic (*prim) (const generic&));
  primitive (const routine& fun);
  friend inline bool is_nil (const primitive& fun);
};
INDIRECT_NULL_IMPL (primitive, primitive_rep)

inline bool is_nil (const primitive& fun) {
  return fun.rep == NULL; }
inline syntactic flatten (const primitive& fun) {
  if (is_nil (fun)) return as_syntactic (GEN_NIL);
  else return as_syntactic (fun->name); }

WRAP_INDIRECT_IMPL(inline,primitive)
PRINT_SUGAR(,primitive);

} // namespace mmx
#endif // __PRIMITIVE_HPP
