
/******************************************************************************
* MODULE     : symbol.hpp
* DESCRIPTION: A symbol points to another object but is stored only once
*              unless explicitely specified
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_SYMBOL_HPP
#define __MMX_SYMBOL_HPP
#include <basix/table.hpp>

//! \file

namespace mmx {
#define TMPL_DEF template<typename C, typename V= eq_table>
#define TMPL template<typename C, typename V>
#define Symbol symbol<C,V>
#define Symbol_rep symbol_rep<C,V>
TMPL class symbol_rep;
TMPL class symbol;
TMPL inline nat hash (const Symbol& s);
TMPL inline bool operator == (const Symbol& s1, const Symbol& s2);
TMPL inline bool operator != (const Symbol& s1, const Symbol& s2);
TMPL inline nat soft_hash (const Symbol& s);
TMPL inline bool eq (const Symbol& s1, const Symbol& s2);
TMPL inline bool neq (const Symbol& s1, const Symbol& s2);

/******************************************************************************
* Symbol class and its representation class
******************************************************************************/

TMPL_DEF
class symbol_rep REP_STRUCT {
private:
  typedef table<Symbol_rep*,C,V> table_type;
  static inline table_type ptr () {
    static table_type t ((Symbol_rep*) NULL);
    return t; };
  static inline table_type all_entries () {
    return copy (ptr ()); }
  static inline Symbol_rep* get_entry (const C& ref) {
    return (inside (ptr ())) -> get (ref); }
  static inline void set_entry (const C& ref, Symbol_rep* val) {
    (inside (ptr ())) -> set (ref)= val; }
  static inline void reset_entry (const C& ref) {
    (inside (ptr ())) -> reset (ref); }

private:
  C ref;

public:
  inline Symbol_rep (const C& ref2): ref (ref2) {
    set_entry (ref, this); }
  inline ~Symbol_rep () {
    reset_entry (ref); }
  friend class Symbol;
  static void show_symbol_entries () {
    cout << "symbols\t= " << all_entries () << "\n"; }
};

TMPL_DEF
class symbol {
INDIRECT_PROTO_2 (symbol, symbol_rep, C, V)
public:
  symbol (const C& ref= C()) {
    rep= Symbol_rep::get_entry (ref);
    if (rep == (Symbol_rep*) NULL) rep= new Symbol_rep (ref);
    else INC_COUNT (rep); }
  inline const C& operator * () const { return rep->ref; };
  friend nat hash LESSGTR (const Symbol& s);
  friend bool operator == LESSGTR (const Symbol& s1, const Symbol& s2);
  friend bool operator != LESSGTR (const Symbol& s1, const Symbol& s2);
  friend nat soft_hash LESSGTR (const Symbol& s);
  friend bool eq LESSGTR (const Symbol& s1, const Symbol& s2);
  friend bool neq LESSGTR (const Symbol& s1, const Symbol& s2);
};
INDIRECT_IMPL_2 (symbol, symbol_rep, typename C, C, typename V, V)

template<> struct symbolic_type_information<symbol<string> > {
  static const nat id= SYMBOLIC_LITERAL; };
template<typename C> struct symbolic_type_information<symbol<vector<C> > > {
  static const nat id= SYMBOLIC_COMPOUND; };

/******************************************************************************
* Routines related to equality testing
******************************************************************************/

TMPL inline nat hash (const Symbol& s) {
  return hash (*s); }
TMPL inline bool operator == (const Symbol& s1, const Symbol& s2) {
  return (*s1) == (*s2); }
TMPL inline bool operator != (const Symbol& s1, const Symbol& s2) {
  return (*s1) != (*s2); }

TMPL inline nat soft_hash (const Symbol& s) {
  return as_hash (s.rep); }
TMPL inline bool eq (const Symbol& s1, const Symbol& s2) {
  return s1.rep == s2.rep; }
TMPL inline bool neq (const Symbol& s1, const Symbol& s2) {
  return s1.rep != s2.rep; }

template<typename C> inline nat hash (const symbol<C,equal_table>& s) {
  return soft_hash (s); }
template<typename C> inline bool
operator == (const symbol<C,equal_table>& s1,
	     const symbol<C,equal_table>& s2) {
  return eq (s1, s2); }
template<typename C> inline bool
operator != (const symbol<C,equal_table>& s1,
	     const symbol<C,equal_table>& s2) {
  return neq (s1, s2); }

/******************************************************************************
* Printing symbols
******************************************************************************/

TMPL void show_symbol_entries () {
  Symbol_rep::show_symbol_entries (); }

TMPL inline syntactic flatten (const Symbol& s) {
  return apply ("symbol", flatten (*s)); }

PRINT_SUGAR(TMPL,Symbol)

#undef TMPL_DEF
#undef TMPL
#undef Symbol
#undef Symbol_rep
} // namespace mmx
#endif // __MMX_SYMBOL_HPP
