
/******************************************************************************
* MODULE     : system.hpp
* DESCRIPTION: Operating system related routines
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_SYSTEM_HPP
#define __MMX_SYSTEM_HPP
#include <basix/string.hpp>
#include <basix/vector.hpp>

//! \file

namespace mmx {

string get_env (const string& var);
int set_env (const string& var, const string& value);
int system (const string& s);
void init_system ();
string user_dir ();
string prefix_dir ();
string sysconf_dir ();
string load_path ();
string eval_system (const string& s);

bool   file_test (const string& file_name, const string& attrs);
bool   file_exists (const string& file_name);
bool   file_is_file (const string& file_name);
bool   file_is_directory (const string& file_name);
string file_in_path (const string& file_path, const string& file_name);
bool   file_is_script (const string& file_name);
string get_dir_name (const string& name);
string get_file_name (const string& name);
string relative_name (const string& base, const string& name);
string resolve_name (const string& base, const string& name);

bool load (const string& file_name, string& s);
bool load (const string& fp, const string& file_name, string& s);
bool save (const string& file_name, const string& s);
bool load_directory (const string& name, vector<string>& s);
bool mkdir (const string& orig_name);

extern void (*embedded_link) (const string& name);
bool dl_exists (const string& name);
void dl_link (const string& name);
void include (const string& name);

} // namespace mmx
#endif // __MMX_SYSTEM_HPP
