
/******************************************************************************
* MODULE     : timer.hpp
* DESCRIPTION: Time related routines
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_TIMER_HPP
#define __MMX_TIMER_HPP
#include <basix/string.hpp>

//! \file

namespace mmx {

//! Relative time of day
nat mmx_time ();
inline string mmx_time_unit () { return("ms"); };

//! User time of the current process and its terminated children
nat mmx_user_time();
inline string mmx_user_time_unit () { return("ms"); };

//! Number of cycles of the cpu.
#if (defined(__GNUC__) && defined(__i386__) && BASIX_HAVE_STDINT_H)
#include<stdint.h>
extern __inline__ nat mmx_cpu_time () {
  uint64_t x;
  __asm__ volatile ("rdtsc" : "=A" (x));
  return (nat) x;
}
#elif (defined(__GNUC__) && defined(__amd64__) && BASIX_HAVE_STDINT_H)
#include<stdint.h>
extern __inline__ nat mmx_cpu_time () {
  uint64_t a, d;
  __asm__ volatile ("rdtsc" : "=a" (a), "=d" (d));
  return (nat) ((d<<32) | a);
}
#else
#include <time.h>
inline nat mmx_cpu_time () {
  return (nat) clock ();
}
#endif
inline string mmx_cpu_time_unit () { return("tk"); };

} // namespace mmx
#endif // __MMX_TIMER_HPP
