
/******************************************************************************
* MODULE     : vector_sort.hpp
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__VECTOR_SORT__HPP
#define __MMX__VECTOR_SORT__HPP
#include <basix/vector.hpp>
#include <basix/operators.hpp>
namespace mmx {

template<typename Op, typename T> static void
sort_sub (vector<T>& a, int start, int end, vector<T>& merge_buf) {
  if (end-start <= 1) return;
  if (end-start == 2) {
    if (!Op::op (a[start], a[start+1])) {
      merge_buf[start]= a[start];
      a[start]= a[start+1];
      a[start+1]= merge_buf[start];
    }
    return;
  }
  int middle= (start+end)>>1; 
  sort_sub<Op> (a, start, middle, merge_buf);
  sort_sub<Op> (a, middle, end, merge_buf);
  int i,j,k;
  for (i=start, j=middle, k=start; (i<middle) && (j<end); )
    if (Op::op (a[i], a[j])) merge_buf[k++]= a[i++];
    else                     merge_buf[k++]= a[j++];
  j=k;
  while (i!=middle) a[k++]= a[i++];
  for (i=start; i<j; i++) a[i]= merge_buf[i];
}

template<typename Op, typename T> void
sort_leq (vector<T>& v) {
  vector<T> merge_buf= fill<T> (N(v)); 
  sort_sub<Op> (v, 0, N(v), merge_buf);
}

template<typename T> inline void
sort (vector<T>& v) {
  sort_leq<less_op> (v);
}

} // namespace mmx
#endif // __MMX__VECTOR_SORT__HPP
