
/******************************************************************************
* MODULE     : mould_symbolic.hpp
* DESCRIPTION: Symbolic moulds
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_MOULD_SYMBOLIC_HPP
#define __MMX_MOULD_SYMBOLIC_HPP
#include <jorix/mould.hpp>
namespace mmx {
#define TMPL template<typename C, typename X, typename V>
#define Mould mould<C,X,V>
#define Mould_rep mould_rep<C,X,V>
#define M list<X>

/******************************************************************************
* Moulds with symbolic entries
******************************************************************************/

TMPL
class symbolic_mould_rep: public mould_rep<C,X,V> {
  generic name;
public:
  inline symbolic_mould_rep (const generic& name2): name (name2) {}
  generic expression () const { return name; }
  C compute (const M& x) {
    M counter= x;
    generic args= comma ();
    while (!is_nil (counter)) {
      args= comma (args, flatten (car (counter)));
      counter= cdr (counter);
    }
    return xaccess (name, args);
  }
  table<bool,X> compute_deps () {
    return table<bool,X> (true); }
};

TMPL Mould
symbolic_mould (const generic& name) {
  return (Mould_rep*) new symbolic_mould_rep<C,X,V> (name);
}

/******************************************************************************
* Helper routines (FIXME: do this in a more elegant way)
******************************************************************************/

#define Symbolic symbolic<generic>

static list<generic>
as_generic_list (const list<Symbolic >& l) {
  if (is_nil (l)) return list<generic> ();
  return cons (*car (l), as_generic_list (cdr (l)));
}

static list<Symbolic >
as_symbolic_list (const list<generic>& l) {
  if (is_nil (l)) return list<Symbolic > ();
  return cons (Symbolic (car (l)), as_symbolic_list (cdr (l)));
}

static list<Symbolic >
solve (const list<Symbolic >& eqs, const list<Symbolic >& vars) {
  // cout << "solve " << eqs << " in " << vars << "\n";
  generic geqs = as<generic> (as_generic_list (eqs));
  generic gvars= as<generic> (as_generic_list (vars));
  vector<Symbolic > gsols= solve (vector<Symbolic > (iterate (geqs)),
				  vector<Symbolic > (iterate (gvars)));
  return as_symbolic_list (list<Symbolic > (iterate (gsols)));
}

#undef Symbolic

/******************************************************************************
* Solving mould equations
******************************************************************************/

TMPL list<C>
xcess (const Mould& m, const list<M >& l) {
  if (is_nil (l)) return list<C> ();
  else return cons (m [car (l)], xcess (m, cdr (l)));
}

template<typename C, typename X, typename Y, typename V>
class solve_mould_rep: public mould_rep<C,X,V> {
  mould<C,Y,V> eq;
  Mould        var;
  nat          p;
public:
  inline solve_mould_rep (const mould<C,Y,V>& eq2, const Mould& var2, nat p2):
    eq (eq2), var (var2), p (p2) {}
  generic expression () const { return solve (flatten (eq), flatten (var)); }
  C compute (const M& x) {
    list<list<Y> > eqs_xl =
      up_till_weight<Y> (dependencies (eq), weight (x) + p);
    list<list<X> > vars_xl=
      up_till_weight<X> (dependencies (var), weight (x));
    list<C> eqs = xcess (eq, eqs_xl);
    list<C> vars= xcess (var, vars_xl);
    list<C> sols= solve (eqs, vars);
    for (; !is_nil (vars_xl); vars_xl= cdr (vars_xl), sols= cdr (sols))
      this->t (car (vars_xl))= car (sols);
    return this->t[x];
  }
  table<bool,X> compute_deps () {
    return table<bool,X> (true); }
};

template<typename C, typename X, typename Y, typename V> Mould
solve_mould (const mould<C,Y,V>& eq, const Mould& var, nat look_ahead= 0) {
  ASSERT (can_list (dependencies (var)) && can_list (dependencies (eq)),
	  "moulds should have a finite number of 'generators'");
  return (Mould_rep*) new solve_mould_rep<C,X,Y,V> (eq, var, look_ahead);
}

TMPL mould<C,nat,V>
decompose (const Mould& what, const Mould& over) {
  mould<C,nat,V> m= symbolic_mould<C,nat,V> (generic ("m"));
  return solve_mould (compose (m, over) - what, m);
}

#undef TMPL
#undef Mould
#undef Mould_rep
#undef M
} // namespace mmx
#endif // __MMX_MOULD_SYMBOLIC_HPP
