
/******************************************************************************
* MODULE     : torn_solve.hpp.hpp
* DESCRIPTION: Polynomial root finding based on torn polynomials
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_TORN_SOLVE_HPP
#define __MMX_TORN_SOLVE_HPP
#include <jorix/torn_evaluate.hpp>
namespace mmx {
template<typename C> struct standard_torn;
#define TMPL_DEF template<typename C, typename V= standard_torn<C> >
#define TMPL template<typename C, typename V>
#define Torn torn<C,V>
#define Torn_vector vector<Torn >
#define Torn_polynomial torn_polynomial<C,V>
#define Torn_polynomial_rep torn_polynomial_rep<C,V>
#define W V::PV
#define Polynomial typename V::PT
#define Exp typename V::ET
TMPL class torn_polynomial_rep;
TMPL class torn_polynomial;
TMPL Torn_vector compute_roots (const Torn_polynomial& tp, nat credit);

/******************************************************************************
* Root finding
******************************************************************************/

TMPL void
quality_roots (const Torn_polynomial& tp, const Torn_vector& tv) {
  Torn_polynomial dp= derive (tp);
  Torn_polynomial ddp= derive (dp);
  Torn_polynomial dddp= derive (dp);
  Torn_vector     pv= eval (tp, tv);
  Torn_vector     dv= eval (dp, tv);
  Torn_vector     ddv= eval (ddp, tv);
  Torn_vector     dddv= eval (dddp, tv);
  for (nat i=0; i<N(tv); i++)
    if (magnitude (pv[i].x) > -100) {
      Torn ref= Torn (1, -pv[i].e);
      Torn v0 = Torn (1, 0) * pv[i];
      Torn v1 = tv[i] * dv[i];
      Torn v2 = tv[i] * tv[i] * ddv[i] / Torn (2, 0);
      Torn v3 = tv[i] * tv[i] * tv[i] * dddv[i] / Torn (6, 0);
      cout << i << ": "
	   << (v0 * ref).e << ", "
	   << (v1 * ref).e << ", "
	   << (v2 * ref).e << ", "
	   << (v3 * ref).e << "\n";
      //cout << i << ": "
      //<< magnitude (pv[i].x) << ", "
      //<< magnitude (dv[i].x) << "\t"
      //<< tv[i] << ", " << pv[i] << ", " << dv[i] << "\n";
    }
}

TMPL Torn_vector
linear_roots (const Torn_polynomial& tp) {
  ASSERT (N(tp->p) == 2, "polynomial should have degree one");
  return vec<Torn > (Torn (-tp->p[0] / tp->p[1], tp->ex[0] - tp->ex[1]));
}

TMPL Torn_vector
quadratic_roots (const Torn_polynomial& tp) {
  ASSERT (N(tp->p) == 3, "polynomial should have degree two");
  ASSERT (N(tp->ab) == 2, "single slope expected");
  C a= tp->p[2], b= tp->p[1], c= tp->p[0];
  C d= sqrt (square (b) - 4*a*c);
  C x1= -b + d, x2= -b - d, x= (abs (x1) > abs (x2)? x1: x2);
  C z1= x / (2 * a), z2= (2 * c) / x;
  Torn v1 (z1, tp->ex[0] - tp->ex[1]);
  Torn v2 (z2, tp->ex[0] - tp->ex[1]);
  //cout << "===========================================\n";
  //cout << a << " z^2 + " << b << " z + " << c << "\n";
  //cout << "    " << z1 << " -> " << c + z1 * (b + z1 * a) << "\n";
  //cout << "    " << z2 << " -> " << c + z2 * (b + z2 * a) << "\n";
  if (v1.e < v2.e || (v1.e == v2.e && abs (z1) < abs (z2))) {
    //quality_roots (tp, vec<Torn > (v1, v2));
    return vec<Torn > (v1, v2);
  }
  else {
    //quality_roots (tp, vec<Torn > (v2, v1));
    return vec<Torn > (v2, v1);
  }
}

TMPL Torn_vector
subdivide_roots (const Torn_polynomial& tp, nat credit) {

  //vector<nat> ab= vec<nat> (tp->ab[0]);
  //cout << "tp->ab= " << tp->ab << "\n";

  Torn_vector tv;
  nat start= 0;
  while (start+1 < N(tp->ab)) {
    nat end= start+1;
    while (end+1 < N(tp->ab)) {
      nat i1 = tp->ab[end-1];
      nat i2 = tp->ab[end];
      nat i3 = tp->ab[end+1];
      Exp sl1= (tp->ex[i2] - tp->ex[i1]) / (i2 - i1);
      Exp sl2= (tp->ex[i3] - tp->ex[i2]) / (i3 - i2);
      if (sl1 - sl2 >= V::prec () + V::disc ()) break;
      end++;
    }
    tv << compute_roots (range (tp, tp->ab[start], tp->ab[end] + 1), credit);

    //cout << start << " -- " << end << ", "
    //<< tp->ab[start] << " -- " << tp->ab[end] << "\n";
    start= end;

    //ab= append (ab, vec<nat> (tp->ab[start]));

  }

  /*
  if (N(ab) > 2) {
    cout << "====================================================\n";
    //cout << "Abscissae " << N(tp->p) << ", " << tp->ab << ", " << ab << "\n";
    //cout << "tp= " << tp << "\n";
    for (nat i=0; i+1<N(ab); i++) {
      //cout << ab[i] << " -- " << ab[i+1] << "\n";
      //cout << "tp= " << range (tp, ab[i], ab[i+1] + 1) << "\n";
      //cout << "tv= " << range (tv, ab[i], ab[i+1]) << "\n";
      quality_roots (range (tp, ab[i], ab[i+1] + 1),
		     range (tv, ab[i], ab[i+1]));
      
    }
    cout << "===\n";
    //cout << "ex= " << tp->ex << "\n";
    quality_roots (tp, tv);
  }
  */

  return tv;
}

TMPL Torn_vector
improve_roots (const Torn_polynomial& tp, const Torn_vector& tv) {
  Torn_polynomial dp= derive (tp);
  vector<nat>     ind;
  Torn_vector     pv= eval (tp, tv, ind);
  Torn_vector     dv= eval (dp, tv);
  Torn_vector     rv= fill<Torn > (N(tv));
  for (nat i=0; i<N(tv); i++)
    rv[i]= tv[i] - pv[i] / (dv[i] + (Torn (ind[i], 0) * pv[i]) / tv[i]);
  return rv;
}

TMPL Torn_vector
graeffe_roots (const Torn_polynomial& tp, nat credit) {
  Torn_polynomial tp0= extract_mod (tp, 0, 2);
  Torn_polynomial tp1= extract_mod (tp, 1, 2);
  Torn_polynomial gp = graeffe (tp);
  Torn_vector gv = subdivide_roots (gp, credit);
  Torn_vector tv = fill<Torn > (N(gv));
  Torn_vector tv0= eval (tp0, gv);
  Torn_vector tv1= eval (tp1, gv);
  for (nat i=0; i<N(gv); i++) {
    Torn z = sqrt (gv[i]);
    Torn v0= tv0[i] + z * tv1[i];
    Torn v1= tv0[i] - z * tv1[i];
    if (v0.e < v1.e || (v0.e == v1.e && abs (v0.x) < abs (v1.x)))
      tv[i]= z;
    else
      tv[i]= -z;
  }
  //cout << "gv= " << gv << "\n";
  //cout << "tv= " << tv << "\n";
  tv= improve_roots (tp, tv);
  //cout << "-------------------------------------------\n";
  //quality_roots (gp, gv);
  //cout << "---\n";
  //quality_roots (tp, tv);
  return tv;
}

TMPL Torn_vector
compute_roots (const Torn_polynomial& tp, nat credit) {
  //cout << "Roots " << credit << ", " << tp << "\n";
  if (N (tp->p) == 2) return linear_roots (tp);
  if (N (tp->p) == 3 && N (tp->ab) == 2) return quadratic_roots (tp);
  ASSERT (credit > 0, "multiple roots not yet treated");
  return graeffe_roots (tp, credit - 1);
}

TMPL Torn_vector
compute_roots (const Torn_polynomial& tp) {
  double credit=
    (3.0 * log (1.0 + N(tp->p)) + 2.0 * log (V::prec ())) / log (2.0);
  Torn_vector tv= compute_roots (tp, (nat) credit);
  quality_roots (tp, tv);
  return tv;
}

#undef TMPL_DEF
#undef TMPL
#undef Torn
#undef Torn_vector
#undef Torn_polynomial
#undef Torn_polynomial_rep
#undef W
#undef Polynomial
#undef Exp
} // namespace mmx
#endif // __MMX_TORN_SOLVE_HPP
