
/******************************************************************************
* MODULE     : base_evaluator.hpp
* DESCRIPTION: The base evaluator
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __BASE_EVALUATOR_HPP
#define __BASE_EVALUATOR_HPP
#include <basix/evaluator.hpp>
#include <mmxlight/environment.hpp>

//! \file

namespace mmx {

class base_evaluator_rep: public evaluator_rep {
protected:
  environment env;

public:
  virtual void* get_internal_data () const;
  //virtual environment get_environment () const;
  //virtual void set_environment_type (nat tp) const;
  //virtual nat  get_environment_type () const;
  virtual void set (const generic& var, const generic& val) const;
  virtual void reset (const generic& var) const;
  virtual void overload (const generic& var, const generic& val, nat p) const;
  virtual bool contains (const generic& var) const;
  virtual generic get (const generic& var) const;
  virtual bool get (const generic& var, generic& val) const;

  virtual generic eval  (const generic& x) const;
  virtual vector<generic> eval_cdr (const vector<generic>& x) const;
  virtual generic construct (const generic& x) const;
  virtual generic apply (const generic& fun) const;
  virtual generic apply (const generic& fun, const generic& x1) const;
  virtual generic apply (const generic& fun,
			 const generic& x1, const generic& x2) const;
  virtual generic apply (const generic& fun, const vector<generic>& v) const;

public:
  inline base_evaluator_rep ();
  inline base_evaluator_rep (const evaluator& ev);
};

evaluator base_evaluator ();
evaluator base_evaluator (const evaluator& ev);

environment get_environment (const evaluator& ev);
void        set_environment_type (const evaluator& ev, nat tp);
nat         get_environment_type (const evaluator& ev);

inline base_evaluator_rep::base_evaluator_rep ():
  env (global_environment ()) {}
inline base_evaluator_rep::base_evaluator_rep (const evaluator& ev):
  env (local_environment (get_environment (ev))) {}

} // namespace mmx
#endif // __BASE_EVALUATOR_HPP
