
/******************************************************************************
* MODULE     : block_polynomial_integer.hpp
* DESCRIPTION: Kronecker multiplication of multivariate integer polynomials
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_BLOCK_POLYNOMIAL_INTEGER_HPP
#define __MMX_BLOCK_POLYNOMIAL_INTEGER_HPP
#include <algebramix/polynomial_integer.hpp>
#include <multimix/block_polynomial.hpp>
namespace mmx {
struct kronecker_block_polynomial: vector_naive {};

void kronecker_encode (integer& dest, const integer* src,
		       const vector<nat>& ords, const vector<nat>& vbits);
void kronecker_decode (integer* dest, const vector<nat>& ords,
		       const vector<nat>& vbits, const integer& src);
void kronecker_mul (integer* dest,
		    const integer* src1, const vector<nat>& ords1,
		    const integer* src2, const vector<nat>& ords2);
void kronecker_square (integer* dest,
		       const integer* src1, const vector<nat>& ords1);

#define TMPL template<typename C>
#define Polynomial block_polynomial<C,kronecker_block_polynomial>

TMPL Polynomial
operator * (const Polynomial& p1, const Polynomial& p2) {
  nat d= dim (p1);
  ASSERT (dim (p2) == d, "dimensions don't match");
  if (N (p1) == 0 || N (p2) == 0) return Polynomial (vec<nat> (d));
  vector<nat> ord1= ord (p1);
  vector<nat> ord2= ord (p2);
  vector<nat> ordr= ord1 + ord2 - 1;
  nat n= big_mul (ordr);
  C* r= mmx_new<C> (n, C(0));
  kronecker_mul (r, seg (p1), ord1, seg (p2), ord2);
  return Polynomial (r, ordr);
}

#undef TMPL
#undef Polynomial

} // namespace mmx
#endif // __MMX_BLOCK_POLYNOMIAL_INTEGER_HPP
