
/******************************************************************************
* MODULE     : lex_series.hpp
* DESCRIPTION: Multivariate power series using a lexicographic representation
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_LEX_SERIES_HPP
#define __MMX_LEX_SERIES_HPP
#include <basix/table.hpp>
#include <multimix/block_polynomial_rational.hpp>
#include <multimix/block_polynomial_floating.hpp>
namespace mmx {
struct lex_series_naive;
#define Lex_series_variant(C) \
  lex_series_variant_helper<C>::SV
#define TMPL_DEF \
  template<typename C, typename V=typename Lex_series_variant(C)>
#define TMPL template<typename C, typename V>
#define Series lex_series<C,V>
#define Series_rep lex_series_rep<C,V>
#define Implicit_series_rep implicit_lex_series_rep<C,V>
#define Variable nat
#define Monomial vector<int>
#define Helper V::template truncated_helper<C>
#define Truncated typename Helper::truncated_type
#define Precision typename Helper::precision_type
TMPL class lex_series;
TMPL syntactic flatten (const Series& f, bool block= false);

/******************************************************************************
* Variant helpers
******************************************************************************/

class below_iterator_rep: public iterator_rep<Monomial > {
  Monomial m;  // current monomial
  nat      i;  // current index
  nat      p;  // product of degrees

public:
  inline below_iterator_rep (const Monomial& m2, nat i2= 0):
    m (m2), i (i2), p (big_mul (m)) {}

protected:
  inline bool is_busy () { return i<p; }
  inline void advance () { i++; }
  inline Monomial current () {
    nat k, q= 1;
    Monomial r= fill<int> (N(m));
    for (k=0; k<N(m); k++) {
      r[k]= (i / q) % m[k];
      q *= m[k];
    }
    return r;
  }
  inline iterator_rep<Monomial >* clone () {
    return new below_iterator_rep (m, i); }
};

inline iterator<Monomial >
below (const Monomial& m) {
  return iterator<Monomial > (new below_iterator_rep (m));
}

inline iterator<Monomial >
before (const Monomial& m) {
  return iterator<Monomial > (new below_iterator_rep (m + Monomial (1)));
}

inline bool
before (const Monomial& m1, const Monomial& m2) {
  return m1 <= m2;
}

  /*
struct lex_series_naive: public vector_naive {
  typedef nat variable_type;
  template<typename C>
  struct truncated_helper {
    typedef table<C,Monomial > truncated_type;
    typedef nat precision_type;
    static inline truncated_type
    zero (nat d, const precision_type& p) {
      (void) d; (void) p;
      return truncated_type (0); }
    static inline bool
    contains (const truncated_type& t, const Monomial& m) {
      return t->contains (m); }
    static inline C
    get (const truncated_type& t, const Monomial& m) {
      return t[m]; }
    static inline void
    set (truncated_type& t, const Monomial& m, const C& c) {
      t[m]= c; }
    static inline precision_type
    shift (const precision_type& p, int sh, nat v) {
      (void) v; return max (0, ((int) p) + sh); }
    static inline precision_type
    increase (const precision_type& p) {
      return p << 1; }
    static inline precision_type
    maximum (const precision_type& p1, const precision_type& p2) {
      return max (p1, p2); }
    static inline precision_type
    total_degree (nat deg) {
      return deg; }
    static inline precision_type
    as_precision (const Monomial& m) {
      if (N(m) == 0) return 1;
      else return big_max (m) + 1; }
  };

  static inline Monomial multiply (const Monomial& m1, const Monomial& m2) {
    return m1+m2; }
  static inline Monomial divide (const Monomial& m1, const Monomial& m2) {
    return m1-m2; }
  static inline Monomial monomial () {
    return Monomial (); }
  static inline Monomial monomial (nat d) {
    return fill<int> (0, d); }
  static inline Monomial monomial (nat i, nat d, nat p=1) {
    Monomial v= fill<int> (0, d); v[i]= p; return v; }
};
*/

struct lex_series_block: public vector_naive {
  typedef nat variable_type;
  template<typename C>
  struct truncated_helper {
    struct truncated_type {
      block_polynomial<C>    p;
      block_polynomial<bool> done;
      inline truncated_type (const vector<nat>& v): p (v), done (v) {}
    };
    typedef nat precision_type;
    static inline truncated_type
    zero (nat d, const precision_type& p) {
      vector<nat> orders= fill<nat> (p, d);
      return truncated_type (orders); }
    static inline bool
    contains (const truncated_type& t, const Monomial& m) {
      ASSERT (m <= Monomial (ord (t.p) - 1), "out of range");
      return t.done[m]; }
    static inline C
    get (const truncated_type& t, const Monomial& m) {
      ASSERT (m <= Monomial (ord (t.p) - 1), "out of range");
      return t.p[m]; }
    static inline void
    set (truncated_type& t, const Monomial& m, const C& c) {
      ASSERT (m <= Monomial (ord (t.p) - 1), "out of range");
      t.p[m]= c; t.done[m]= true; }
    static inline void
    add (truncated_type& t, const Monomial& m, const C& c) {
      if (m <= Monomial (ord (t.p) - 1))
	t.p[m] += c; }
    static inline void
    commit (truncated_type& t, const Monomial& m) {
      ASSERT (m <= Monomial (ord (t.p) - 1), "out of range");
      t.done[m]= true; }
    static inline precision_type
    shift (const precision_type& p, int sh, nat v) {
      (void) v; return max (0, ((int) p) + sh); }
    static inline precision_type
    increase (const precision_type& p) {
      return p << 1; }
    static inline precision_type
    maximum (const precision_type& p1, const precision_type& p2) {
      return max (p1, p2); }
    static inline precision_type
    total_degree (nat deg) {
      return deg; }
    static inline precision_type
    as_precision (const Monomial& m) {
      if (N(m) == 0) return 1;
      else return big_max (m) + 1; }
  };

  static inline Monomial multiply (const Monomial& m1, const Monomial& m2) {
    return m1+m2; }
  static inline Monomial divide (const Monomial& m1, const Monomial& m2) {
    return m1-m2; }
  static inline Monomial monomial () {
    return Monomial (); }
  static inline Monomial monomial (nat d) {
    return fill<int> (0, d); }
  static inline Monomial monomial (nat i, nat d, nat p=1) {
    Monomial v= fill<int> (0, d); v[i]= p; return v; }
};

template<typename C>
struct lex_series_variant_helper {
  //typedef lex_series_naive SV;
  typedef lex_series_block SV;
};

/******************************************************************************
* Global variables
******************************************************************************/

extern nat mmx_lex_series_precision;     // current series precision
extern nat mmx_lex_series_cancel_order;  // order for zero testing

/******************************************************************************
* Dense multivariate power series
******************************************************************************/

TMPL_DEF
class lex_series_rep REP_STRUCT {
public:   // should be protected
  nat d;
  Precision cur_p;
  Truncated coeffs;

public:
  inline lex_series_rep (nat d2):
    d (d2), cur_p (0), coeffs (Helper::zero (d, 0)) {
      ASSERT (d<1000, "suspicious dimension"); }
  inline virtual ~lex_series_rep () {}
  inline Series me () const;
  virtual syntactic expression () const = 0;
  virtual void set_precision (const Precision& p) {
    Precision np= Helper::maximum (cur_p, p);
    if (np != cur_p) {
      //cout << "{" << d << ": " << cur_p << " -> " << np << "} ";
      cur_p = np;
      coeffs= Helper::zero (d, np); } }
  virtual void increase_precision (const Precision& p) {
    Precision np= Helper::maximum (cur_p, p);
    if (np != cur_p && d != 0)
      set_precision (Helper::maximum (Helper::increase (cur_p), p)); }
  virtual C compute (const Monomial& m) = 0;
  friend class Series;
};

TMPL_DEF
class lex_series {
INDIRECT_PROTO_2 (lex_series, lex_series_rep, C, V)
public:
  lex_series ();
  template<typename T>
  lex_series (const T& c, nat d= 0);
  template<typename T>
  lex_series (const lex_series<T,V>& f);
  template<typename T>
  lex_series (const lex_series<T,V>& f, nat d);
  lex_series (const Monomial& m);
  const Series operator [] (nat i) const;
  const C operator [] (const Monomial& m) const;
};
INDIRECT_IMPL_2 (lex_series, lex_series_rep,
		 typename C, C, typename V, V)

STYPE_TO_TYPE(TMPL,scalar_type,Series,C);

/******************************************************************************
* Basic series routines
******************************************************************************/

TMPL inline Series
Series_rep::me () const {
  return Series (this, true);
}

TMPL inline void
increase_precision (const Series& f, const Precision& p) {
  if (Helper::maximum (f->cur_p, p) != f->cur_p)
    inside (f) -> increase_precision (p);
}

template<typename C,typename V>
class coefficient_lex_series_rep: public Series_rep {
protected:
  Series f;
  nat    i;
public:
  inline coefficient_lex_series_rep (const Series& f2, nat i2):
    Series_rep (f2->d-1), f (f2), i (i2) {}
  syntactic expression () const {
    return apply ("coefficient", flatten (f), flatten (i)); }
  virtual void increase_precision (const Precision& p) { 
    Series_rep::increase_precision (p);
    if (this->d > 1) mmx::increase_precision (f, p); }
  virtual C compute (const Monomial& m) {
    return f[append (m, vec<int> (i))]; }
};

TMPL const Series
Series::operator [] (nat i) const {
  Series f= rep->me ();
  if (f->d == 0) {
    ASSERT (f->d > 0, "non-zero dimension expected"); 
    return i==0? f: Series (0, 0);
  }
  return (Series_rep*)
    new coefficient_lex_series_rep<C,V> (f, i);
}

TMPL const C
Series::operator [] (const Monomial& mm) const {
  Monomial m= mm;
  if (N(m) != rep->d) {
    if (N(m) > rep->d) {
      for (nat i= rep->d; i<N(m); i++)
	if (m[i] != 0) return 0;
      m= range (m, 0, rep->d);
    }
    else m= append (m, fill<int> (0, rep->d - N(m)));
  }
  rep->increase_precision (Helper::as_precision (m));
  if (!Helper::contains (rep->coeffs, m)) {
    C c= rep->compute (m);
    Helper::set (rep->coeffs, m, c);
  }
  return Helper::get (rep->coeffs, m);
}

/******************************************************************************
* Implicit series
******************************************************************************/

TMPL_DEF
class implicit_lex_series_rep: public Series_rep {
public:
  Series   eq;
  Series   init;
  Monomial lim;
public:
  inline implicit_lex_series_rep (nat d):
    Series_rep (d), lim (V::monomial (d)) {}
  virtual Series initialize (Series& init, Monomial& lim) = 0;
  virtual void increase_precision (const Precision& p) {
    Series_rep::increase_precision (p);
    mmx::increase_precision (eq, p); }
  inline C compute (const Monomial& m) {
    if (before (lim, m)) return eq[m];
    else return init[m]; }
};

TMPL
class implicit_container_lex_series_rep: public Series_rep {
  Series f;
public:
  implicit_container_lex_series_rep (const Series& f2):
    Series_rep (f2->d), f (f2) {
      Implicit_series_rep* rep=
        (Implicit_series_rep*) f.operator -> ();
      rep->eq= rep->initialize (rep->init, rep->lim); }
  ~implicit_container_lex_series_rep () {
    Implicit_series_rep* rep=
      (Implicit_series_rep*) f.operator -> ();
    rep->eq= Series (C(0), f->d); }
  syntactic expression () const { return flatten (f); }
  virtual void increase_precision (const Precision& p) {
    Series_rep::increase_precision (p);
    mmx::increase_precision (f, p); }
  C compute (const Monomial& m) {
    return f[m]; }
};

TMPL Series
implicit (const Series& f) {
  return (Series_rep*) new implicit_container_lex_series_rep<C,V> (f);
}

/******************************************************************************
* Hashing and equality testing
******************************************************************************/

HARD_TO_SOFT_IDENTITY_SUGAR(TMPL,Series)

template<typename Op, typename C, typename V> nat
unary_hash (const Series& f, nat n= mmx_lex_series_cancel_order) {
  if (f->d == 0) return Op::op (f[V::monomial ()]);
  else {
    nat h= 1221;
    for (nat i=0; i<n; i++)
      h= (h<<5) + (h>>27) + unary_hash<Op> (f[i], n-i);
    return h;
  }
}

TMPL inline nat
hash (const Series& s) {
  return unary_hash<hash_op> (s);
}

template<typename Op,typename C,typename V> inline bool
binary_test (const Series& f, const Series& g,
	     nat n= mmx_lex_series_cancel_order)
{
  if (f->d < g->d) return binary_test<Op> (Series (f, g->d), g, n);
  else if (g->d < f->d) return binary_test<Op> (f, Series (g, f->d), n);
  else if (f->d == 0) return Op::op (f[V::monomial ()], g[V::monomial ()]);
  else {
    for (nat i=0; i<n; i++)
      if (!binary_test<Op> (f[i], g[i]))
	return false;
    return true;
  }
}

TMPL bool operator == (const Series& f1, const Series& f2) {
  return binary_test<equal_op> (f1, f2); }
TMPL bool operator != (const Series& f1, const Series& f2) {
  return !binary_test<equal_op> (f1, f2); }

/******************************************************************************
* Printing
******************************************************************************/

inline syntactic
flatten_monomial (const Monomial& m) {
  syntactic p= 1;
  for (nat i=0; i<N(m); i++)
    p= p * pow (access (syntactic ("z"), flatten (i+1)), flatten (m[i]));
  return p;
}

TMPL syntactic
flatten_until (const Series& f, const Monomial& m, bool block) {
  if (N(m) == f->d) return flatten (f[m]) * flatten_monomial (m);
  else {
    syntactic s= 0;
    nat prec= mmx_lex_series_precision;
    for (nat i= 0; ((int) i) < ((int) prec) - (block? 0: big_add (m)); i++)
      s= s + flatten_until (f, append (m, vec<int> (i)), block);
    return s;
  }
}

TMPL syntactic
flatten (const Series& f, bool block) {
  nat deg= mmx_lex_series_precision;
  increase_precision (f, Helper::total_degree (deg));
  return flatten_until (f, Monomial (), block) + syntactic ("...");
}

PRINT_SUGAR(TMPL,Series)

/******************************************************************************
* Explicit series
******************************************************************************/

TMPL
class table_lex_series_rep: public Series_rep {
  table<C,Monomial > t;
public:
  table_lex_series_rep (const table<C,Monomial >& t2, nat d):
    Series_rep (d), t (t2) {}
  syntactic expression () const {
    syntactic s;
    for (iterator<Monomial > it= entries (t); busy (it); ++it)
      s= s + flatten (t[*it]) * flatten_monomial (*it);
    return s; }
  C compute (const Monomial& m) { return t[m]; }
};

TMPL
Series::lex_series () {
  rep= new table_lex_series_rep<C,V> (table<C,Monomial > (0), 0);
}

TMPL template<typename T>
Series::lex_series (const T& c, nat d) {
  table<C,Monomial > t (0);
  t[fill<int> (0, d)]= as<C> (c);
  rep= new table_lex_series_rep<C,V> (t, d);
}

TMPL
Series::lex_series (const Monomial& m) {
  table<C,Monomial > t (0);
  t[m]= C (1);
  rep= new table_lex_series_rep<C,V> (t, N(m));
}

template<typename C, typename V, typename K, typename W>
class cast_lex_series_rep: public Series_rep {
  lex_series<K,W> f;
public:
  cast_lex_series_rep (const lex_series<K,W>& f2, nat d):
    Series_rep (d), f (f2) {}
  syntactic expression () const { return flatten (f); }
  void increase_precision (const Precision& p) { 
    Series_rep::increase_precision (p);
    mmx::increase_precision (f, p); }
  C compute (const Monomial& m) { return as<C> (f[m]); }
};

TMPL template<typename T>
Series::lex_series (const lex_series<T,V>& f) {
  rep= new cast_lex_series_rep<C,V,T,V> (f, f->d);
}

TMPL template<typename T>
Series::lex_series (const lex_series<T,V>& f, nat d) {
  ASSERT (f->d <= d, "can only increase dimension");
  rep= new cast_lex_series_rep<C,V,T,V> (f, d);
}

/******************************************************************************
* Abstract operations on series
******************************************************************************/

template<typename Op,typename C,typename V>
class unary_lex_series_rep: public Series_rep {
protected:
  Series f;
public:
  inline unary_lex_series_rep (const Series& f2):
    Series_rep (f2->d), f(f2) {}
  syntactic expression () const {
    return Op::op (flatten (f)); }
  virtual void increase_precision (const Precision& p) { 
    Series_rep::increase_precision (p);
    mmx::increase_precision (f, p); }
  virtual C compute (const Monomial& m) { return Op::op (f[m]); }
};

template<typename Op,typename C,typename V> inline Series
unary_lex_series (const Series& f) {
  return (Series_rep*) new unary_lex_series_rep<Op,C,V> (f);
}

template<typename Op,typename C,typename V>
class unary_implicit_lex_series_rep: public Implicit_series_rep {
protected:
  Series f;
  Series init;
  nat    nr_init;
public:
  unary_implicit_lex_series_rep (const Series& f2):
    Implicit_series_rep (f2->d), f (f2), init (0), nr_init (0) {}
  unary_implicit_lex_series_rep (const Series& f2, const Series& i2):
    Implicit_series_rep (f2->d), f (f2), init (i2), nr_init (1) {}
  syntactic expression () const {
    return Op::op (flatten (f)); }
  virtual void increase_precision (const Precision& p) {
    Implicit_series_rep::increase_precision (p);
    mmx::increase_precision (f, p); }
  Series initialize (Series& init2, Monomial& lim2) {
    if (f->d == 0)
      return Series (Op::op (f[V::monomial ()]), 0);
    else if (nr_init > 0) {
      init2= init;
      lim2 = V::monomial (this->d - 1, this->d, nr_init);
    }
    else if (Op::nr_init () > 0) {
      ASSERT (Op::nr_init () <= 1, "wrong number of initial conditions");
      init2= Op::op (f[0]);
      lim2 = V::monomial (this->d - 1, this->d, 1);
    }
    return Op::def (Series (this->me ()), f);
  }
};

template<typename Op,typename C,typename V> inline Series
unary_implicit_lex_series (const Series& f) {
  Series_rep* rep= new unary_implicit_lex_series_rep<Op,C,V> (f);
  return implicit (Series (rep));
}

template<typename Op,typename C,typename V> inline Series
unary_implicit_lex_series (const Series& f, const C& c) {
  Series_rep* rep= new unary_implicit_lex_series_rep<Op,C,V> (f, c);
  return implicit (Series (rep));
}

template<typename Op,typename C,typename V>
class binary_lex_series_rep: public Series_rep {
protected:
  Series f, g;
public:
  inline binary_lex_series_rep (const Series& f2, const Series& g2):
    Series_rep (max (f2->d, g2->d)), f(f2), g (g2) {}
  syntactic expression () const {
    return Op::op (flatten (f), flatten (g)); }
  virtual void increase_precision (const Precision& p) {
    Series_rep::increase_precision (p);
    mmx::increase_precision (f, p);
    mmx::increase_precision (g, p); }
  virtual C compute (const Monomial& m) {
    return Op::op (f[m], g[m]); }
};

template<typename Op,typename C,typename V> inline Series
binary_lex_series (const Series& f, const Series& g) {
  return (Series_rep*) new binary_lex_series_rep<Op,C,V> (f, g);
}

template<typename Op,typename C,typename V>
class binary_implicit_lex_series_rep: public Implicit_series_rep {
protected:
  Series f, g;
public:
  binary_implicit_lex_series_rep (const Series& f2, const Series& g2):
    Implicit_series_rep (max (f2->d, g2->d)), f (f2), g (g2) {}
  syntactic expression () const {
    return Op::op (flatten (f), flatten (g)); }
  virtual void increase_precision (const Precision& p) {
    Implicit_series_rep::increase_precision (p);
    mmx::increase_precision (f, p);
    mmx::increase_precision (g, p); }
  Series initialize (Series& init, Monomial& lim) {
    if (f->d == 0 && g->d == 0)
      return Series (Op::op (f[V::monomial ()], g[V::monomial ()]), 0);
    else if (Op::nr_init () > 0) {
      ASSERT (Op::nr_init () <= 1, "wrong number of initial conditions");
      init= Op::op (f[0], g[0]);
      lim = V::monomial (this->d - 1, this->d, 1);
    }
    return Op::def (Series (this->me ()), f, g);
  }
};

template<typename Op,typename C,typename V> inline Series
binary_implicit_lex_series (const Series& f, const Series& g) {
  Series_rep* rep= new binary_implicit_lex_series_rep<Op,C,V> (f, g);
  return implicit (Series (rep));
}

template<typename Op,typename C,typename V,typename X>
class binary_scalar_lex_series_rep: public Series_rep {
protected:
  Series f;
  X x;
public:
  inline binary_scalar_lex_series_rep (const Series& f2, const X& x2):
    Series_rep (f2->d), f(f2), x (x2) {}
  syntactic expression () const {
    return Op::op (flatten (f), flatten (x)); }
  virtual void increase_precision (const Precision& p) {
    Series_rep::increase_precision (p);
    mmx::increase_precision (f, p); }
  virtual C compute (const Monomial& m) { return Op::op (f[m], x); }
};

template<typename Op,typename C,typename V,typename X> inline Series
binary_scalar_lex_series (const Series& f, const X& x) {
  return (Series_rep*)
    new binary_scalar_lex_series_rep<Op,C,V,X> (f, x);
}

/******************************************************************************
* Coefficient-wise operations
******************************************************************************/

TMPL Series
operator + (const Series& f, const Series& g) {
  return binary_lex_series<add_op> (f, g);
}

TMPL Series
operator + (const C& c, const Series& f) {
  return binary_lex_series<add_op> (Series (c), f);
}

TMPL Series
operator + (const Series& f, const C& c) {
  return binary_lex_series<add_op> (f, Series (c));
}

TMPL Series
operator - (const Series& f) {
  return unary_lex_series<neg_op> (f);
}

TMPL Series
operator - (const Series& f, const Series& g) {
  return binary_lex_series<sub_op> (f, g);
}

TMPL Series
operator - (const C& c, const Series& f) {
  return binary_lex_series<sub_op> (Series (c), f);
}

TMPL Series
operator - (const Series& f, const C& c) {
  return binary_lex_series<sub_op> (f, Series (c));
}

TMPL Series
operator * (const Series& f, const C& c) {
  return binary_scalar_lex_series<rmul_op> (f, c);
}

TMPL Series
operator * (const C& c, const Series& f) {
  return binary_scalar_lex_series<lmul_op> (f, c);
}

TMPL Series
operator / (const Series& f, const C& c) {
  return binary_scalar_lex_series<rdiv_op> (f, c);
}

TMPL Series
quo (const Series& f, const C& c) {
  return binary_scalar_lex_series<rquo_op> (f, c);
}

TMPL Series
rem (const Series& f, const C& c) {
  return binary_scalar_lex_series<rrem_op> (f, c);
}

/******************************************************************************
* Other linear time operations
******************************************************************************/

TMPL
class lshiftz_lex_series_rep: public Series_rep {
  Series f;
  int sh;
public:
  lshiftz_lex_series_rep (const Series& f2, int sh2):
    Series_rep (f2->d), f (f2), sh (sh2) {}
  syntactic expression () const {
    return lshiftz (flatten (f), flatten (sh)); }
  void increase_precision (const Precision& p) {
    Series_rep::increase_precision (p);
    mmx::increase_precision (f, Helper::shift (p, -sh, this->d - 1)); }
  C compute (const Monomial& m) {
    nat d= this->d;
    return sh > m[d-1]? C(0): f[V::divide (m, V::monomial (d-1, d, sh))]; }
};

TMPL Series
lshiftz (const Series& f, const int& shift) {
  return (Series_rep*) new lshiftz_lex_series_rep<C,V> (f, shift);
}

TMPL
class restrict_lex_series_rep: public Series_rep {
  Series   f;
  Variable v;
  nat      deg;
public:
  restrict_lex_series_rep
    (const Series& f2, const Variable& v2, nat deg2):
      Series_rep (f2->d), f (f2), v (v2), deg (deg2) {}
  syntactic expression () const {
    return apply ("restrict", flatten (f), flatten (v), flatten (deg)); }
  void increase_precision (const Precision& p) {
    Series_rep::increase_precision (p);
    mmx::increase_precision (f, p); }
  C compute (const Monomial& m) {
    if (N(m) < v || ((nat) m[v]) != deg) return C(0);
    else return f[m]; }
};

TMPL Series
restrict (const Series& f, const Variable& v, nat deg) {
  return (Series_rep*) new restrict_lex_series_rep<C,V> (f, v, deg);
}

TMPL
class derive_lex_series_rep: public Series_rep {
  Series f;
  Variable v;
public:
  derive_lex_series_rep (const Series& f2, const Variable& v2):
    Series_rep (f2->d), f (f2), v (v2) {}
  syntactic expression () const {
    return derive (flatten (f), flatten (v)); }
  void increase_precision (const Precision& p) {
    Series_rep::increase_precision (p);
    mmx::increase_precision (f, Helper::shift (p, 1, v)); }
  C compute (const Monomial& m) {
    return C (m[v] + 1) * f[V::multiply (V::monomial (v, this->d), m)]; }
};

TMPL Series
derive (const Series& f) {
  return (Series_rep*) new derive_lex_series_rep<C,V> (f, f->d-1);
}

TMPL Series
derive (const Series& f, const Variable& v) {
  return (Series_rep*) new derive_lex_series_rep<C,V> (f, v);
}

TMPL
class xderive_lex_series_rep: public Series_rep {
  Series f;
  Variable v;
public:
  xderive_lex_series_rep (const Series& f2, const Variable& v2):
    Series_rep (f2->d), f (f2), v (v2) {}
  syntactic expression () const {
    return derive (flatten (f), flatten (v)); }
  void increase_precision (const Precision& p) {
    Series_rep::increase_precision (p);
    mmx::increase_precision (f, p); }
  C compute (const Monomial& m) {
    return C(m[v]) * f[m]; }
};

TMPL Series
xderive (const Series& f, const Variable& v) {
  return (Series_rep*) new xderive_lex_series_rep<C,V> (f, v);
}

TMPL
class integrate_lex_series_rep: public Series_rep {
  Series f;
  Variable v;
  C c;
public:
  integrate_lex_series_rep
    (const Series& f2, const Variable& v2, const C& c2= 0):
      Series_rep (max (f2->d, v2)), f (f2), v (v2), c (c2) {
        ASSERT (v2 < 1000, "suspicious integration variable"); }
  syntactic expression () const {
    if (c == C(0)) return integrate (flatten (f), flatten (v));
    else return integrate (flatten (f), flatten (v)) + flatten (c); }
  void increase_precision (const Precision& p) {
    Series_rep::increase_precision (p);
    mmx::increase_precision (f, Helper::shift (p, -1, v)); }
  C compute (const Monomial& m) {
    if (m == 0) return c;
    else if (m[v] == 0) return C (0);
    else return f[V::divide (m, V::monomial (v, this->d))] / C (m[v]); }
};

TMPL Series
integrate (const Series& f) {
  return (Series_rep*) new integrate_lex_series_rep<C,V> (f, f->d-1);
}

TMPL Series
integrate (const Series& f, const Variable& v) {
  return (Series_rep*) new integrate_lex_series_rep<C,V> (f, v);
}

TMPL Series
integrate_init (const Series& f, const C& c) {
  return (Series_rep*)
    new integrate_lex_series_rep<C,V> (f, f->d-1, c);
}

TMPL Series
integrate_init (const Series& f, const Variable& v, const C& c) {
  return (Series_rep*) new integrate_lex_series_rep<C,V> (f, v, c);
}

/******************************************************************************
* Multiplication and division
******************************************************************************/

TMPL
class mul_lex_series_rep:
  public binary_lex_series_rep<mul_op,C,V> {
public:
  inline mul_lex_series_rep (const Series& f, const Series& g):
    binary_lex_series_rep<mul_op,C,V> (f, g) {}
  C compute (const Monomial& m) {
    C acc= 0;
    for (iterator<Monomial > it= before (m); busy (it); ++it)
      acc += this->f[*it] * this->g[V::divide (m, *it)];
    return acc;
  }
};

TMPL Series
operator * (const Series& f, const Series& g) {
  return (Series_rep*) new mul_lex_series_rep<C,V> (f, g);
}

struct var_div_op: public div_op {
  template<typename C> static inline C
  def (const C& me, const C& f, const C& g) {
    return (f - lshiftz (me * rshiftz (g))) / g[0]; }
};

TMPL Series
operator / (const C& c, const Series& f) {
  return binary_implicit_lex_series<var_div_op> (Series (c), f);
}

TMPL Series
operator / (const Series& f, const Series& g) {
  if (f->d > g->d) return f * (C(1) / g);
  return binary_implicit_lex_series<var_div_op> (f, g);
}

ARITH_INT_SUGAR(TMPL,Series,C);

/******************************************************************************
* Elementary functions
******************************************************************************/

struct var_sqrt_op: public sqrt_op {
  template<typename C> static inline C
  def (const C& me, const C& f) {
    return integrate (derive (f) / (2*me)); }
};

TMPL Series
sqrt (const Series& f) {
  return unary_implicit_lex_series<var_sqrt_op> (f);
}

TMPL Series
sqrt_init (const Series& f, const Series& c) {
  return unary_implicit_lex_series<var_sqrt_op> (f, c);
}

TMPL Series
exp (const Series& f) {
  return unary_implicit_lex_series<exp_op> (f);
}

TMPL Series
log (const Series& f) {
  return unary_implicit_lex_series<log_op> (f);
}

TMPL Series
log_init (const Series& f, const Series& c) {
  return unary_implicit_lex_series<log_op> (f, c);
}

TMPL Series
pow (const Series& f, const C& c) {
  return exp (c * log (f));
}

TMPL Series
pow (const Series& f, const Series& g) {
  return exp (g * log (f));
}

// NOTE: cos, sin and tan defined in lex_series_linear.hpp

TMPL Series
acos (const Series& f) {
  return unary_implicit_lex_series<acos_op> (f);
}

TMPL Series
acos_init (const Series& f, const Series& c) {
  return unary_implicit_lex_series<acos_op> (f, c);
}

TMPL Series
asin (const Series& f) {
  return unary_implicit_lex_series<asin_op> (f);
}

TMPL Series
asin_init (const Series& f, const Series& c) {
  return unary_implicit_lex_series<asin_op> (f, c);
}

TMPL Series
atan (const Series& f) {
  return unary_implicit_lex_series<atan_op> (f);
}

TMPL Series
atan_init (const Series& f, const Series& c) {
  return unary_implicit_lex_series<atan_op> (f, c);
}

INV_TRIGO_SUGAR(TMPL,Series)
HYPER_SUGAR(TMPL,Series)
INV_HYPER_SUGAR(TMPL,Series)
ARG_HYPER_SUGAR(TMPL,Series)

#undef TMPL_DEF
#undef TMPL
#undef Series
#undef Series_rep
#undef Implicit_series_rep
#undef Variable
#undef Monomial
#undef Helper
#undef Truncated
#undef Precision
} // namespace mmx
#endif // __MMX_LEX_SERIES_HPP
