
/******************************************************************************
* MODULE     : transcoefficient.hpp
* DESCRIPTION: Transcoefficient numbers
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_TRANSCOEFFICIENT_HPP
#define __MMX_TRANSCOEFFICIENT_HPP
#include <basix/table.hpp>
#include <basix/timer.hpp>
#include <numerix/ball.hpp>
#include <numerix/twin.hpp>
namespace mmx {

struct var_twin {};

STMPL
class twin<ball<floating<> >,generic,var_twin> {
MMX_ALLOCATORS;
private:
  ball<floating<> > x1;
  generic           x2;

protected:
  static nat random_value;
  static table<ball<floating<> >,generic> parameter_value;
  inline ball<floating<> > as_ball (const generic& x) {
    if (!parameter_value->contains (x)) {
      random_value= (random_value << 7) ^ mmx_time ();
      nat mask= (1 << 30) - 1;
      ball<floating<> > b ((int) (random_value & mask));
      parameter_value[x]= b >> 30;
      cout << x << " -> " << parameter_value[x] << "\n";
    }
    return parameter_value[x];
  }

public:
  inline twin (): x1 (0), x2 (0) {}
  inline twin (const int& x):
    x1 (x), x2 (construct (x)) {}
  inline twin (const integer& x):
    x1 (x), x2 (construct (x)) {}
  inline twin (const rational& x):
    x1 (x), x2 (construct (x)) {}
  inline twin (const generic& x):
    x1 (as_ball (x)), x2 (x) {}
  inline twin (const ball<floating<> >& y1, const generic& y2):
    x1 (y1), x2 (y2) {}
  template<typename V> inline
  twin (const twin<ball<floating<> >,generic,V>& x):
    x1 (car (x)), x2 (cdr (x)) {}
  friend ball<floating<> >
  car LESSGTR (const twin<ball<floating<> >,generic,var_twin>& x);
  friend generic
  cdr LESSGTR (const twin<ball<floating<> >,generic,var_twin>& x);
};

typedef twin<ball<floating<> >,generic,var_twin> transcoefficient;

}
#endif // __MMX_TRANSCOEFFICIENT_HPP
