
/******************************************************************************
* MODULE     : complex_double.hpp
* DESCRIPTION: Complexified doubles
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_COMPLEX_DOUBLE_HPP
#define __MMX_COMPLEX_DOUBLE_HPP
#include <basix/double.hpp>
#include <numerix/complex.hpp>
namespace mmx {

/******************************************************************************
* Printing complex doubles
******************************************************************************/

template<> inline syntactic
flatten (const complex<double>& z) {
  if (abs (Im (z)) < abs (Re (z)) * 1.0e-4)
    return flatten (Re (z));
  if (abs (Re (z)) < abs (Im (z)) * 1.0e-4)
    return gaussian (flatten (0), flatten (Im (z)));
  else return gaussian (flatten (Re (z)), flatten (Im (z)));
}

/******************************************************************************
* Avoid overflows for complex<double>
******************************************************************************/

STMPL inline complex<double>
operator / (const complex<double>& z1, const complex<double>& z2) {
  if (abs (Re (z2)) >= abs (Im (z2))) {
    double f= Im (z2) / Re (z2);
    double c= 1.0 / (Re (z2) + Im (z2) * f);
    return complex<double> (c * (Re (z1) + Im (z1) * f),
			       c * (Im (z1) - Re (z1) * f));
  }
  else {
    double f= Re (z2) / Im (z2);
    double c= 1.0 / (Re (z2) * f + Im (z2));
    return complex<double> (c * (Re (z1) * f + Im (z1)),
			       c * (Im (z1) * f - Re (z1)));
  }
}

// TODO: abs, sqrt and in-place operations

} // namespace mmx
#endif // __MMX_COMPLEX_DOUBLE_HPP
