
/******************************************************************************
* MODULE     : mmx_gmp.hpp
* DESCRIPTION: Interface to gmp
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_GMP_HPP
#define __MMX_GMP_HPP
#include <gmp.h>
#include <basix/exception.hpp>

namespace mmx {

#define BITS_PER_LIMB __GMP_BITS_PER_MP_LIMB
#define DEFAULT_LOW   __GMP_BITS_PER_MP_LIMB

extern gmp_randstate_t gmp_randstate;

inline void mpz_add_si (mpz_t dest, const mpz_t arg, long int plus) {
  if (plus >= 0) mpz_add_ui (dest, arg, plus);
  else mpz_sub_ui (dest, arg, -plus); }
inline void mpz_sub_si (mpz_t dest, const mpz_t arg, long int plus) {
  if (plus >= 0) mpz_sub_ui (dest, arg, plus);
  else mpz_add_ui (dest, arg, -plus); }
inline void mpz_si_sub (mpz_t dest, long int plus, const mpz_t arg) {
  if (plus >= 0) mpz_ui_sub (dest, plus, arg);
  else { mpz_add_ui (dest, arg, -plus); mpz_neg (dest, dest); } }
inline void mpz_mul_2si (mpz_t dest, const mpz_t arg, long int shift) {
  if (shift >= 0) mpz_mul_2exp (dest, arg, shift);
  else mpz_fdiv_q_2exp (dest, arg, -shift); }
inline void mpq_mul_2si (mpq_t dest, const mpq_t arg, long int shift) {
  if (shift >= 0) mpq_mul_2exp (dest, arg, shift);
  else mpq_div_2exp (dest, arg, -shift); }

}
#endif // __MMX_GMP_HPP
