
/******************************************************************************
* MODULE     : modulus_naive_integer.hpp
* DESCRIPTION: modulus for modular arithmetic over integers
* COPYRIGHT  : (C) 2008  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__MODULUS_NAIVE_INTEGER__HPP
#define __MMX__MODULUS_NAIVE_INTEGER__HPP
#include <numerix/integer.hpp>
#include <numerix/modulus_naive.hpp>

namespace mmx {

#define TMPL template<typename C, typename M>

/******************************************************************************
* Variant
******************************************************************************/
// The modulus 'p' is always taken to be positive
// The modulars are in the range [0, p-1]

struct modulus_normalization_naive_integer {

  template<typename C> static inline bool
  normalize (C& p) {
    if (p < 0) p = -p;
    return true;
  }
};

template<typename V>
struct modulus_inv_naive_integer : public V {

  TMPL static inline void
  inv_mod (C& a, const M& m) {
    if (m.p == 0) {
      if (a == 1 || a == -1)
	return;
      else
	ERROR ("inv_mod: argument is not invertible");  
    }
    a = invert_modulo (a, m.p);
    if (a == 0 && m.p != 1)
      ERROR ("inv_mod: argument is not invertible"); }

  TMPL static inline void
  inv_mod (C& dest, const C& s, const M& m) {
    dest = s;
    inv_mod (dest, m); }
};

template<typename V>
struct modulus_encoding_naive_integer : public V {
  TMPL static inline void 
  encode_mod (C& dest, const C& s, const M& m) {
    if (s < 0 && m.p != 0) {
      dest = - s;
      V::reduce_mod (dest, m);
      dest = m.p - dest;
    }
    else
      dest = s;
      V::reduce_mod (dest, m);
  }
  TMPL static inline void 
  decode_mod (C& dest, const C& s, const M& m) {
    (void) m;
    dest = s; }
};

struct modulus_naive_integer:
  public modulus_encoding_naive_integer<
         modulus_div_naive<
         modulus_inv_naive_integer<
	 modulus_mul_naive<
	 modulus_add_naive<
         modulus_reduction_naive<
	 modulus_normalization_naive_integer> > > > > > {};

/******************************************************************************
* Helpers specializations for hardware integers
******************************************************************************/

template<>
struct modulus_variant_helper<integer> {
  typedef modulus_naive_integer MV; };

#undef TMPL
} // namespace mmx
#endif //__MMX__MODULUS_NAIVE_INTEGER__HPP
