#ifndef SYNAPS_MPOL_MPOLDST_CONVERT_H
#define SYNAPS_MPOL_MPOLDST_CONVERT_H
#include <subdivix/type.hpp>

namespace mmx {

  namespace MPOLDST {
    //--------------------------------------------------------------------
    /** Convert a multivariate polynomial to univariate one, formally by substuting
	all the variables but the variable of index \c ind by 1.
    **/
     template<class UPOL,class MPOL>
     UPOL convert(const MPOL & p, unsigned ind, meta::As<UPOL> )
    {
      typedef typename UPOL::value_type coeff_t;
      unsigned N = degree(p,ind)+1;
      UPOL r(N, AsSize());
      for(typename MPOL::const_iterator it=p.begin(); it != p.end(); it++)
	r[(*it)[ind]]+= monomial_convert(*it,ind,meta::As<coeff_t>());
      return r;
    }

    //--------------------------------------------------------------------
    /** Convert a multivariate polynomial to univariate polynomial in the variable i1 
	with coefficients which are univariate polynomials in the variable i2. The other 
	variables are replace by 1.
    **/
    template<class UPOL,class MPOL>
    UPOL convert(const MPOL & p, unsigned i1, unsigned i2, meta::As<UPOL>)
    {
      typedef typename UPOL::value_type coeff_t;
      unsigned N = degree(p,i1)+1;
      int d1;
      UPOL r(N, AsSize());
      //      std::fill( p.begin(), p.end(), coeff_t(0));
      for(typename MPOL::const_iterator it=p.begin(); it != p.end(); it++)
	{
	  d1 = (*it)[i1];
	  //	  if (d1>=0) 
	  {
	    r[d1] += monomial_convert(*it, i2, meta::As<coeff_t>());
	  }
	}
      return r;
    }
  }
}
#endif // SYNAPS_MPOL_MPOLDST_CONVERT_H
