/*********************************************************************
*      This file is part of the source code of SYNAPS kernel.        *
*   Author(s): B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
#ifndef SYNAPS_MPOL_Binomial_H
#define SYNAPS_MPOL_Binomial_H
//#include <algebrix/pow.hpp>

#include <assert.h>

//======================================================================
namespace mmx {
//======================================================================
#ifndef WITH_POW
#define WITH_POW
  template<class T>
  T pow(const T& a, int i)
  {
    assert(i>=0);
    if(i == 1) return a;
    if(i > 0)
      {
	if (a == T(1))  return a;
	if (a == T(-1)) return (i % 2  == 0) ? T(1) : T(-1);
	T y(a);
	T z(1);
	unsigned int n = i;
	unsigned int m;
	while (n > 0) {
	  m = n / 2;
	  if (n %2 ) 
	    {
	      z *= y;
	    }
	  y = y * y;
	  n = m;
	}
	return z;
      }
    return T(1);
  }
#endif //WITH_POW
  //--------------------------------------------------------------------
/** Binomial coefficients **/
template <typename T>  T binomial(int n, int p)
{
  T n1=1;

  if(p!=0)
    {
      if((n-p)<p) p=n-p;
      int n2=n;
      for(int i=1;i<=p;i++,n2--) n1=n1*n2/T(i);
    }

  return(n1);
}

//--------------------------------------------------------------------
/** Expansion of coeff*(x_i+a)^d, where
    i is the index of the variable x,
    a is the shift,
    d is the degree. **/
template < class MP >
MP binomial( typename MP::coeff_t coeff, int i, int d, typename MP::coeff_t a)
{
  MP polynom;
  typename MP::coeff_t tmp;
  for ( int k = 0 ; k <= d ; k++ )
    {
      tmp =  coeff * binomial<typename MP::coeff_t>(d, k) * pow(a,d-k);
      polynom += MP( typename MP::monom_t( tmp , i, k ) );
    }
  return polynom;
}
//======================================================================
} //  namespace mmx 
//======================================================================
#endif // SYNAPS_MPOL_Binomial_H


