
/********************************************************************
 *   This file is part of the source code of the SYNAPS library.
 *   Author(s): F.Carreras, B.Mourrain, O.Ruatta, C.Tanasescu
 *   $Id: Gcd.hpp,v 1.1 2005/07/11 08:12:26 mourrain Exp $
 ********************************************************************/
#ifndef SYNAPS_UPOL_GCD_H
#define SYNAPS_UPOL_GCD_H
//--------------------------------------------------------------------
//#include <algebrix.hpp>
#include <subdivix/type.hpp>
#include <subdivix/upoldse.hpp>
#include <subdivix/sturm_seq.hpp>
//--------------------------------------------------------------------

namespace mmx {

template<class C> void contentcoeff( C & a, const C & c ) { a = c; };
template<class C, class POL>
void contentcoeff( C & tmp, const POL & p)
{
  //  typedef typename POL::value_type coeff_type;
  //  typedef typename POL::size_type size_type;
  tmp = C(0);
  int d = degree(p);
  for(int i = 0; i <= d; i++)
    {
      C b;
      contentcoeff(b,p[i]);
      if ( b != C(0) )
	{
	  tmp = gcd(tmp,b);
	};
    };
}
template<class C, class UPOL> 
  C contentcoeff( const UPOL & p ) { C tmp; contentcoeff(tmp,p); return tmp; };
/**
           content of a polynomial over an euclidian ring
**/
template<class POL>
typename POL::value_type content(const POL & p)
{
  typedef typename POL::value_type coeff_type;
  typedef typename POL::size_type size_type;

  if ( p ==0 ) return coeff_type(1);
  else 
    {
  size_type d = degree(p);
  coeff_type tmp = p[0];
  for(size_type i = 1; i <= d; i++)
    tmp = gcd(tmp,p[i]);
  return  tmp;
    }
}
/**
         Content of a polynomial with double coefficient
**/
template<class R>
double content(const upoldse<double, R > & p)
{
  double r=1.;
  if (degree(p)>=0)
        r = p[degree(p)];
  return r;
}


/**
 *                     GCD by Sturm-Habicht sequence
 **/

template<class C, class R>
upoldse<C, R> gcd(const upoldse<C, R> & P, const upoldse<C, R> & Q)
{
  typedef SturmSeq<C,upoldse<C,R>,std::vector<upoldse<C,R> > > stha_t;
  if ((degree(P) < 0))
	return Q;
  else if (degree(Q) < 0)
	return P;
  else if ((degree(P) == 0) || (degree(Q) ==0))
	return upoldse<C,R>(1);
  else {
    stha_t s(P,Q,HABICHT());
    typename stha_t::reverse_iterator it = s.rep().rbegin();
    for(; (it!= s.rep().rend()) && (*it == 0) ; ++it);
    upoldse<C,R> RES=*it;
    C tmp=content(RES);
    RES/=tmp;
    return RES;
  }
}


template<class C, class R>
upoldse<C,R> lcm(const upoldse<C,R> & p, const upoldse<C,R> & q)
{
  return (p/gcd(p,q))*q;
}


// 
template<class C>
upoldse<upoldse<C> > gcd2(const upoldse<upoldse<C> > & p,const upoldse<upoldse<C> > & q)
{
  typedef upoldse<upoldse<C> > PolPol;
  typedef upoldse<C> Pol;
  PolPol g = gcd(p,q);
  if (degree(g) >= 0){
    C tmp = content(g[0]);
    for(int i=1;i <= degree(g);++i){
      if(g[i]!=0)
	tmp = gcd(tmp,content(g[i]));
    }
    for(int i=0;i<=degree(g);i++)
      g[i]/=tmp;
  }
  return g;
}

/** Function to consider bivariate polynomial of two variables as a 
* polynomials of one variable with coefficients univariate polynomials of 
* the other variable and that remove content  
**/ 
template<class C>
upoldse<upoldse<C> > remove_content2(const upoldse<upoldse<C> > & p)
{
  typedef upoldse<upoldse<C> > PolPol;
  typedef upoldse<C> Pol;
  PolPol res = p;
  Pol g = res[0];
  /** loop to compute the content **/
  for(int i = 1; i <= degree(res); i++){
    if (p[i] != 0) 
      g = gcd(g,res[i]);
  }
  /** remove content if it is needed **/
  if (degree(g) != 0 ){
    for(int i = 0; i <= degree(res); i++)
      res[i] = res[i]/g;
  }
  return res;
}


template<class C> inline
bool gcd_coeffs_stop(const C& p, const meta::false_t & ) { return p == C(1); };
template<class C> inline
bool gcd_coeffs_stop(const C& p, const meta::true_t &  ) { return degree(p) == 0; };
template<class C> inline
bool gcd_coeffs_stop(const C& p ) { return gcd_coeffs_stop(p, typename ispolynomial<C>::T() ); };

template< typename C, class R >
C gcd_coeffs( const upoldse<C,R> & pol )
{
  typedef upoldse<C,R> polynom;
  typedef typename polynom::const_iterator iterator;
  typedef typename polynom::value_type coeff;
  coeff p;
  iterator b = pol.begin();
  ++b;
  p = gcd( *(pol.begin()), *b );
  ++b;
  while( b != pol.end() && ! gcd_coeffs_stop(p)  ) { 
    p = gcd(p,*b);
    ++b; 
  };
  
  if ( b != pol.end() ) return C(1); 
  
  return p;
};
  
  


} //namespace mmx
#endif // SYNAPS_UPOL_GCD_H

