#ifndef SYNAPS_LINALG_ARRAY2D_H
#define SYNAPS_LINALG_ARRAY2D_H
//======================================================================
//#include <algebrix/Norm.hpp>
#include <subdivix/MATRDSE_m.hpp>
//#include <subdivix/iterator_stride.hpp>
#include <subdivix/linalg_rep1d.hpp>


namespace mmx {

//======================================================================
namespace linalg {
//----------------------------------------------------------------------
  typedef unsigned size_type;
  //extern double TOL;
//----------------------------------------------------------------------
/** 
   Class derived from @linalg::rep1d<C>@, adding the 
   number of rows  @nbrow_@ and the number of columns @nbcol_@
   of the matrix. The matrix is stored by column in the array @rep1d<C>@.
   The function @resize(m,n)@ allocates the size for a matrix of size
   $m \times n$.
**/
template <class C>
struct rep2d : public rep1d<C> {
  typedef C*                              col_iterator;
//  typedef stride_iterator<C,C*,size_type> row_iterator;
 
  size_type nbrow_, nbcol_, lda_;
 
  rep2d() {}
  rep2d(const rep2d<C> & v): 
    linalg::rep1d<C>(v), nbrow_(v.nbrow_), nbcol_(v.nbcol_), lda_(v.lda_) {}
  rep2d(size_type i,size_type j=1): 
    linalg::rep1d<C>(i*j), nbrow_(i),nbcol_(j), lda_(i) {}
  rep2d(size_type i, size_type j, C* t): 
    linalg::rep1d<C>(i*j,t), nbrow_(i), nbcol_(j), lda_(i)  {}
  rep2d(const char * nm);

  rep2d<C> & operator=(const rep2d<C> & v);
 
  C &       operator()(size_type i, size_type j) 
  {
    assert(i < nbrow_ && j < nbcol_);
    return this->tab_[i+lda_*j];
  }

  const C & operator()(size_type i, size_type j) const 
  {
    return this->tab_[i+lda_*j];
  }  

  size_type nbrow() const {return nbrow_;}
  size_type nbcol() const {return nbcol_;}
  size_type lda()   const {return lda_;}

  void resize(size_type i,size_type j)
    {
      nbrow_=i; nbcol_=j; lda_ = i;  
      this->linalg::rep1d<C>::resize(i*j); 
    }
  
};
//--------------------------------------------------------------------
template <class C>
rep2d<C>::rep2d(const char * nm)
{
  std::ifstream ins(nm, std::ios::in);
  ins >> nbrow_ >> nbcol_;
  lda_ = nbrow_;
  this->size_ = this->nbrow_*this->nbcol_; this->tab_ = new C[this->size_];

  for(size_type i=0; i< this->size_; i++) ins >>this->tab_[i];
  ins.close();
}
//----------------------------------------------------------------------
template <class C>
rep2d<C> & rep2d<C>::operator=(const rep2d<C> & v) 
{
  this->rep1d<C>::operator=(v);
  this->nbrow_ = v.nbrow(); this->nbcol_ = v.nbcol(); this->lda_ = v.lda();
  return *this;
}
//----------------------------------------------------------------------
template<class C>
typename linalg::rep2d<C>::col_iterator 
  create(typename linalg::rep2d<C>::col_iterator s,
	 const linalg::rep2d<C> & a, unsigned int i, unsigned int j) 
{
  typedef  typename linalg::rep2d<C>::col_iterator iterator;
  return iterator(&a(i,j));
}
//----------------------------------------------------------------------
// template<class C>
// typename linalg::rep2d<C>::row_iterator 
//   create(typename linalg::rep2d<C>::row_iterator s,
// 	 const linalg::rep2d<C> & a, unsigned int i, unsigned int j) 
// {
//   typedef  typename linalg::rep2d<C>::row_iterator iterator;
//   return iterator((C*)(&a(i,j)),a.lda());
// }
// //----------------------------------------------------------------------
// template<class C>
// typename linalg::rep2d<C>::row_iterator 
//   begin_row(const linalg::rep2d<C> & a, unsigned int i) 
// {
//   return typename linalg::rep2d<C>::row_iterator((C*)(&a(i,0)),a.lda());
// }
//----------------------------------------------------------------------
template<class C>
typename linalg::rep2d<C>::col_iterator 
  begin_col(const linalg::rep2d<C> & a, unsigned int i) 
{
  return typename linalg::rep2d<C>::col_iterator((C*)(&a(0,i)));
}
//----------------------------------------------------------------------
// template<class C>
// typename linalg::rep2d<C>::row_iterator 
//   end_row(const linalg::rep2d<C> & a, unsigned int i) 
// {
//   return typename linalg::rep2d<C>::row_iterator((C*)(&a(i,a.nbcol())),a.lda());
// }
//----------------------------------------------------------------------
template<class C>
typename linalg::rep2d<C>::col_iterator 
  end_col(const linalg::rep2d<C> & a, unsigned int i) 
{
  return typename linalg::rep2d<C>::col_iterator(&a(a.lda(),i));
}
//----------------------------------------------------------------------
} //CLOSE linalg
//====================================================================
} //namespace mmx
//----------------------------------------------------------------------
#endif // SYNAPS_LINALG_ARRAY2D_H

