//#include "matrdse_row_col_access.hpp"

namespace mmx {

/// matrix += matrix
template<class C1, class R1, 
	 class C2, class R2 > inline
void add( matrdse<C1,R1> & r, const matrdse<C2,R2> & a )
{
  using namespace MATRDSE;
  add(r.rep(),a.rep());

};

/// matrix = matrix + matrix
template<class C1, class R1, 
	 class C2, class R2, 
	 class C3, class R3> inline
void add( matrdse<C1,R1> & r, const matrdse<C2,R2> & a, const matrdse<C3,R3> & b )
{
  if ( &r == &a ) { add(r,b); return;};
  if ( &r == &b ) { add(r,a); return;};
  r.rep().resize(a.nbrow(),b.nbcol());
  using namespace MATRDSE;
  add(r.rep(),a.rep(),b.rep());

};

/// matrix = matrix - matrix
template<class C1, class R1,
	 class C2, class R2, 
	 class C3, class R3> inline
void sub( matrdse<C1,R1>& r, const matrdse<C2,R2> & a, const matrdse<C3,R3> & b )
{
  using MATRDSE::sub;
  r.rep().resize(a.nbrow(),b.nbcol());
  sub(r.rep(),a.rep(),b.rep());
}

/// matrix = matrix * matrix
template<class C1, class R1,
	 class C2, class R2,
	 class C3, class R3> inline
void mul(matrdse<C1,R1> & r, const matrdse<C2,R2> & a, const matrdse<C3,R3> & b )
{
  using namespace MATRDSE;
  if ( &r == &a ) { mul(r,b); return; };
  if ( &r == &b ) { matrdse<C3,R3> cb = b; mul(r,a,cb); return; };
  r.rep().resize(a.nbrow(),b.nbcol()); 
  init(r,C1(0));
  mul(r.rep(),a.rep(),b.rep());
}

/// matrix = matrix * matrix
template<class C1, class R1,
	 class C2, class R2 >
void mul(matrdse<C1,R1> & r, const matrdse<C2,R2> & a )
{
  using namespace MATRDSE;
  matrdse<C1,R1> tmp = r;
  mul(r,tmp,a);
}
  
/// vector = matrix * vector
template<class Cr, class Rr, 
	 class Cm, class Rm, 
	 class Cv, class Rv >
void mul( vectdse<Cr,Rr> & r, const matrdse<Cm,Rm> & m, const vectdse<Cv,Rv> & v ) 
{
  {
    using namespace VECTOR;
    r.rep().resize( m.nbrow() );
    init(r.rep(),Cr(0));
  }
  using namespace MATRDSE;
  mul_vect(r.rep(),m.rep(),v.rep());
};

/// matrix = matrix * scalar
template<class C1, class R1, 
	 class C2, class R2, 
	 class C3> inline
void mul( matrdse<C1,R1> & r, const matrdse<C2,R2> & a, const C3 & b )
{
  using namespace MATRDSE;
  r.rep().resize( a.nbrow(),a.nbcol());
  mul_ext(r.rep(),a.rep(),b);
}

/// matrix = scalar * matrix
template<class C1, class R1, 
	 class C2, class R2, 
	 class C3> inline
void mul( matrdse<C1,R1> & r,  const C3 & b, const matrdse<C2,R2> & a)
{
  using namespace MATRDSE;
  r.rep().resize( a.nbrow(),a.nbcol());
  mul_ext(r.rep(),a.rep(),b);
}

/// matrix *= scalar
template<class C1, class R1, class C2> inline
void mul( matrdse<C1,R1> & r, const C2 & x )
{
  using namespace MATRDSE;
  mul_ext(r.rep(),x);
};

/// matrix = matrix / scalar
template<class C1, class R1, 
	 class C2, class R2,
	 class C3> inline
void div( matrdse<C1,R1> & r, const matrdse<C2,R2> & a, const C3 & x )
{
  using namespace MATRDSE;
  r.resize( a.nbrow(), a.nbcol() );
  div_ext(r.rep(),a.rep(),x);
};

/// matrix /= scalar 
template<class C1, class R1, class C2> inline
void div( matrdse<C1,R1> & r, const C2 & x )
{
  using namespace MATRDSE;
  div_ext(r.rep(),x);
};


// template<class C, class R>
// void submatrix( matrdse<C,R> & sm,   const matrdse<C,R> & m, const Range2d & rg )
// {
//   using namespace MATRDSE;
//   submatrix(sm.rep(),m.rep(),rg);
// };

// template<class Ca, class Ra, class Cb, class Rb>
// void col(  matrdse<Ca,Ra>& sm, const matrdse<Cb,Rb> & m, const Range & rg )
// {
//   submatrix(sm,m,Range2d(0,m.nbrow()-1,rg.i1,rg.i2));
// };

// template<class Ca, class Ra, class Cb, class Rb>
// void row(  matrdse<Ca,Ra>& sm, const matrdse<Cb,Rb> & m, const Range & rg )
// {
//   submatrix(sm,m,Range2d(rg.i1,rg.i2,0,m.nbcol()-1));
// };

/// Row of index @i@ of a matrix (local view).
template <class Ca, class Ra, class Cb, class Rb>  inline 
void row( vectdse<Ca,Ra> & rt, const matrdse<Cb,Rb> & a, int i)
{
  rt.resize( a.nbcol() );
  for ( int k = 0; k < rt.size(); k ++ ) rt[k] = a(i,k);
}

template<class C, class R> 
//vectdse<C, row_access< matrdse<C,R> > > 
vectdse<C> 
row( const matrdse<C,R> & m, int i )
{
  // return vectdse<C, row_access< matrdse<C,R> > >
  //   (row_access< matrdse<C,R> >(m,i));
    vectdse<C> r(m.nbcol(), AsSize()) ;
    for(unsigned j = 0 ; j < m.nbcol() ; j++)
	r[j] = m(i,j) ;
    return r ;
};

template<class C, class R>
vectdse<C>
col( const matrdse<C,R> & m, int i )
{
  // return vectdse<C, col_access< matrdse<C,R> > >
  //   (col_access< matrdse<C,R> >(m,i));
    vectdse<C> r(m.nbrow(), AsSize()) ;
    for(unsigned j = 0 ; j < m.nbrow() ; j++)
	r[j] = m(j,i) ;
    return r ;
};

//----------------------------------------------------------------------
/// Column of index @i@ of a matrix (local view).
template <class Ca, class Ra, class Cb, class Rb>  inline 
void col( vectdse<Ca,Ra> & rt, const matrdse<Cb,Rb> & a, int i)
{
  rt.resize( a.nbrow() );
  for ( int k = 0; k < rt.size(); k ++ ) rt[k] = a(k,i);
}

// #include <subdivix/traits_ringof.hpp>
// namespace numerix 
// {
// //   template<class C, class R> // matrdse<C,R> is exact as soon as C is exact
// //   struct isexact_< matrdse<C,R> > { typedef typename isexact< C >::T T; };
//   template<class C, class R> // matrdse<C,R> define a ring
//   struct isring_< matrdse<C,R> > { typedef true_t T; };
  
//   template<class Ca, class Ra, class Cb, class Rb > // result of ring operations over matrdse
//   struct ringof_< matrdse<Ca,Ra>, matrdse<Cb,Rb>  >  { 
//     typedef typename ringof<Ca,Cb>::T C;            // 
//     typedef typename instanceof<Ra,C>::T RaC;
//     typedef typename instanceof<Rb,C>::T RbC;
//     typedef typename sup_container<RaC,RbC>::T      R;
//     typedef matrdse<C,R>                            T;
//   };
  
//   template<class C, class R> struct structureof< matrdse<C,R> > { typedef structure::matrix T; };
// }

#define head template<class Ca, class Ra, class Cb, class Rb>
#define parm0 matrdse<Ca,Ra>
#define parm1 matrdse<Cb,Rb>
  declare_binary_operator(head,parm0,parm1,meta::_add_,operator+);
  declare_binary_operator(head,parm0,parm1,meta::_sub_,operator-);
  declare_binary_operator(head,parm0,parm1,meta::_mul_,operator*);
#undef head
#undef parm0
#undef parm1
#define head  template<class Ca,class Ra>
#define parm0 matrdse<Ca,Ra>
#define parm1 typename matrdse<Ca,Ra>::value_type
  declare_binary_operator(head,parm0,parm1,meta::_mul_,operator*);
  declare_binary_operator(head,parm1,parm0,meta::_mul_,operator*);
  declare_binary_operator(head,parm0,parm1,meta::_div_,operator/);
  declare_binary_operator(head,parm1,parm0,meta::_div_,operator/);
#undef parm1
#undef  head
#define head template<class Ca,class Ra,class K>
#define parm1 typename K::integer
  declare_binary_operator(head,parm0,parm1,meta::_mul_,operator*);
  declare_binary_operator(head,parm1,parm0,meta::_mul_,operator*);
  declare_binary_operator(head,parm0,parm1,meta::_div_,operator/);
  declare_binary_operator(head,parm1,parm0,meta::_div_,operator/);
#undef parm1
#undef  head
#define head template<class Ca,class Ra,class K>
#define parm1 typename K::floating
  declare_binary_operator(head,parm0,parm1,meta::_mul_,operator*);
  declare_binary_operator(head,parm1,parm0,meta::_mul_,operator*);
  declare_binary_operator(head,parm0,parm1,meta::_div_,operator/);
  declare_binary_operator(head,parm1,parm0,meta::_div_,operator/);
#undef parm1
#undef  head
#define head template<class Ca,class Ra,class K>
#define parm1 typename K::rational
  declare_binary_operator(head,parm0,parm1,meta::_mul_,operator*);
  declare_binary_operator(head,parm1,parm0,meta::_mul_,operator*);
  declare_binary_operator(head,parm0,parm1,meta::_div_,operator/);
  declare_binary_operator(head,parm1,parm0,meta::_div_,operator/);
#undef parm1
#undef head
#define head template<class Ca, class Ra, class Cb, class Rb>
#define parm1 vectdse<Cb,Rb>
  declare_binary_operator(head,parm0,parm1,meta::_mul_,operator*);
#undef head
#undef parm0
#undef parm1
#define head template<class C, class R>
#define parm0 matrdse<C,R>
// declare_binary_operator(head,parm0,Range2d,linalg::_submatrix_,submatrix);
#undef head
#undef parm0
#define head template<class C, class R>
#define parm0 matrdse<C,R>
//declare_binary_operator(head,parm0,int,linalg::_row_,row);
//declare_binary_operator(head,parm0,int,linalg::_col_,col);
//declare_binary_operator(head,parm0,Range,linalg::_row_,row);
//declare_binary_operator(head,parm0,Range,linalg::_col_,col);
#undef head
#undef parm0

} //namespace mmx
