/********************************************************************
 *   This file is part of the source code of the SYNAPS library.
 *   Description: rounding mode for Numerix floating type
 *   Author(s): G.M. Tzoumas, NKU Athens
 ********************************************************************/
#ifndef SYNAPS_MPFR_ROUNDING_MODE_HPP
#define SYNAPS_MPFR_ROUNDING_MODE_HPP
//--------------------------------------------------------------------

#include <numerix/floating.hpp>
#include <mpfr.h>
#include <subdivix/rounding_mode.hpp>
//--------------------------------------------------------------------
namespace mmx {
//--------------------------------------------------------------------
namespace numerics
{
  template<> struct LongVersion<floating<> > { typedef floating<> result_t; };
  template<> struct fpu_rounding<floating<> > 
  { 
      typedef mpfr_rnd_t rnd_t;
      inline static rnd_t  getrnd() { return mmx_rounding_mode; };
      inline static void   setrnd( rnd_t r ) { mmx_rounding_mode = r; };
      inline static rnd_t rnd_up() { return GMP_RNDU;};
      inline static rnd_t rnd_dw() { return GMP_RNDD;};
      inline static rnd_t rnd_nr() { return GMP_RNDN;};
      inline static rnd_t rnd_z()  { return GMP_RNDZ;};
      static const int id = 123;  // george: what's this?

      private:      
        rnd_t m_prev;
        bool  m_chg;
      public:
      
      fpu_rounding( rnd_t rnd ) 
      {
          m_chg = false;
          m_prev = getrnd();
          
          if ( m_prev != rnd ) 
          { 
              m_chg = true;
              setrnd(rnd); 
          };
      };
      ~fpu_rounding() { 
          if ( m_chg ) setrnd(m_prev);
      };
      
        
  };
    
}; //namespace numerics
//======================================================================
} //namespace mmx
//======================================================================
#endif
