#ifndef SYNAPS_MPOLDSE_VIEW_C
#define SYNAPS_MPOLDSE_VIEW_C
#include <algorithm>

namespace mmx {

template<class C, class R>
template<class X> inline
mpoldse<C,R>::mpoldse( const X& x )
{ 
  let::assign(*this,x); 
}

template<class C,class R> inline
mpoldse<C,R> & mpoldse<C,R>::operator= ( const char *   name ) { 
  let::assign(*this,mpoldst<C>(name));
  return *this;
};

template<class C,class R>
template<class X> inline
mpoldse<C,R> & mpoldse<C,R>::operator= ( const X & x ) { 
    using let::assign;
    assign(*this,x);
    return *this;
};

template<class C, class R> inline
mpoldse<C, R>::mpoldse( int nvr, const int * szs, const int * vrs ) : data(nvr,szs,vrs) {};

template < class C, class R > inline 
mpoldse < C, R >::mpoldse (const C & x): data (x) {};
template < class C, class R >
template < class O, class E > inline 
mpoldse < C, R >::mpoldse (const mpoldst < C, O, E > &mpol) : data (mpol) {};
template < class C, class R > inline
mpoldse < C, R >::mpoldse (const char *s) : data(mpoldst<C>(s)) {};

template<class C, class R> inline const C& mpoldse<C,R>::operator[]( int i ) const { return rep()[i]; };
template<class C, class R> inline  C& mpoldse<C,R>::operator[]( int i ) { return rep()[i]; };
template < class C, class R > inline 
mpoldse < C, R > &mpoldse < C, R >::operator+= (const mpoldse < C, R > &mp)
{
  using namespace multivariate_dense;
  add(rep(),mp.rep());
  return *this;
};

template < class C, class R > inline
mpoldse < C, R > & mpoldse < C, R >::operator+= (const C & x)
{
  using namespace multivariate_dense;
  add (rep (), x);
  return *this;
};

template < class C, class R > inline
mpoldse < C, R > &mpoldse < C, R >::operator-= (const mpoldse < C, R > &mp)
{
  using namespace multivariate_dense;
  sub (rep (), mp.rep ());
  return *this;
};

template < class C, class R > inline
  mpoldse < C, R > &mpoldse < C, R >::operator-= (const C & x)
{
  using namespace multivariate_dense;
  sub (rep (), x);
  return *this;
};

template < class C, class R > inline
  mpoldse < C, R > &mpoldse < C, R >::operator*= (const mpoldse < C, R > &mp)
{
  using namespace multivariate_dense;
  mul (rep (), mp.rep ());
  return *this;
};

template < class C, class R > inline
  mpoldse < C, R > &mpoldse < C, R >::operator*= (const C & x)
{
  using namespace multivariate_dense;
  mul (rep (), x);
  return *this;
};

template < class C, class R > inline
bool mpoldse < C, R >::operator== (const mpoldse & mp) const 
{
  return rep () == mp.rep ();
};

template < class C, class R >
template < class O, class E > inline
mpoldse < C, R > &mpoldse < C, R >::operator= (const mpoldst < C, O, E > &mpol)
{
  using namespace multivariate_dense;
  this->~mpoldse();
  new (this) mpoldse(mpol);
  return *this;
};

// texp / arithmetic
template < class C, class R > inline 
void add (mpoldse < C, R > &r, const mpoldse < C, R > &a )
{
  using namespace multivariate_dense;
  add (r.rep (), a.rep () );
};
template < class C, class R > inline 
void add (mpoldse < C, R > &r, const mpoldse < C, R > &a, const mpoldse < C, R > &b)
{
  using namespace multivariate_dense;
  add (r.rep (), a.rep (), b.rep ());
};

template < class C, class R > inline void
add (mpoldse < C, R > &r, const mpoldse < C, R > &a, const C & b)
{
  using namespace multivariate_dense;
  add (r.rep (), a.rep (), b);
};

template < class C, class R > inline void
add (mpoldse < C, R > &r, const C & a, const mpoldse < C, R > &b)
{
  using namespace multivariate_dense;
  add (r, b, a);
};

template < class C, class R > inline void
sub (mpoldse < C, R > &r, const mpoldse < C, R > &a )

{
  using namespace multivariate_dense;
sub (r.rep (), a.rep () );
}
;
template < class C, class R > inline void
sub (mpoldse < C, R > &r, const mpoldse < C, R > &a, const mpoldse < C,
     R > &b)
{
  using namespace multivariate_dense;
  sub (r.rep (), a.rep (), b.rep ());
};

template < class C, class R > inline void
sub (mpoldse < C, R > &r, const C & a, const mpoldse < C, R > &b)
{
  using namespace multivariate_dense;
  sub (r.rep (), a, b.rep ());
};

template < class C, class R > inline void
sub (mpoldse < C, R > &r, const mpoldse < C, R > &a, const C & b)
{
  using namespace multivariate_dense;
  sub (r.rep (), a.rep (), b);
};

template < class C, class R > inline void
mul (mpoldse < C, R > &r, const mpoldse < C, R > &a )
{
  using namespace multivariate_dense;
  mul (r.rep (), a.rep () );
};

template < class C, class R > inline void
mul (mpoldse < C, R > &r, const mpoldse < C, R > &a, const mpoldse < C,
     R > &b)
{
  using namespace multivariate_dense;
  mul (r.rep (), a.rep (), b.rep ());
};

template < class C, class R > inline void
mul (mpoldse < C, R > &r, const mpoldse < C, R > &a, const C & b)
{
  using namespace multivariate_dense;
  mul (r.rep (), a.rep (), b);
};

template < class C, class R > inline void
mul (mpoldse < C, R > &r, const C & a, const mpoldse < C, R > &b)
{
  using namespace multivariate_dense;
  mul (r.rep (), b.rep (), a);
};

/*
template < class C, class R >
template < class S >
mpoldse < C, R > &mpoldse < C, R >::operator= (const texp < S > &exp)
{
  eval (*this, exp);
  return *this;
};

template < class C, class R >
template < class S >
mpoldse < C, R > &mpoldse < C, R >::operator+= (const texp < S > &exp)
{
  mpoldse<C,R> tmp;
  eval(tmp,exp);
  add(*this,tmp);
  return *this;
};


template < class C, class R >
template < class S >
mpoldse < C, R > &mpoldse < C, R >::operator-= (const texp < S > &exp)
{
  mpoldse<C,R> tmp;
  eval(tmp,exp);
  sub(*this,tmp);
  return *this;
};

template < class C, class R >
template < class S >
mpoldse < C, R > &mpoldse < C, R >::operator*= (const texp < S > &exp)
{
  mpoldse<C,R> tmp;
  eval(tmp,exp);
  mul(*this,tmp);
  return *this;
};

*/
namespace meta {
  template < class C, class R > struct sup_ <C, mpoldse < C, R > >
  { typedef mpoldse < C, R > T; };

  template < class C, class R > struct sup_ <mpoldse < C, R >, C >
  { typedef mpoldse < C, R > T; };

  template<class Op, class Ca, class Ra, class Cb, class Rb> 
  struct binary_operator_prototype< Op, mpoldse<Ca,Ra>, mpoldse<Cb,Rb> >
  {
    typedef typename ringof<Ca,Cb>::T C;
    typedef typename instanceof<Ra,C>::T R;
    typedef mpoldse<C,R> F;
    typedef mpoldse<C,R> U;
    typedef mpoldse<C,R> V;
  };
  
  template<class Op, class Ca, class Ra, class Cb>
  struct binary_operator_prototype< Op, mpoldse<Ca,Ra>, Cb >
  {
    typedef typename ringof<Ca,Cb>::T C;
    typedef typename instanceof<Ra,C>::T R;
    typedef mpoldse<C,R> F;
    typedef mpoldse<C,R> U;
    typedef C V;
  };

  template<class Op, class Ca, class Ra, class Cb>
  struct binary_operator_prototype< Op, Cb,  mpoldse<Ca,Ra> >
  {
    typedef binary_operator_prototype<Op, mpoldse<Ca,Ra>, Cb > X;
    typedef typename X::F F;
    typedef typename X::U V;
    typedef typename X::V U;
  };
};  
//--------------------------------------------------------------------
template<class Ca,class Ra>
mpoldse<Ca,Ra> operator-(const mpoldse<Ca,Ra>& p) { return p* Ca(-1); }

#define head template<class Ca,class Ra, class Cb, class Rb>
#define parm0 mpoldse<Ca,Ra>
#define parm1 mpoldse<Cb,Rb>
  declare_binary_operator(head,parm0,parm1,meta::_add_,operator+);
  declare_binary_operator(head,parm0,parm1,meta::_sub_,operator-);
  declare_binary_operator(head,parm0,parm1,meta::_mul_,operator*);
#undef head
#undef parm0
#undef parm1

#define head template<class Ca,class Ra>
#define parm0 mpoldse<Ca,Ra>
#define parm1 Ca
   declare_binary_operator(head,parm0,parm1,meta::_mul_,operator+);
   declare_binary_operator(head,parm1,parm0,meta::_mul_,operator+);
   declare_binary_operator(head,parm0,parm1,meta::_mul_,operator-);
   declare_binary_operator(head,parm1,parm0,meta::_mul_,operator-);
   declare_binary_operator(head,parm0,parm1,meta::_mul_,operator*);
   declare_binary_operator(head,parm1,parm0,meta::_mul_,operator*);
#undef head
#undef parm0
#undef parm1


template < class C, class R >
std::ostream & operator<<(std::ostream & o, const mpoldse < C, R > &mp)
{
  print (o, mp.rep());
  return o;
};

template < class C, class R >
  mpoldse < C, R > operator^(const mpoldse < C, R > &a, unsigned n)
{
  return pow (a, n);
};

template < class C, class R > bool mpoldse < C, R >::operator== (const C & c) const 
{
  return rep () == c;
};

template < class C, class R >
  bool operator== (const C & c, const mpoldse < C, R > &mp)
{
  return mp.rep () == c;
};

template<class C, class R > inline 
int mpoldse<C,R>::nvars() const { return rep().nvr(); };


template<class C, class R> inline int mpoldse<C,R>::varid( int i )  const { 
  return rep().env.vrs()[i];
};



/** 
    Multivariate Polynomial Differentiation.
    Compute the partial derivative of @polynom@ w.r.t @v@ the vth variable of @polynom@.
    The result is stored in @derivative@.
 **/
template<class Coeff, class Rep> inline
mpoldse<Coeff,Rep> diff(const mpoldse<Coeff,Rep>& polynom, int v )
{
  mpoldse<Coeff,Rep> derivative;
  diff(derivative.rep(),polynom.rep(),v);
  return derivative;
};

/**
   Multivariate Polynomial Evaluation.
   The type {\tt Prm } must a container type, @parameters@ corresponds to the $0..nvr()$ coordinates.
   The computation is done in the {\tt Sup Type} of the {\tt Contained Type } of {\tt Prm} and {\tt Coeff}.
   The type {\tt Result } is the type of the result, it is assumed to have an operator= on the computation type.
**/
template<class Result, class Coeff, class Rep, class Parameters> inline
Result eval(const mpoldse<Coeff,Rep>& polynom, const Parameters& parameters ) 
{
  Result result;
  eval(result, polynom.rep(), parameters );
  return result;
};

} //namespace mmx
#endif
