# ifndef MPOLDST_LET_H
# define MPOLDST_LET_H

# include <subdivix/let.hpp>
# include <subdivix/mpoldst.hpp>

namespace mmx {

namespace let  {

  /** Convert a multivariate polynomial to a polynomial which double 
      coefficients.
  **/
  template<class U, class P, class S, class C, class O, class R> inline void assign(mpoldst<U,P,S>& r, const mpoldst<C,O,R> & p)
  {
      typedef typename mpoldst<double,P,S>::monom_t monom_t;
      U c;
      for(typename mpoldst<C,O,R>::const_iterator it=p.begin(); it !=p.end(); ++it) {
	let::assign(c,it->coeff());
	monom_t m(c,it->rep());
	r+=m;
      }
  }


  /// transform the polynomial the multivariate polynomial @mp@$(x_0,..,x_v,..,x_n)$
  /// as @p@$(x_v)$ with coefficient in $C[x_0,...,x_{v-1},x_{v+1},..,x_n]$
  template<class Ca, class Ra,
	   class Cb, class Ob, class Rb >
  void convert( upoldse<Ca,Ra> & p, const mpoldst<Cb,Ob,Rb> & mp, int v )
  {
    int dv;
    p.resize( (dv = degree(mp,v))+1 );
    std::fill( p.begin(), p.end(), 0);
	
    typedef typename mpoldst<Cb,Ob,Rb>::const_iterator pmit;
    typedef typename mpoldst<Cb,Ob,Rb>::monom_t monom;
    using namespace let;
    for ( pmit it = mp.begin(); it != mp.end(); it ++ )
      {
	monom m(*it);
	if ( m.size() > v ) m.rep()[v] = 0;
	p[ (*it)[v] ] += convert(m,meta::As<Ca>());
      };
  };
  
  template<class Ca, class Ra,
	   class Cb, class Ob, class Rb,
	   class Rp, class Ru>
  void convert( upoldse< upoldse<Ca,Ra>, Rp >& p, 
		const upoldse< mpoldst<Cb,Ob,Rb>, Ru >&  ump, int v )
  {
    p.resize( ump.size() );
    for ( int i = 0; i < ump.size(); i++ ) convert(p[i],ump[i],v);
  };
  
  /// convert mp in C[$...,x_a,...,x_b,...$] into (C[...])[xa,xb]
  template<class Ca, class Ra, 
	   class Rp, 
	   class Cb, class Ob, class Rb >

  void convert( upoldse< upoldse<Ca,Ra>, Rp >& p, 
		const mpoldst<Cb,Ob,Rb> & mp, int a, int b )
  {
      upoldse< mpoldst<Cb,Ob,Rb>, Rb > tmp;
    convert( tmp, mp, a );
    convert( p,  tmp, b );
  };
  
  template<class C, class O, class Ra, class Rb, class Rc>
  void convert( upoldse< upoldse<C,Ra>, Rb >& p, 
		const mpoldst<C,O,Rc> & mp, 
		const char * a, const char * b, 
		const Variables & vars = Variables::default_  )
  {
    convert( p, mp, vars[a], vars[b] );
  };

  template<class C, class O, class Rm, class Ru> inline
  void convert( mpoldst<C,O,Rm> & result, const upoldse< mpoldst<C,O,Rm> , Ru >  & p, int v )  
  {
    int nvar = NbVar(p);
    int tmp[ nvar + 1 ];
//     mpoldst<C,O,Rm> result;
    for ( int i = 0; i < p.size(); i ++ ) 
      for ( typename mpoldst<C,O,Rm>::coeff_t::const_iterator itc = p[i].begin(); itc != p[i].end(); itc++ )
	{
	  typename mpoldst<C,O,Rm>::monom_t tmp = *itc;
	  if ( tmp.rep.size() <= v ) tmp.rep().resize( v+1 );
	  tmp.rep()[v] = i;
	  result += tmp;
	};
  };

} ; // namespace let

} ; // namespace mmx

# endif
