#include <iostream>
#include <list>

#include <subdivix/mpoldst_xml.hpp>
#include <subdivix/solver_mvbst.hpp>


template<class MPOL,class MTH>
void run_solver(int argc, char** argv, const MTH & mth) 
{
  using namespace mmx;
  double total=0,start,end;

  using std::cout;
  using std::endl;

  int N=0;
  bool verbose=false;
  for(int i=0;i<argc;i++)
    if (std::string(argv[i]) == std::string("-h"))
      {
	std::cout<<"Usage: "<<argv[0]<<" file [-h|-v]"<<std::endl;
	return;
      }
    else if (std::string(argv[i]) == std::string("-v"))
      verbose=true;
    else if (std::string(argv[i]) == std::string("-n"))
      N = atoi(argv[i+1]);

  std::ifstream in(argv[1],std::ios::in);
  if(!in.good())
    {
      std::cout<<"file: "<<argv[1]<<std::endl;
      return;
    }
  std::string s;

  std::ifstream in0(argv[1],std::ios::in);
  std::string dom = xml::read_tag(in0,"<domain>");
  
  std::ifstream in1(argv[1],std::ios::in);
  int dim = atoi(xml::read_tag(in0,"<dimension>").data());
  xml::print(std::cout,dim,"dimension");
  std::cout<<std::endl;

  typedef MPOL mpol_t;
  std::list<mpol_t>  L;
  mpol_t p;
  int nbp=0,deg=0,bzb=1, nbv =0;
  while(xml::read(in,p) && (N==0 || nbp < N)) {
    L.push_back(p);
    deg= std::max(deg,degree(p));
    nbv= std::max(nbv,nbvars(p));
    bzb*=degree(p);
    nbp++;
  }
  if(verbose)
    for(typename std::list<mpol_t>::const_iterator it = L.begin(); it != L.end(); ++it)
      {
	xml::print(std::cout,*it,"polynomial")<<std::endl;
      }
  xml::print(std::cout,nbp,"nb_pol");
  xml::print(std::cout,nbv,"nb_var");
  std::cout<<std::endl;
  xml::print(std::cout,deg,"max_deg");
  xml::print(std::cout,bzb,"bezout_bnd");
  std::cout<<std::endl;

  std::vector<double> dmn(2*nbv);
  for(int i=0; i<2*nbv; i+=2)
    {
      dmn[i]=0; dmn[i+1]=1;
    }
  std::istringstream ins(dom.data());
  for(int i=0; i<2*nbv && ins.good();i++) ins >> dmn[i];

  std::cout<<"<domain>";
  for(int i=0; i<2*nbv;i++) std::cout<<dmn[i]<<" ";
  std::cout<<"</domain>"<<std::endl;


  start = clock();
  Seq< std::vector<double> > sol = solve(L,mth,dmn);
  end = clock();
  if(verbose)
    xml::print(std::cout, sol,"solution")<<std::endl;
  int nr=sol.size();

  total += (end - start);
  
  xml::print(std::cout,nr,"nb_root");
  cout << "<time>" << (total / CLOCKS_PER_SEC) << "</time> ";
  if(!verbose)
    cout << "<speed>" << int(nr/(total / CLOCKS_PER_SEC)) << "</speed>";
  std::cout<<std::endl;
}

template<class MPOL, class MTH>
void run_rootof(int argc, char** argv, const MTH & mth) 
{
  using namespace mmx;
  double total=0,start,end;

  using ::std::cout;
  using ::std::endl;

  bool verbose=false;
  int iters=0, deg=0, N=0;
  for(unsigned i=0;i<argc;i++)
    if (std::string(argv[i]) == std::string("-h"))
      {
	std::cout<<"Usage: "<<argv[0]<<" file [-h|-v]"<<std::endl;
	std::cout<<"\t -h  display this message"<<std::endl;
	std::cout<<"\t -v  verbose mode"<<std::endl;
	return;
      }
    else if (std::string(argv[i]) == std::string("-v"))
      verbose=true;
    else if (std::string(argv[i]) == std::string("-n"))
      N = atoi(argv[i+1]);

  ::std::ifstream in(argv[1],std::ios::in);
  ::std::string s;

  ::std::ifstream in0(argv[1],std::ios::in);
  ::std::string d0 = xml::read_tag(in,"domain");
  ::std::cout <<"---> "<<d0<<std::endl;

  //  xml::read(in,s,"name");
  //  xml::print(std::cout,s,"name")<<std::endl;
  typedef MPOL mpol_t;
  std::list<mpol_t>  L;
  mpol_t p;
  int nbp=0;

  while(mmx::xml::read(in,p) && (N==0 || nbp < N)) {
    L.push_back(p);
    deg= std::max(deg,Degree(p));    
    nbp++;
  }

 if(verbose)
   for(typename std::list<mpol_t>::const_iterator it = L.begin(); it != L.end(); ++it)
     {
       xml::print(std::cout,*it,"polynomial")<<std::endl;
     }

  cout << "<nb_pol>" <<  nbp << "</nb_pol>";
  cout << "<degree>" <<  deg << "</degree>";

  start = clock();
  solve(L,mth);
  end = clock();
  total += (end - start);
  
  if(verbose)
    {
      xml::print(std::cout,solve(L,mth),"solution")<<std::endl;
      cout << "<time>" << (total / CLOCKS_PER_SEC) << "</time> ";
    }
}
