/*******************************************************************
 *   This file is part of the source code of the SYNAPS kernel.
 *   Author(s): B. Mourrain, GALAAD, INRIA
 ********************************************************************/
#ifndef subdivix_solver_hpp
#define subdivix_solver_hpp
//====================================================================
namespace mmx {

#define solver_default_precision 20

  template<class real>  struct Homography
  {
    real a,b,c,d;
    Homography(): a(1),b(0),c(0),d(1) {}
    Homography(const Homography& H): a(H.a),b(H.b),c(H.c),d(H.d) {}
    Homography(const real& A, const real& B, const real& C, const real&D): a(A),b(B),c(C),d(D) {}
  };

  struct Isolate     
  { 
    static const bool isole=true; 
  };
  struct Approximate 
  { 
    static const bool isole=false; 
    double eps;
    Approximate(): eps(1e-6) {}
    Approximate(double e): eps(e) {}
  };

  template<class P> struct solver_helper 
  {
    typedef P                       solver_t;
    typedef typename solver_t::solutions_t solutions_t;
  };


  template<class POL, class SLV>
  typename SLV::solutions_t solve(const POL& p, const SLV& s)
  {
    typedef typename SLV::solver_t solver_t;
    typename SLV::solutions_t sol;
    solver_t::solve(sol,p);
    return sol;
  }
  
  template<class POL, class SLV>
  typename SLV::solutions_t solve(const POL& p, SLV& s)
  {
    typedef typename SLV::solver_t solver_t;
    typename SLV::solutions_t sol;
    solver_t::solve(sol,p);
    return sol;
  }
  

  template<class POL,class SLV,class Domain>
  typename SLV::solutions_t solve(const POL& p, const SLV& s, const Domain& D)
  {
    typedef typename SLV::solver_t  solver_t;
    typename SLV::solutions_t sol;
    solver_t::solve(sol,p,D);
    return sol;
  }

}
//====================================================================
#endif  //subdivix_solver_uvp_hpp
