/********************************************************************
 *   This file is part of the source code of the SYNAPS library.
 *   Author(s): J.P. Pavone, GALAAD, INRIA
 *   $Id: subdivix.h,v 1.1 2005/07/11 10:03:51 jppavone Exp $
 ********************************************************************/
/** \file  subdivix.h                                                 **/
#ifndef SYNAPS_SOLVE_mslvbst_H
#define SYNAPS_SOLVE_mslvbst_H
/********************************************************************/
#include <set>
#include <vector>
#include <math.h>
#include <subdivix/solver.hpp>
#define REP(x) x
namespace mmx {

template<class C>
std::ostream & operator<<(std::ostream & os, const std::vector<C> & V)
{
  return VECTOR::print(os,V);
}

}


#include <subdivix/mpoldst.hpp>
#include <subdivix/method.hpp>
#include <subdivix/support.hpp>
#include <subdivix/mpoldse.hpp>
#include <subdivix/Seq.hpp>
//====================================================================
namespace mmx {
//====================================================================
namespace subdivix
{
#define SBDSLV_RD       0
#define SBDSLV_RDL      1
#define SBDSLV_RDS      2
#define SBDSLV_RDRDL    3
#define SBDSLV_RDRDLRDS 4
#define SBDSLV_RDRDS    5
#define SBDSLV_RDLRDS   6
#define SBDSLV_RDSRDL   7
#ifndef SBDSLV_MTH
#define SBDSLV_MTH SBDSLV_RDRDL
#endif
  
  template< class system, int mth > /* selection de methodes */
  struct select_mth
  { typedef subdivix::method<system,strgy::newton<system>,rdslv::parallel,sbdrl::parametric> result_t; };
  template< class system > 
  struct select_mth<system,SBDSLV_RD>
  { typedef subdivix::method<system,strgy::simple<system>,rdslv::parallel,sbdrl::parametric> result_t; };
  template< class system > 
  struct select_mth<system,SBDSLV_RDL>
  { typedef subdivix::method<system,strgy::newton<system>,rdslv::parallel,sbdrl::parametric> result_t; };
  template< class system>
  struct select_mth<system,SBDSLV_RDRDL>
  { typedef subdivix::method<system,strgy::newton< system, strgy::simple<system> >,rdslv::parallel,sbdrl::parametric> 
    result_t; };
  template< class system >
  struct select_mth<system,SBDSLV_RDRDLRDS>
  { typedef subdivix::method<system,strgy::cvmatrix< system, strgy::newton<system,strgy::simple<system> > >,rdslv::parallel,sbdrl::parametric> result_t; };
  template< class system > 
    struct select_mth<system,SBDSLV_RDS>
  { typedef subdivix::method<system,strgy::cvmatrix<system>,rdslv::parallel,sbdrl::parametric> result_t; };

  template< class system > 
    struct select_mth<system,SBDSLV_RDRDS>
  { typedef subdivix::method<system,strgy::cvmatrix<system, strgy::simple<system> >,rdslv::parallel,sbdrl::parametric> result_t; };
  
  template< class system > 
    struct select_mth<system,SBDSLV_RDLRDS>
  { typedef subdivix::method<system,strgy::cvmatrix<system, strgy::newton<system> >,rdslv::parallel,sbdrl::parametric> result_t; };

  template< class system > 
    struct select_mth<system,SBDSLV_RDSRDL>
  { typedef subdivix::method<system,strgy::cvmatrix<system, strgy::newton<system> >,rdslv::parallel,sbdrl::parametric> result_t; };
  
  template< class T, class eenv_t, class MPOL >
  void fill_data( T * data, eenv_t * env, const MPOL& p, binomials<T>& benv  )
  {
    typename MPOL::const_iterator mn; 
    std::fill(data,data+env->sz(),0); 
    mpolfill(data,p,*env);
    convertm2b (data,*env,benv);
  };
  template< class monom >
  void scan_monom_env ( std::set< int >& env, const monom& m )
  { for ( unsigned i = 0; i < (m.rep()).size(); i++ ) if ( (m.rep())[i] != 0 ) env.insert(i); };
}

//====================================================================
/** Class which specifies a subdivision solver for multivariate polynomials.
   The class @T@ is the type of coefficients for the polynomials.
   The type used to store the domain is @DOMAIN@.
   The preconditioning and reduction and subdivision strategy is let 
   by the integer @mth@.
   \param SBDSLV_RD is performing simple reductions.
   \param SBDSLV_RDL is performing reductions and a local preconditioning.
   \param SBDSLV_RDS is performing reductions and a global preconditioning.
   \param SBDSLV_RDRDL is performing two steps of reductions and a local preconditioning. 
   \param SBDSLV_RDRDLRDS is performing two steps of reductions and a local preconditioning and again a reduction.
**/
template<class C, int MTH>
struct solver_multivariate_bernstein
{
  typedef C coeff_t;
  typedef subdivix::system<C> system_t;
  
  static C eps;
  static system_t* sys;

  /// Construction of the solver for the precision \c precision.
  // solver_multivariate_bernstein( const C& precision = (C)1e-6 ): eps(precision), sys(0){};

  
  static void set_precision(int prec) { using std::pow; eps = pow(2.0,-prec); }

  template< class MPDI >  static
  void init_bernstein( MPDI  first, MPDI last)
  {

    int    ne = 0;
    MPDI eptr;
    for ( eptr = first; eptr != last; eptr ++ ) ne++;
    int nvrs[ne];
    const int *szs[ne];
    int e;
    //    int mxvr = 0;
    std::set<int> gvrs;
    for ( e = 0, eptr = first; eptr != last; eptr ++, e++  ) { 
      nvrs[e] = eptr->nvr(); 
      szs[e]  = eptr->szs();
      const int * tmp = eptr->vrs();
      for ( int v = 0; v < eptr->nvr(); v ++ ) gvrs.insert(tmp[v]);
    };
    int mxv = *(gvrs.rend());
    int lnames[ mxv+1 ];
    int nv = 0;
    for ( std::set<int>::const_iterator it = gvrs.begin(); it != gvrs.end(); lnames[*it++] = nv++ );
    int * vrdat = new int[ nv * ne ];
    int * vrptr = vrdat;
    const int *  vrs[ne];
    e = 0;
    for ( eptr = first; eptr != last; vrptr+= eptr->nvr(), e++, eptr ++ ) {
      vrs[e] = vrptr;
      const int * tmp = eptr->vrs();
      for ( int v = 0; v < eptr->nvr();  v ++ )
	vrptr[v] = lnames[ tmp[v] ];	
    };
    //    std::cout << ne << " " << nv << std::endl;
    assert(ne>=nv);
    sys = new subdivix::system<C>( ne, nv, nvrs, (const int**)vrs, (const int**)szs, eps );
    e=0; for ( MPDI eptr = first; eptr != last; add( REP(*eptr), e++  ), eptr++ );
    delete[] vrdat;
  };
  
  static void add( const multivariate_dense::bernstein<C>& data, int e )
  {
    sys->set_bernstein2( data.begin(), e );
  };
  
  void add( const multivariate_dense::monomial<C>& data, int e ) 
  {
    sys->set_monomial2( data.begin(), e );
  };
  
  static int run( std::vector<C> & R, int& nv )
  {
    R.resize(0);
    typename subdivix::select_mth<system_t,MTH>::result_t __mth;
    //int e = 0;for ( MPDI eptr = first; eptr != last; add( eptr->rep(), e++  ), eptr++ );
    //std::vector<C> tmp;
    subdivix::system_ctrl< std::vector< C > >  intf(R);
    int nvars = sys->nvr();
    __mth.launch(sys,intf,0);

    int nsols = R.size() / (2*nvars);
    int nsols2 = subdivix::clean_result( &R[0], nvars, nsols,(C)1e-3 );
    R.resize(nsols2 * nvars) ;
    int ns = nsols2;
    nv = nvars;
    return ns;
  };
  
  static int run( std::vector<C> & sol, C* dom, int& nv )
  {
    sol.resize(0);
    typename subdivix::select_mth<system_t,MTH>::result_t __mth;
    subdivix::system_ctrl< std::vector< C > >  intf(sol);
    int nvars = sys->nvr();
    //    C* dmns = &dom[0];
    // if ( (int)dom.size() >= 2*nvars )
//       {
// 	dmns = new C[ 2*nvars ];
// 	for ( int i = 0; i < 2*nvars; i ++ ) let::assign(dmns[i],dom[i]);
//       };
    __mth.launch(sys,intf,dom);

    int nsols = sol.size() / (2*nvars);
    int nsols2 = subdivix::clean_result( &sol[0], nvars, nsols,(C)1e-3 );
    sol.resize(nsols2 * nvars) ;
    int ns = nsols2;
    nv = nvars;
    return ns;
  };






  //--------------------------------------------------------------------
  /** This is the general interface function for solving multivariate 
      polynomial systems. 
      \param l is a sequence of polynomials (with @begin()@ and @end()@ member functions)
      \param mth specifies the strategies and the domain for the resolution.
      \param verbose set to @true@ output complete information on the resolution.
  **/
  template< class output, class MPC, class Domains >
  static void solve( output& res, const MPC& l, const Domains & dom, 
		     bool verbose = false )
  {
    typedef C T;
    typedef typename MPC::value_type          MPoly;
    typedef MPoly                             mpoly_t;
    typedef MPC                               plist_t;
    typedef typename plist_t::const_iterator  pptr_t;
    typedef typename MPoly::const_iterator    mptr_t;
    typedef typename MPoly::monom_t           monom_t;
    typedef solver_multivariate_bernstein<C, MTH>           mth_t;
    typedef int                               sz_t;
    sz_t  nvars = 0, neqs, * nvrs, ** vrs, ** szs; 
    int e;
    pptr_t cpl;

    // recuperation de l'environement global
    for ( neqs = 0, cpl = l.begin(); cpl != l.end(); cpl++, neqs++  )
      for ( mptr_t mn = cpl->begin(); mn != cpl->end(); mn++ )
	nvars = std::max(nvars,mn->nvars()+1);
    // recuperation de l'environement de chaque equation
    std::set<int> * seenvs = new std::set< int > [ neqs ];
    for ( e = 0, cpl = l.begin(); e < neqs; e++, cpl++  )
      for ( mptr_t mn = cpl->begin(); mn != cpl->end(); mn++ ) 
	subdivix::scan_monom_env(seenvs[e],*mn);
    
    //  assert(sizeof(sz_t)==sizeof(sz_t*));
    sz_t  s = 0;
    for ( e = 0; e < neqs; s += seenvs[e].size(), e++ );
    nvrs = new sz_t [neqs+2*s];
    vrs  = new sz_t*[2*neqs];
    szs  = vrs + neqs;
    sz_t * base = nvrs + neqs;
    std::set<int>::iterator seenvit;  
    for ( e = 0, cpl = l.begin(); e < neqs; base += 2*nvrs[e++], cpl++ )
      {
	nvrs[e] = seenvs[e].size();
	vrs [e] = base;
	szs [e] = base + nvrs[e];
	int k = 0;
	for ( seenvit = seenvs[e].begin(); k < nvrs[e]; seenvit++, k++  ) 
	{
	  vrs[e][k]  = *seenvit;
	  szs[e][k]  = MPOLDST::degree( *cpl, *seenvit ) + 1;
	};
      };
    
    std::vector<C> result;
    C * dmns = 0;
    if ( (sz_t)dom.size() >= 2*nvars )
      {
	dmns = new C[ 2*nvars ];
	for ( int i = 0; i < 2*nvars; i ++ ) let::assign(dmns[i],dom[i]);
      };
    
    //  chrono tm;
    std::vector<T> tmp;
    
    {  /* remplissage et appel */
      using namespace subdivix;
      binomials<T> bznv;
      typedef  subdivix::system<T> system_t;
      system_t sys( neqs, nvars, nvrs, (const int**)vrs, (const int**)szs, (T)eps);
      for ( e = 0, cpl = l.begin(); e < neqs; e ++, cpl++  )
	subdivix::fill_data( sys.data(e), sys.env(e), *cpl, bznv );
      subdivix::system_ctrl< std::vector< T > >  intf(tmp);
      //    tm.start();
      typename subdivix::select_mth<system_t,MTH>::result_t __mth;
      __mth.launch(&sys,intf,dmns);
      //tm.stop();
      if ( verbose )
	{
	  std::cout << " iterations  = " << __mth.m_niter << std::endl;
	  std::cout << "subdivisions = "<< __mth.m_nsbd  << std::endl;
	};
    };
    
    if ( !verbose && dmns ) delete[] dmns;
    delete[] seenvs;
    delete[] nvrs;
    delete[] vrs;
    if ( verbose )
      {
	std::cout << "<domain>";
	if  ( dmns )
	  {
	    std::cout << std::endl;
	    for ( int i = 0; i < nvars; i ++ )
	      std::cout << "\t" 
			<< "x" << i//monom_t::var_of_index(i) 
			<< " = [ " << dmns[2*i] 
			<< "," 
			<< dmns[2*i+1] 
			<< " ]";
	  }
	else std::cout << "[0,1]^" << nvars;
	std::cout << "</domain>\n";
	//	std::cout << "\treductions   = " << bsys.nrest() << std::endl;
	//	std::cout << "\titerations   = " << bsys.niter() << std::endl;
	//      std::cout << "<solver_time>"; tm.write(std::cout); 
	//      std::cout <<"</solver_time>"<< std::endl;	
	if ( dmns ) delete[] dmns;
      };
    
    unsigned nsols = tmp.size() / (2*nvars);
    T boxes[nsols*2*nvars];
    std::copy(&tmp[0], &tmp[0]+2*nvars*nsols,boxes);
    int nsols2 = subdivix::clean_result( &tmp[0], nvars, nsols,(T)1e-3 );
    if ( verbose ) { 
      //    xml::print(std::cout,nsols,"nb_box")<< std::endl;
      for ( unsigned i = 0; i < nsols; i ++ )
	{
	  std::cout << "\t[ ";
	  for (int j = 0; j < nvars; j ++ )
	    std::cout << (boxes[2*(i*nvars+j)]+boxes[2*(i*nvars+j)+1])/2.0 << " ";
	  std::cout << "]";
	  std::cout << std::endl;
	};
      //    xml::print(std::cout,nsols2,"nb_sol")<< std::endl; 
      //    if ( nsols ) vctops::print(&tmp[0],nvars*nsols );
    };
    
    
    tmp.resize( nsols2 * nvars );
    //  cout <<"-----"<<endl;
    std::vector<T> vsol(nvars);
    //    Seq<vectdse<T> > res;
    for(int i=0;i<nsols2;++i)
      {
	for(int j=0; j<nvars; j++)
	  vsol[j]=tmp[i*nvars+j];
	res.push_back(vsol);
      }
    //    return res;
  };

  template< class output, class MPC>
  static void solve_bernstein(output& sol, const MPC& l)
  {
     init_bernstein(l.begin(), l.end());
     std::vector<C> res; int ns, nv; 
     ns=run(sol,nv);
     //     std::cout<<"system: nv="<<nv<< " ns="<< ns<<std::endl;

  }


  template< class output, class MPL, class DOM>
  static void solve_bernstein(output& sol, const MPL& eqs, const DOM& dom)
  {
     init_bernstein(eqs.begin(), eqs.end());
     int nvars = sys->nvr();
     int ns, nv; 
     DOM dmns(dom);
     if((int)dom.size()>=2*nvars)
       ns=run(sol,&dmns[0],nv);
     else
       ns=run(sol,nv);

     //     std::cout<<"system: nv="<<nv<< " ns="<< ns<<std::endl;

  }
  //  ~solver_multivariate_bernstein() { delete sys; };

};

template<class C, int M> C solver_multivariate_bernstein<C,M>::eps = std::pow(2.0,-solver_default_precision);
template<class C, int M> subdivix::system<C>* solver_multivariate_bernstein<C,M>::sys = 0;

//====================================================================
template<class C,int MTH= SBDSLV_RDRDL> struct MVBST 
{
  static void set_precision(unsigned prec) 
  {
    solver_multivariate_bernstein<C,MTH>::set_precision(prec);
  }
};
//====================================================================
template<class C, int MTH> struct solver_helper< MVBST<C,MTH> >
{
  //    typedef Seq<Interval<typename K::rational> > solutions_t;
  typedef solver_multivariate_bernstein<C,MTH> solver_t;
};
//====================================================================
template< class MPC, class Domains,  class C, int MTH>
Seq<std::vector<C> > solve( const MPC& l, MVBST<C,MTH> slv, const Domains& D, 
			    bool verbose = false  )
{
  Seq<std::vector<C> > sol; 
  solver_multivariate_bernstein<C,MTH>::solve(sol,l,D,verbose);
  return sol;
}
//====================================================================
template< class MPC, class Domains,  class C, int MTH>
Seq<std::vector<C> > solve_bernstein( const MPC& l, MVBST<C,MTH> slv, const Domains& D, 
				      bool verbose = false  )
{
  Seq<std::vector<C> > sol; 
  solver_multivariate_bernstein<C,MTH>::solve_bernstein(sol,l,D,verbose);
  return sol;
}
//====================================================================
} //namespace mmx
/********************************************************************/
#endif //SYNAPS_SOLVE_mslvbst_H
