/*******************************************************************
 *   This file is part of the source code of the SYNAPS kernel.
 *   Author(s): J.P. Pavone GALAAD, INRIA
 *              B. Mourrain GALAAD, INRIA
 ********************************************************************/
#ifndef subdivix_solver_sleeve_hpp
#define subdivix_solver_sleeve_hpp
/********************************************************************/
#include <subdivix/solver.hpp>
#include <subdivix/IEEE754.hpp>
#include <subdivix/solver_binary.hpp>
//--------------------------------------------------------------------
namespace mmx {
  //====================================================================  
  template<class K>
  struct binary_isolate: public K 
  {
    static data_t data;

    template<class output, class real> static inline
    void get(output& sol, const real& first = 0, const real& last =1)
    {
      typedef typename output::value_type root_t;
      typedef root_t                      as_root;
      //      unsigned s = o.size();
      real a,b;
      //      o.resize(s+2*size());
      real scale = last-first;
      for ( unsigned i = 0; i < data.nb_sol(); i ++ )
	{
	  data_t::convert(a,data.m_res[i].a,first, scale);
	  data_t::convert(b,data.m_res[i].b,first, scale);
	  sol << as<root_t>(Interval<real>(a,b));
	};
    };

    template<class output, class real> static inline
    void get(output & sol, const Homography<real>& H)
    {
      typedef typename output::value_type root_t;
      typedef root_t                      as_root;
      typedef typename root_t::value_type bound_t;
      //      unsigned s = o.size();
      bound_t u,v;
      real dt= H.a*H.d-H.b*H.c;
      if(dt>0)
	for (unsigned i = 0; i < data.nb_sol(); i ++ )
	  {
	    data.convert(u,data.m_res[i].a,H);
	    data.convert(v,data.m_res[i].b,H);
	    sol << as_root(u,v);
	  }
      else
	for ( int i = data.nb_sol()-1; i >=0 ; i --)
	  {
	    data.convert(u,data.m_res[i].b,H);
	    data.convert(v,data.m_res[i].a,H);
	    sol << as_root(u,v);
	  }
    }
  };
  template <class K>  data_t binary_isolate<K>::data(true);
  //====================================================================
  template<class K>
  struct binary_approximate : public K 
  {
    typedef unsigned unsigned_t;
    typedef typename K::integer integer;
    static data_t data;
    
    static inline 
    integer binomial(const integer& n,const integer& p){return K::binomial(n,p);}

    //    template<class C> static inline C as_root(const C& a, const C& b) {return (a+b)/2;}

    template<class output, class real> static inline
    void get(output& sol, const real& first = 0, const real& last =1)
    {
      typedef typename output::value_type root;
      real a,b;
      real scale= last-first;
      for (unsigned i = 0; i < data.nb_sol(); i ++ )
	{
	  data_t::convert(a,data.m_res[i].a, first, scale);
	  data_t::convert(b,data.m_res[i].b, first, scale);
	  sol << as<root>(Interval<real>(a,b));
	}
    } 

    template<class output, class real> static inline
    void get(output& sol, const Homography<real>& H)
    {
      typedef typename output::value_type as_root;
      //root_t u,v;
      unsigned s=data.nb_sol();
      std::vector<real> u(s);
      real v;
      real dt=H.a*H.d-H.b*H.c;

      if(dt>0)
	{
	  {	
	    numerics::rounding<real> rnd(numerics::rnd_dw());
	    for ( unsigned i = 0; i < s; i ++ )
	      {
		data.convert(u[i],data.m_res[i].a,H);
	      }
	  }
	  {
	    numerics::rounding<real> rnd(numerics::rnd_up());
	    for ( unsigned i = 0; i < s; i ++ )
	      {
		data.convert(v,data.m_res[i].b,H);
		sol << as_root(u[i],v);
	      }
	  }
	}
      else
	{
	  {	
	    numerics::rounding<real> rnd(numerics::rnd_dw());
	    for (int i = s-1; i >=0; i -- )
	      data.convert(u[i],data.m_res[i].a,H);
	  }
	  {
	    numerics::rounding<real> rnd(numerics::rnd_up());
	    for (int i = s-1; i >=0; i -- )
	      {
		data.convert(v,data.m_res[i].b,H);
		sol << as_root(v,u[i]);
	      }
	  }
	}


//       if(dt>0)
// 	for (unsigned i = 0; i < data.nb_sol(); i ++ )
// 	  {
// 	    data.convert(u,data.m_res[i].a,H);
// 	    data.convert(v,data.m_res[i].b,H);
// 	    u+=v;u/=2;
// 	    sol << u;
// 	  }
//       else
// 	for (int i = data.nb_sol()-1; i>=0; i-- )
// 	  {
// 	    data.convert(u,data.m_res[i].a,H);
// 	    data.convert(v,data.m_res[i].b,H);
// 	    u+=v;u/=2;
// 	    sol << u;
// 	  }
    }

  };

  template <class K>  data_t binary_approximate<K>::data(false);

  //====================================================================
  /** Solver class using sleeve approximation with coefficients of type \c C. **/
  template < class K>
  struct binary_sleeve_subdivision : public K
  {

    typedef typename K::integer  integer;
    typedef typename K::rational rational;
    //    typedef typename K::floating floating;
    //    typedef typename K::ieee     ieee;

    typedef double                      creal_t;
    typedef unsigned                    sz_t; 
    typedef binary_subdivision<K>       Base_t;
    typedef typename Base_t::unsigned_t unsigned_t;
    typedef res_t                       Domain_t;
  
    //    sleeve(): Base_t(1e-6){}
    //    sleeve(creal_t e): Base_t(e){}

    static inline void alloc( sz_t s, sz_t deep  ) 
    {
      K::data.alloc(s,2*s,deep); 
    };
    
    static void barre( char c, unsigned n )
    {
      char  _bar[ n+1 ];
      std::fill(_bar,_bar+n, c);
      _bar[n] = 0;
      std::cout << _bar << std::endl;
    };
    
    static void writebounds( creal_t * pup, creal_t * pdw, unsigned s )
    {
      Base_t::print(pup,s); std::cout << std::endl;
      Base_t::print(pdw,s); std::cout << std::endl;
    };
   
    static inline 
    bool glue( sz_t cup, sz_t cdw, int d )
    {
      if ( K::data.bckb() == K::data.m_dmn[d] )
	{
	  K::data.bckb() = K::data.m_dmn[d];
	  K::data.bckb().m += 1;
	  if ( cup ) K::data.m_cup = cup;
	  if ( cdw ) K::data.m_cdw = cdw;
	  return true;
	};
      return false;
    };
    
    static inline 
    void mstore( sz_t cup, sz_t cdw, int d  ) 
    {
      if ( K::data.m_res.size() == 0 || !glue(cup,cdw,d)  )
	{
	  K::data.mstore(d);
	  K::data.m_cup = cup;
	  K::data.m_cdw = cdw;
	  //std::cout << "Push ";
	};
      //      writebck();
    };
  

    static inline
    void dwsplit( creal_t * r, creal_t * l, sz_t s )
    {
      numerics::rounding<creal_t> rnd( numerics::rnd_dw() );
      Base_t::split(r,l,s);
    };

    static inline 
    void upsplit( creal_t * r, creal_t * l, sz_t s )
    {
      numerics::rounding<creal_t> rnd( numerics::rnd_up() );
      Base_t::split(r,l,s);
    };
    
    static void Loop( bool isole = true )
    {
      //      std::cout << "LOOP\n";
      numerics::rounding<creal_t> rnd(numerics::rnd_dw());
      creal_t * pup, * pdw;
      int d;
      sz_t a,s,cup,cdw;
      unsigned_t v;
      s = K::data.m_s;
      pup = K::data.m_data;
      pdw = K::data.m_data + s;
      K::data.m_dmn[0].m = 0;
      K::data.m_dmn[0].e = 0;
      for( a = 0, d = 0; d >= 0; d--, a -= 2*s   ) 
	{
	  cup = Base_t::sgncnt(pup+a,s);
	  cdw = Base_t::sgncnt(pdw+a,s);
	  if ( cup || cdw )
	    {
	      if ( isole && cup == cdw && cup == 1 ){ 
 		//std::cout << " (b) \n";
		K::data.sstore(d);
		continue; 
	      };
	      if ( K::data.m_dmn[d].e == K::data.m_limit-1 ) 
		{ 
   		  // std::cout << " (c) "<<cup<<" "<<cdw<<"\n";
 		  if ( cup == cdw && cup == 1 )
 		    K::data.sstore(d);
		  else
		    mstore(cup,cdw,d); 
		  continue; 
		};
	      dwsplit(pdw+a,pdw+a+2*s,s);    
	      upsplit(pup+a,pup+a+2*s,s);
	      Base_t::split(K::data.m_dmn[d],K::data.m_dmn[d+1]);
	      d += 2;
	      a += 4*s;
	      continue;
	    }
	  else
	    {
	      //std::cout<<"  (a) ";
	      sz_t k =0, sv=0;
	      for (  k = 0; k < K::data.m_s; k ++ )
		{
		  if ( ( pup[a+k] > 0 ) != ( pdw[a+k] > 0 ) )
		    {
		      sv++;
		      if(sv>1)
			{ 
			  mstore(cup,cdw,d); 
			  break; 
			};
		    }
		}
	      //std::cout<<std::endl;
	    };
	};
      // std::cout << "END LOOP\n";
    };
    

    template<class input> static 
    void run_loop( const input& in, const creal_t& eps, const meta::true_t&   )
    {
      //      binary_subdivision<K> slv(eps);
      sz_t prec = numerics::bitprec(eps);
      alloc(in.size(),prec);
      std::copy(in.begin(),in.end(),K::data.m_data);
      std::copy(in.begin(),in.end(),K::data.m_data+K::data.m_s);
      Loop(K::data.isole);
    };

    const Domain_t& operator[]( int i ) const  { return K::data.m_res[i]; };
    Domain_t&       operator[]( int i ) { return K::data.m_res[i]; };
    
  
    template<class output, class input> static inline
    void solve_bernstein( output& out, const input& in)
    {
      run(in,K::data.eps);
      //      solutions<typename output::value_type>::get(this,out);
    }

    template<class VECT,class POL> static 
    void init_pol(VECT& ubp, VECT& dbp, const POL& r,  unsigned sz, bool neg = false); 
    
    template<class VECT,class POL,class Q>
    void init_pol(VECT& ubp, VECT& dbp, const POL& r,  unsigned sz, const Q& u, const Q&v);


    template<class input> static inline
    void run( const input& in, const creal_t& eps)
    { 
      typedef typename numerics::is_rounded<creal_t>::result_t round_t;
      run_loop(in,eps,round_t()); 
    }
    
    template<class input> static 
    void run( const input& up, const input& dw)
    {
      assert(up.size()==dw.size());
      sz_t prec = numerics::bitprec(K::data.eps);
      alloc(up.size(),prec);
      std::copy(up.begin(),up.end(),K::data.m_data);
      std::copy(dw.begin(),dw.end(),K::data.m_data+K::data.m_s);
      Loop(K::data.isole);
    }
    
    template <typename output, typename POL> static 
    void solve_positive(output& sol, const POL& r, bool pos=true);

    //--------------------------------------------------------------------
    /** Compute the roots of the polynomial \c r expressed in the monomial basis. If
	If \c MTH=Isolate, the ordered sequence of isolating intervals is stored in \c sol. 
	If \c MTH=Approx, the ordered sequence of approximation of the roots is stored in \c sol.
    **/ 
    template <typename output, typename POL> static 
    void solve(output& sol, const POL& r) 
    {
      //      this->root_bound = - bound_root(r,Cauchy<C>());
      solve_positive(sol,r,false);
      //      this->root_bound = - this->root_bound;
      solve_positive(sol,r,true);
    }

    /** Compute the roots of the polynomial \c r expressed in the monomial
	basis, in the interval [u,v].
	If \c MTH=Isolate, the ordered sequence of isolating intervals is stored in \c sol. 
	If \c MTH=Approx, the ordered sequence of approximation of the roots is stored in \c sol.
    **/ 
    template <typename output, typename POL, typename Q> static 
    void solve(output& sol, const POL& r, const Q& u, const Q& v);

    //--------------------------------------------------------------------
    template<class output, class input,class real,class MTH> static inline
    void solve_bernstein( output& sol, 
			  const input& up, const input& dw, 
			  const real& u, const real& v, const MTH& mth)
    { 
      K::data.isole=mth.isole;
      run(up,dw);
      get(sol,u,v);
      //      solutions<typename output::value_type>::get(this,out,u,v);
    }
  
    //--------------------------------------------------------------------
    /** Compute the roots of the polynomial \c r expressed in the bernstein
	basis, on the interval [u,v].
	If \c MTH=Isolate, the ordered sequence of isolating intervals is stored in \c sol. 
	If \c MTH=Approx, the ordered sequence of approximation of the roots is stored in \c sol.
    **/ 
    template <typename output, typename POL ,typename real, typename MTH> static 
    void solve_bernstein(output& sol, const POL& r, const real& u,const real& v, const MTH& mth )
    {
      typedef typename output::value_type    root_t;    
      using let::assign;
      
      unsigned sz= r.size();
      std::vector<creal_t> ubp( sz ), dbp( sz );
      {
	numerics::rounding<creal_t> rnd( numerics::rnd_up() );
	for(unsigned i=0;i<sz;i++) ubp[i] =as<creal_t>(r[i]);
      }
      {
	numerics::rounding<creal_t> rnd( numerics::rnd_dw() );
	for(unsigned i=0;i<sz;i++) dbp[i] =as<creal_t>(r[i]);
      }
      K::data.isole=mth.isole; 
      run(ubp,dbp);
      get(sol,u,v);
      //      solutions<typename output::value_type>::get(this, sol, u,v);
    }
  };
  //====================================================================
  template<class K>
  template<class VECT,class POL>
  void binary_sleeve_subdivision<K>::init_pol(VECT& ubp, VECT& dbp, const POL& p, unsigned sz, bool pos)
  {
    typedef typename K::integer  integer;
    typedef typename K::rational rational;

    if(pos)
      K::data.root_bound = bound_root(p,Cauchy<creal_t>());
    else
      K::data.root_bound = - bound_root(p,Cauchy<creal_t>());

    //    std::cout <<"Root bound: "<< this->root_bound <<std::endl;

    using let::assign;
    
    rational mx = VECTOR::max_abs(p);
    if( mx == 0 ) return;
 
    std::vector<rational> bp( sz );
    for ( unsigned i = 0; i < sz; i ++ )
      {
	bp[i] = as<double>(p[i])/as<double>(mx*binomial<integer>(sz-1,i));
	//	Interval<double> I=as_interval_double(p[i],mx*binomial<integer>(sz-1,i));
	//	std::cout<<I<<std::endl;
      }
 
    if(!pos) for(unsigned i=1;i<sz;i+=2) bp[i]=-bp[i];

    {
      numerics::rounding<creal_t> rnd( numerics::rnd_up() );
      for ( unsigned i = 0; i < sz; i ++ )
	  ubp[i]=as<creal_t>(bp[i]);
    }
    {
      numerics::rounding<creal_t> rnd( numerics::rnd_dw() );
      for ( unsigned i = 0; i < sz; i ++ )
	  dbp[i]=as<creal_t>(bp[i]);
    }

  }
  //--------------------------------------------------------------------
  template<class K>
  template<class VECT,class POL,class Q>
  void binary_sleeve_subdivision<K>::init_pol(VECT& ubp, VECT& dbp, const POL& r,unsigned sz,const Q& u, const Q&v)
  {
    using let::assign;
  
#if 1
    //    PRINT_DEBUG("Conversion using exact arithmetic");
    rational U,V;
    assign(U,u);
    assign(V,v);
    //    std::vector<interval> bp( sz );
    std::vector<rational> bp(sz);
    BEZIER::monomial_to_bezier(bp,r,sz,U,V);
    rational mx = VECTOR::max_abs(bp);
    VECTOR::div(bp,mx);
    // unsigned i;
    //    for (i = 0; i < sz && bp[i].upper()-bp[i].lower()<as<floating>(1.e-16); i ++ );
    //    if(i<sz) std::cout<<"refine: "<<i<<" " <<(bp[i].upper()-bp[i].lower())<<std::endl;
    //    double l; assign(l, rational(V-U));
    //    std::cout<<"\t init: "<<sign_variation(bp)<<" "<<l<<std::endl;
    {
      numerics::rounding<creal_t> rnd( numerics::rnd_up() );
      for ( unsigned i = 0; i < sz; i ++ ) ubp[i]=as<creal_t>(bp[i]);
    }
    {
      numerics::rounding<creal_t> rnd( numerics::rnd_dw() );
      for ( unsigned i = 0; i < sz; i ++ ) dbp[i]=as<creal_t>(bp[i]);
    };
   
#else
    //    PRINT_DEBUG("Using rounding arithmetic");
    O ud, vd;
    {
      numerics::rounding<creal_t> rnd( numerics::rnd_dw() );
      ud=as<creal_t>(u);
    };
    {
      numerics::rounding<creal_t> rnd( numerics::rnd_up() );
      vd=as<creal_t>(v);
      BEZIER::monomial_to_bezier(ubp,r,r.size(),ud,vd);
    };
    {
      numerics::rounding<creal_t> rnd( numerics::rnd_dw() );
      BEZIER::monomial_to_bezier(dbp,r,r.size(),ud,vd);
    };

    for ( unsigned i= 0; i < ubp.size(); i ++ )
      if ( dbp[i] > ubp[i] ) { 
	std::swap(dbp[i], ubp[i]);
      };
#endif

  }
  //------------------------------------------------------------------
  template<class K>
  template <typename output, typename Poly>
  void binary_sleeve_subdivision<K>::solve_positive(output& sol, const Poly& r, bool pos)
  {
    unsigned sz=r.size();
    std::vector<creal_t> ubp(sz), dbp(sz);

    init_pol(ubp,dbp,r,sz,pos);
    run(ubp,dbp);
    Homography<creal_t> H(1,0,-1,1);
    if(!pos) {H.c=1;H.d=-1;}
    K::get(sol,H);
  }
  //--------------------------------------------------------------------
  template <class K>
  template <typename output, typename POL ,typename Q>
  void binary_sleeve_subdivision<K>::solve(output& sol, const POL& r, const Q& u, const Q& v)
  {
    unsigned sz= r.size();
    std::vector<double> ubp(sz), dbp(sz);
    init_pol(ubp,dbp,r,sz,u,v);
    run(ubp,dbp,true);
    //    run(ubp,dbp,this->eps,mth.isolate);
    //    solutions<typename output::value_type>::get(this,sol,u,v);
  }

  //====================================================================
  template<class K>
  struct solver_sleeve_isolate:
    public binary_sleeve_subdivision<binary_isolate<K> > 
  {
    typedef solver_sleeve_isolate<K> solver_t;
    typedef Seq<Interval<typename K::ieee> >  solutions_t;
  };

  template<class K>
  struct solver_sleeve_approximate:
    public binary_sleeve_subdivision<binary_approximate<K> > 
  {
    typedef solver_sleeve_approximate<K> solver_t;
    typedef Seq<Interval<typename K::ieee> >  solutions_t;
  };
  //====================================================================
  template<class M, class K=IEEE754> struct Sleeve{};

  template<class K> struct Sleeve<Isolate,K>
  {
    typedef solver_sleeve_isolate<K>       solver_t;
    typedef typename solver_t::solutions_t solutions_t;
  };

  //--------------------------------------------------------------------
  template<class K> struct Sleeve<Approximate,K>
  {
    typedef solver_sleeve_approximate<K>   solver_t;
    typedef typename solver_t::solutions_t solutions_t;

    static inline void set_precision(unsigned p)
    {
      binary_approximate<K>::set_precision(p);
    }
  };

//====================================================================
} //namespace mmx
#endif  //subdivix_solver_sleeve_hpp
