/*******************************************************************
 *   This file is part of the source code of the SYNAPS kernel.
 *   Author(s): J.P. Pavone GALAAD, INRIA
 *              B. Mourrain GALAAD, INRIA
 ********************************************************************/
#ifndef synaps_solver1d_sleeve_H
#define synaps_solver1d_sleeve_H
/********************************************************************/
#include <subdivix/Seq.hpp>
#include <subdivix/binomial.hpp>
#include <subdivix/upoldse_bound.hpp>
#include <subdivix/rounding_mode.hpp>
#include <subdivix/sign.hpp>
#include <subdivix/bezier_rep1d.hpp>
#include <subdivix/solver.hpp>
#include <subdivix/solver_uvp_bigunsigned.hpp>
#include <subdivix/IEEE754.hpp>
#include <assert.h>
#include <vector>
#include <algorithm>
//====================================================================
namespace mmx {

#ifndef WITH_AS
#define WITH_AS
  template<typename T,typename F>
  struct cast_helper {
    static inline T cv (const F& x) { return x; }
  };
  
  template<typename T,typename F> inline 
  T as (const F& x) {  return cast_helper<T,F>::cv (x); }
#endif

  //  template<class C> inline void operator<<(std::vector<C>& v, const C& x) { v.push_back(x) ; }

  template<class C, int N> struct Interval;
  template<class C> struct WithRounding { typedef meta::false_t T;};
  template<class C, int N> struct WithRounding<Interval<C,N> > { typedef typename WithRounding<C>::T T;};
  template<>               struct WithRounding<double> { typedef meta::true_t T;};


  //--------------------------------------------------------------------
  template<class K> struct Simple {};
  template<class K> struct Sleeve {};

  template<class C, class K = Sleeve<IEEE754> > struct Solver ;

  struct Approx { static const bool isolate= false; };
  struct Isole  { static const bool isolate= true;  };
  //--------------------------------------------------------------------

  template< class C , class K >
  struct Solver<C, Simple<K> >
  {
   
    typedef typename K::integer  integer;
    typedef typename K::rational rational;
    //    typedef typename K::floating floating;
    //    typedef typename K::ieee     ieee;


    typedef unsigned sz_t;
    //  typedef unsigned unsigned_t;
    static const sz_t bitquo = numerics::bit_resolution<C>::nbit / numerics::hdwi<sz_t>::nbit;
    static const sz_t bitrem = numerics::bit_resolution<C>::nbit % numerics::hdwi<sz_t>::nbit;
    typedef BigUnsigned<bitquo+(sz_t)(bitrem!=0)> unsigned_t;
    typedef C creal_t;


    inline
    static void split( creal_t * r, creal_t * l, sz_t sz )
    {
      creal_t * er, * p;
      er = r + (sz-1);
      for ( er = r + (sz-1); er != r; er--, l++ )
	{
	  *l = *r;
	  for ( p = r; p != er; p ++ )
	    *p = (*p+*(p+1))/2;
	};
      *l = *r;
    };
    
    sz_t sgncnt ( creal_t const *  b , sz_t sz )
    {
      creal_t const * last = b + sz;
      sz_t c;
      int prv = (((*b>0)<<1)|(*b<0));
      int crr;
      for ( c = 0; b != last; crr = (((*b>0)<<1)|(*b<0)), c += prv != crr, prv = crr, b ++ );
      return c;
    };
    
    //    struct dmn_t
    //    {
    //      unsigned   h;
    //      unsigned_t v;
    //    };
    
    struct bound
    {
      unsigned  e; 
      unsigned_t m;
      bool operator==( const bound& b )
      {
	if ( e < b.e )
	  {
	    return (m << (b.e-e)) == b.m;
	  };
	return (b.m << (e-b.e)) == m;
      };
    };
    
    inline void split( bound& r, bound& l )
    {
      r.m <<= 1;
      l.m = r.m;
      r.m |=  1;
      r.e ++;
      l.e = r.e;
    };
    
    struct res_t { 
      bound a;
      bound b;
      bool t;  
    };
    
    creal_t  *m_data;
    bound    *m_dmn ;
    sz_t  m_limit;
    sz_t  m_s;
    C eps, root_bound;
    
    std::vector<res_t> m_res;
    
    Solver(): eps(1e-6)        { m_data = 0; };
    Solver(const C& e): eps(e) { m_data = 0; };
    
    ~Solver() { Free(); };
    
    inline 
    void Free() 
    { 
      if ( m_data ) 
	{ 
	  delete[] m_data; 
	  delete[] m_dmn; 
	  m_res.resize(0); 
	}; 
    };
        
    inline 
    void alloc( sz_t s, sz_t cs, sz_t deep  )
    { 
      Free();
      m_s = s;
      m_limit = std::min((sz_t)numerics::hdwi<unsigned_t>::nbit,deep);
      m_data  = new creal_t [ cs*m_limit ];
      m_dmn   = new bound   [ m_limit ];
      m_res.resize(0);
    };
  
    inline 
    void alloc( sz_t s, sz_t deep ) 
    { 
      alloc(s,s,deep); 
    };
  
    inline const 
    bound& bcka() const 
    { 
      return m_res.back().a; 
    };

    inline
    bound& bcka()
    {
      return m_res.back().a;
    };
    
    inline const 
    bound& bckb() const 
    {
      return m_res.back().b;
    };
    
    inline
    bound& bckb()
    {
      return m_res.back().b;
    };
    
    inline 
    bool& bckt() 
    {
      return m_res.back().t;
    };
    
    inline 
    bool bckt() const
    {
      return m_res.back().t;
    };
    
    inline 
    void setbck( const bound& a, const bound& b, bool type )
    {
      bcka() = a;
      bckb() = b;
      bckt() = type;
    };
    
    inline 
    void pshbck()
    {
      m_res.push_back(res_t());
    };
    
    inline 
    void sstore( int d, int t = 1 ) 
    {
      // m_data + d*m_s;
      pshbck();
      bcka() = m_dmn[d];
      bckb() = bcka();
      bckb().m += t;
      bckt() = true;
    };
  
    inline 
    void bstore( int d, int t = 1 ) 
    {
      // m_data + d*m_s;
      pshbck();
      bcka() = m_dmn[d];
      bcka().m += t;
      bckb() = bcka();
      bckt() = true;
    };
   inline 
    void estore( int d, int t = 1 ) 
    {
      // m_data + d*m_s;
      pshbck();
      bckb() = m_dmn[d];
      bckb().m += t;
      bcka() = bckb();
      bckt() = true;
    };
  
    inline 
    void mstore( int d ) 
    {
      sstore(d);
      bckt() = false;
    };
    
    inline
    void print( creal_t * p, sz_t  n )
    {
      std::cout << "[";
      for ( sz_t i = 0; i < n-1; i ++ )
	std::cout << p[i] << ",";
      std::cout << p[n-1] << " ]";
    };

    
    void Loop( bool isole = true )
    {
      //      std::cout <<"Starting simple solver loop"<<std::endl;
      creal_t * pup, * pdw;
      int d;
      sz_t h,a,s,cup,cdw;
      unsigned_t v;
      s = m_s;
      pup = m_data;
      m_dmn[0].m = 0;
      m_dmn[0].e = 0;
      for( a = 0, d = 0; d >= 0; d--, a -= s   ) 
	{
	  cup = sgncnt(pup+a,s);
	  if ( cup ) 
	    {
	      if ( isole && cup == 1 ) { sstore(d); continue; };
	      if ( this->m_dmn[d].e == m_limit-1 )
		{
		  if ( cup == 1 ) 
		    {
		      if ( *(pup+a) != 0 )
			{
			  if ( *(pup+a+s-1) == 0 )
			    sstore(d,0);
			  else 
			    sstore(d);
			};
		    }
		  else  
		    {
		      mstore(d);
		      //std::cout<<"s "<<cup<<" "<<d<<" "<<m_limit<<std::endl;
		    }
		  continue;
		};
	      split(pup+a,pup+a+s,s);
	      split(this->m_dmn[d],this->m_dmn[d+1]);
	      d += 2;
	      a += 2*s;
	    };
	};
      //std::cout <<"End of simple solver loop"<<std::endl;
    };

    template<class real>
    void Convert( real& a, const bound& b, const real & first = 0, const real & scale = 1  )
    {
      unsigned_t m(b.m);
      numerics::hdwi<unsigned_t>::reverse(b.e+1,m);
      a = first;
      real x = scale;
      for ( unsigned i = 0; i < b.e + 1; x /= 2, i ++ )
	{
	  if ( m & 1 ) a += x;
	  m >>= 1;
	};
    };

    template<class real, class coeff>
    void Convert( real& a, const bound& b, const Homography<coeff>& H)
    {
      unsigned_t m(b.m);
      numerics::hdwi<unsigned_t>::reverse(b.e+1,m);
      a = H.b;
      real d = -H.d;
      real x = H.a, y = -H.c;
      for ( unsigned i = 0; i < b.e + 1; x /= 2, y/=2, i ++ )
	{
	  if ( m & 1 ) {a += x; d+=y;}
	  m >>= 1;
	};
      //      std::cout <<"\n\t d= "<<d<<" a= "<<a<< std::endl;
      if(d==0 )
	{
	  //  std::cout <<"\n\t rb= "<<this->root_bound<< std::endl;
	  a = real(this->root_bound);
	}
      else
	{
	  a/= (-d);
	  //	  std::cout <<"\n\t d!=0 a= "<<a<< std::endl;
	}
    };

    sz_t size() const  { return this->m_res.size(); };
    sz_t nb_sol() {return this->m_res.size();}


    //--------------------------------------------------------------------
    template<class output>
    void get_flags( output& o )
    {
      unsigned s = o.size();
      o.resize( s + o.size() );
      for ( sz_t i = 0; i < this->m_res.size(); o[s+i] = this->m_res[i++].t );
    };
    
    //--------------------------------------------------------------------
    template<class input>
    void run_loop( const input& in, const creal_t& eps, bool isole, meta::false_t)
    {
      sz_t prec = numerics::bitprec(eps);
      alloc(in.size(),prec);
      std::copy(in.begin(),in.end(),this->m_data);
      Loop(isole);
    };
    

  };
  //====================================================================
  // Helper for extracting the solutions.
  template<class sol_t, class V = typename WithRounding<sol_t>::T >  struct solutions 
  {
    template<class SLV, class output, class real> static inline
    void get(SLV* slv, output& sol, const real& first = 0, const real& last =1)
    {
      real a,b;
      real scale= last-first;
      for (typename SLV::sz_t i = 0; i < slv->nb_sol(); i ++ )
	{
	  slv->Convert(a,slv->m_res[i].a, first, scale);
	  slv->Convert(b,slv->m_res[i].b, first, scale);
	  sol.push_back((a+b)/2);
	}
    }

    template<class SLV, class output, class real> static inline
    void get(SLV* slv, output& sol, const Homography<real>& H)
    {
      typedef typename output::value_type sol_t;
      sol_t u,v;
      real dt=H.a*H.d-H.b*H.c;
      if(dt>0)
	for (typename SLV::sz_t i = 0; i < slv->nb_sol(); i ++ )
	  {
	    slv->Convert(u,slv->m_res[i].a,H);
	    slv->Convert(v,slv->m_res[i].b,H);
	    u+=v;u/=2;
	    sol << u;
	  }
      else
	for (int i = slv->nb_sol()-1; i>=0; i-- )
	  {
	    slv->Convert(u,slv->m_res[i].a,H);
	    slv->Convert(v,slv->m_res[i].b,H);
	    u+=v;u/=2;
	    sol << u;
	  }
    }
  };

  //--------------------------------------------------------------------
  template<class C, int N>  struct solutions< Interval<C,N>, meta::false_t>
  {
    template<class SLV, class output, class real> static inline
    void get(SLV* slv, output& sol, const real& first = 0, const real& last =1)
    {
      typedef typename output::value_type sol_t;
      typedef typename SLV::sz_t          sz_t;
      //      unsigned s = o.size();
      real a,b;
      //      o.resize(s+2*size());
      real scale = last-first;
      for ( sz_t i = 0; i < slv->nb_sol(); i ++ )
	{
	  slv->Convert(a,slv->m_res[i].a,first, scale);
	  slv->Convert(b,slv->m_res[i].b,first, scale);
	  sol <<sol_t(a,b);
	};
    };

    template<class SLV, class output, class real> static inline
    void get(SLV* slv, output & sol, const Homography<real>& H)
    {
      typedef typename output::value_type sol_t;
      typedef typename SLV::sz_t          sz_t;
      typedef typename sol_t::value_type  bound_t;
      //      unsigned s = o.size();
      bound_t u,v;
      real dt= H.a*H.d-H.b*H.c;
      if(dt>0)
	for ( sz_t i = 0; i < slv->nb_sol(); i ++ )
	  {
	    slv->Convert(u,slv->m_res[i].a,H);
	    slv->Convert(v,slv->m_res[i].b,H);
	    sol << sol_t(u,v);
	  }
      else
	for ( int i = slv->nb_sol()-1; i >=0 ; i --)
	  {
	    slv->Convert(u,slv->m_res[i].b,H);
	    slv->Convert(v,slv->m_res[i].a,H);
	    sol << sol_t(u,v);
	  }
    }
  };
  //--------------------------------------------------------------------
  template<class C, int N>  struct solutions< Interval<C,N>, meta::true_t>
  {
    template<class SLV, class output, class real> static inline
    void get(SLV* slv, output& sol, const real& first = 0, const real& last =1)
    {
      typedef typename output::value_type sol_t;
      typedef typename SLV::sz_t          sz_t;
      unsigned s = slv->nb_sol(); //o.size();
      std::vector<C> a(s);
      real b;
      {	
	numerics::rounding<C> rnd(numerics::rnd_dw());
	real scale= last-first;
	for ( sz_t i = 0; i < slv->nb_sol(); i ++ )
	  slv->Convert(a[i],slv->m_res[i].a,first,scale);
      }
      {
	numerics::rounding<C> rnd(numerics::rnd_up());
	real scale= last-first;
	for ( sz_t i = 0; i < slv->nb_sol(); i ++ )
	  {
	    slv->Convert(b,slv->m_res[i].b,first,scale);
	    sol <<sol_t(a[i],b);
	  }
      }
    }
    template<class SLV, class output, class real> static inline
    void get(SLV* slv, output& sol, const Homography<real>& H)
    {
      typedef typename SLV::sz_t          sz_t;
      typedef typename output::value_type sol_t;
      typedef typename sol_t::value_type  bound_t;
      unsigned s = slv->nb_sol();
      std::vector<bound_t> u(s);
      bound_t v;
      real dt= H.a*H.d-H.b*H.c;
      std::cout<<"\t"<<dt<<std::endl;
      if(dt>0)
	{
	  {	
	    numerics::rounding<C> rnd(numerics::rnd_dw());
	    for ( sz_t i = 0; i < s; i ++ )
	      {
		slv->Convert(u[i],slv->m_res[i].a,H);
	      }
	  }
	  {
	    numerics::rounding<C> rnd(numerics::rnd_up());
	    for ( sz_t i = 0; i < s; i ++ )
	      {
		slv->Convert(v,slv->m_res[i].b,H);
		sol << sol_t(u[i],v);
	      }
	  }
	}
      else
	{
	  {	
	    numerics::rounding<C> rnd(numerics::rnd_dw());
	    for (int i = s-1; i >=0; i -- )
	      slv->Convert(u[i],slv->m_res[i].a,H);
	  }
	  {
	    numerics::rounding<C> rnd(numerics::rnd_up());
	    for (int i = s-1; i >=0; i -- )
	      {
		slv->Convert(v,slv->m_res[i].b,H);
		sol << sol_t(v,u[i]);
	      }
	  }
	}
    }
  };
  //====================================================================
  /** Solver class using sleeve approximation with coefficients of type \c C. **/
  template < class C , class K>
  struct Solver<C,Sleeve<K> > : Solver<C,Simple<K> >
  {

    typedef typename K::integer  integer;
    typedef typename K::rational rational;
    //    typedef typename K::floating floating;
    //    typedef typename K::ieee     ieee;

    typedef C                           creal_t;
    typedef unsigned                    sz_t; 
    typedef Solver<C,Simple<K> >        Base_t;
    typedef typename Base_t::unsigned_t unsigned_t;
    typedef typename Base_t::res_t      Domain_t;
    typedef typename Base_t::bound      bound;
    sz_t  m_cup;
    sz_t  m_cdw;

    Solver(): Base_t(1e-6){}
    Solver(C e): Base_t(e){}

    inline 
    void alloc( sz_t s, sz_t deep  ) 
    {
      Base_t::alloc(s,2*s,deep); 
    };
    
    void barre( char c, unsigned n )
    {
      char  _bar[ n+1 ];
      std::fill(_bar,_bar+n, c);
      _bar[n] = 0;
      std::cout << _bar << std::endl;
    };
    
    void writebounds( creal_t * pup, creal_t * pdw, unsigned s )
    {
      print(pup,s); std::cout << std::endl;
      print(pdw,s); std::cout << std::endl;
    };
    
    void writedomain( int d )
    {
      std::cout << "Domaine-" << d << std::endl;
      std::cout << "\thauteur = " << this->m_dmn[d].e << std::endl;
      std::cout << "\tvaleur  = " << this->m_dmn[d].m << std::endl;
    };

    void writebck()
    {
      std::cout << "Back " << std::endl;
      std::cout << "\tA = " << this->bcka().e << " , " << this->bcka().m  << std::endl;
      std::cout << "\tB = " << this->bckb().e << " , " << this->bckb().m  << std::endl;
    };
    

    inline 
    bool glue( sz_t cup, sz_t cdw, int d )
    {
      if ( this->bckb() == this->m_dmn[d] )
	{
	  this->bckb() = this->m_dmn[d];
	  this->bckb().m += 1;
	  if ( cup ) m_cup = cup;
	  if ( cdw ) m_cdw = cdw;
	  return true;
	};
      return false;
    };
    
    inline 
    void mstore( sz_t cup, sz_t cdw, int d  ) 
    {
      if ( this->m_res.size() == 0 || !glue(cup,cdw,d)  )
	{
	  Base_t::mstore(d);
	  m_cup = cup;
	  m_cdw = cdw;
	  //std::cout << "Push ";
	};
      //      writebck();
    };
  
    
    inline
    void dwsplit( creal_t * r, creal_t * l, sz_t s )
    {
      numerics::rounding<creal_t> rnd( numerics::rnd_dw() );
      this->split(r,l,s);
    };

    inline 
    void upsplit( creal_t * r, creal_t * l, sz_t s )
    {
      numerics::rounding<creal_t> rnd( numerics::rnd_up() );
      this->split(r,l,s);
    };
    
    void Loop( bool isole = true )
    {
      //  std::cout << "LOOP\n";
      numerics::rounding<creal_t> rnd(numerics::rnd_dw());
      creal_t * pup, * pdw;
      int d;
      sz_t a,s,cup,cdw;
      unsigned_t v;
      s = this->m_s;
      pup = this->m_data;
      pdw = this->m_data + s;
      this->m_dmn[0].m = 0;
      this->m_dmn[0].e = 0;
      for( a = 0, d = 0; d >= 0; d--, a -= 2*s   ) 
	{
	  cup = sgncnt(pup+a,s);
	  cdw = sgncnt(pdw+a,s);
	  if ( cup || cdw )
	    {
	      if ( isole && cup == cdw && cup == 1 ){ 
// 		std::cout << " (b) \n";
		this->sstore(d);
		continue; 
	      };
	      if ( this->m_dmn[d].e == this->m_limit-1 ) 
		{ 
//   		  std::cout << " (c) "<<cup<<" "<<cdw<<"\n";
 		  if ( cup == cdw && cup == 1 )
 		    this->sstore(d);
		  else
		    this->mstore(cup,cdw,d); 
		  continue; 
		};
	      dwsplit(pdw+a,pdw+a+2*s,s);    
	      upsplit(pup+a,pup+a+2*s,s);
	      this->split(this->m_dmn[d],this->m_dmn[d+1]);
	      d += 2;
	      a += 4*s;
	      continue;
	    }
	  else
	    {
	      //std::cout<<"  (a) ";
	      sz_t k =0, sv=0;
	      for (  k = 0; k < this->m_s; k ++ )
		{
		  if ( ( pup[a+k] > 0 ) != ( pdw[a+k] > 0 ) )
		    {
		      sv++;
		      if(sv>1)
			{ 
			  mstore(cup,cdw,d); 
			  break; 
			};
		    }
		}
	      //std::cout<<std::endl;
	    };
	};
      //      std::cout << "END LOOP\n";
    };
    
        
    template<class input>
    void run_loop( const input& in, const creal_t& eps, bool isole, const meta::true_t&   )
    {
      sz_t prec = numerics::bitprec(eps);
      alloc(in.size(),prec);
      std::copy(in.begin(),in.end(),this->m_data);
      std::copy(in.begin(),in.end(),this->m_data+this->m_s);
      Loop(isole);
    };
    
    const Domain_t& operator[]( int i ) const  { return this->m_res[i]; };
    Domain_t&       operator[]( int i ) { return this->m_res[i]; };
    
  
    template<class output, class input,class MTH> inline
    void solve_bernstein( output& out, const input& in, const MTH& mth )
    {
      run(in,this->eps,mth.isolate);
      solutions<typename output::value_type>::get(this,out);
    }

    template<class VECT,class POL>
    void init_pol(VECT& ubp, VECT& dbp, const POL& r,  unsigned sz, bool neg = false); 
    
    template<class VECT,class POL,class Q>
    void init_pol(VECT& ubp, VECT& dbp, const POL& r,  unsigned sz, const Q& u, const Q&v);


    template<class input> inline
    void run( const input& in, const creal_t& eps, bool isole )
    { 
      typedef typename numerics::is_rounded<C>::result_t round_t;
      run_loop(in,eps,isole,round_t()); 
    }
    
    template<class input>
    void run( const input& up, const input& dw, const creal_t& eps, bool isole )
    {
      assert(up.size()==dw.size());
      sz_t prec = numerics::bitprec(eps);
      alloc(up.size(),prec);
      std::copy(up.begin(),up.end(),this->m_data);
      std::copy(dw.begin(),dw.end(),this->m_data+this->m_s);
      Loop(isole);
    }
    
    template <typename output, typename POL, typename MTH >
    void solve_positive(output& sol, const POL& r, const MTH& mth, bool pos=true);

    //--------------------------------------------------------------------
    /** Compute the roots of the polynomial \c r expressed in the monomial basis. If
	If \c MTH=Isolate, the ordered sequence of isolating intervals is stored in \c sol. 
	If \c MTH=Approx, the ordered sequence of approximation of the roots is stored in \c sol.
    **/ 
    template <typename output, typename POL, typename MTH >
    void solve(output& sol, const POL& r, const MTH& mth= Approx()) 
    {
      //      this->root_bound = - bound_root(r,Cauchy<C>());
      this->solve_positive(sol,r,mth,false);
      //      this->root_bound = - this->root_bound;
      this->solve_positive(sol,r,mth,true);
    }

    /** Compute the roots of the polynomial \c r expressed in the monomial
	basis, in the interval [u,v].
	If \c MTH=Isolate, the ordered sequence of isolating intervals is stored in \c sol. 
	If \c MTH=Approx, the ordered sequence of approximation of the roots is stored in \c sol.
    **/ 
    template <typename output, typename POL, typename Q, typename MTH>
    void solve(output& sol, const POL& r, const Q& u, const Q& v, const MTH& m);

    //--------------------------------------------------------------------
    template<class output, class input,class real,class MTH> inline
    void solve_bernstein( output& out, const input& up, const input& dw, 
			  const real& u, const real& v, const MTH& mth)
    { 
      run(up,dw,this->eps,mth.isole); 
      solutions<typename output::value_type>::get(this,out,u,v);
    }
    //--------------------------------------------------------------------
    /** Compute the roots of the polynomial \c r expressed in the bernstein
	basis, on the interval [u,v].
	If \c MTH=Isolate, the ordered sequence of isolating intervals is stored in \c sol. 
	If \c MTH=Approx, the ordered sequence of approximation of the roots is stored in \c sol.
    **/ 
    template <typename output, typename POL ,typename real, typename MTH>
    void solve_bernstein(output& sol, const POL& r, const real& u,const real& v, const MTH& mth )
    {
      typedef typename output::value_type    sol_t;    
      using let::assign;
      
      unsigned sz= r.size();
      std::vector<C> ubp( sz ), dbp( sz );
      {
	numerics::rounding<C> rnd( numerics::rnd_up() );
	for(unsigned i=0;i<sz;i++) ubp[i] =as<C>(r[i]);
      }
      {
	numerics::rounding<C> rnd( numerics::rnd_dw() );
	for(unsigned i=0;i<sz;i++) dbp[i] =as<C>(r[i]);
      }
      
      run(ubp,dbp,this->eps,mth.isolate);
      solutions<typename output::value_type>::get(this, sol, u,v);
    }
  };
  //====================================================================
  template<class C, class K>
  template<class VECT,class POL>
  void Solver<C,Sleeve<K> >::init_pol(VECT& ubp, VECT& dbp, const POL& p,unsigned sz, bool pos)
  {
    typedef typename K::integer  integer;
    typedef typename K::rational rational;

    if(pos)
      this->root_bound = bound_root(p,Cauchy<C>());
    else
      this->root_bound = - bound_root(p,Cauchy<C>());

    //    std::cout <<"Root bound: "<< this->root_bound <<std::endl;

    using let::assign;
    
    rational mx = as<rational>(VECTOR::max_abs(p));
    if( mx == 0 ) return;
 
    std::vector<rational> bp( sz );
    for ( unsigned i = 0; i < sz; i ++ )
      {
	bp[i]=as<rational>(p[i]);
	bp[i]/= binomial<integer>(sz-1,i);
	bp[i]/=mx;
      }
 
    if(!pos) for(unsigned i=1;i<sz;i+=2) bp[i]=-bp[i];

    {
      numerics::rounding<C> rnd( numerics::rnd_up() );
      for ( unsigned i = 0; i < sz; i ++ )
	  ubp[i]=as<C>(bp[i]);
    }
    {
      numerics::rounding<C> rnd( numerics::rnd_dw() );
      for ( unsigned i = 0; i < sz; i ++ )
	  dbp[i]=as<C>(bp[i]);
    }

  }
  //--------------------------------------------------------------------
  template<class C, class K>
  template<class VECT,class POL,class Q>
  void Solver<C,Sleeve<K> >::init_pol(VECT& ubp, VECT& dbp, const POL& r,unsigned sz,const Q& u, const Q&v)
  {
    using let::assign;
  
#if 1
    //    PRINT_DEBUG("Conversion using exact arithmetic");
    rational U,V;
    assign(U,u);
    assign(V,v);
    //    std::vector<interval> bp( sz );
    std::vector<rational> bp(sz);
    BEZIER::monomial_to_bezier(bp,r,sz,U,V);
    rational mx = VECTOR::max_abs(bp);
    VECTOR::div(bp,mx);
    // unsigned i;
    //    for (i = 0; i < sz && bp[i].upper()-bp[i].lower()<as<floating>(1.e-16); i ++ );
    //    if(i<sz) std::cout<<"refine: "<<i<<" " <<(bp[i].upper()-bp[i].lower())<<std::endl;
    //    double l; assign(l, rational(V-U));
    //    std::cout<<"\t init: "<<sign_variation(bp)<<" "<<l<<std::endl;
    {
      numerics::rounding<C> rnd( numerics::rnd_up() );
      for ( unsigned i = 0; i < sz; i ++ ) ubp[i]=as<C>(bp[i]);
    }
    {
      numerics::rounding<C> rnd( numerics::rnd_dw() );
      for ( unsigned i = 0; i < sz; i ++ ) dbp[i]=as<C>(bp[i]);
    };
   
#else
    //    PRINT_DEBUG("Using rounding arithmetic");
    O ud, vd;
    {
      numerics::rounding<C> rnd( numerics::rnd_dw() );
      ud=as<C>(u);
    };
    {
      numerics::rounding<C> rnd( numerics::rnd_up() );
      vd=as<C>(v);
      BEZIER::monomial_to_bezier(ubp,r,r.size(),ud,vd);
    };
    {
      numerics::rounding<C> rnd( numerics::rnd_dw() );
      BEZIER::monomial_to_bezier(dbp,r,r.size(),ud,vd);
    };

    for ( unsigned i= 0; i < ubp.size(); i ++ )
      if ( dbp[i] > ubp[i] ) { 
	std::swap(dbp[i], ubp[i]);
      };
#endif

  }
  //------------------------------------------------------------------
  template<class C, class K>
  template <typename output, typename Poly, typename MTH >
  void Solver<C,Sleeve<K> >::solve_positive(output& sol, const Poly& r, const MTH& mth, bool pos)
  {
    unsigned sz=r.size();
    std::vector<C> ubp(sz), dbp(sz);

    init_pol(ubp,dbp,r,sz,pos);
    run(ubp,dbp,this->eps,mth.isolate);
    Homography<C> H(1,0,-1,1);
    if(!pos) {H.c=1;H.d=-1;}
    solutions<typename output::value_type>::get(this,sol,H);
  }
  //--------------------------------------------------------------------
  template <class C, class K>
  template <typename output, typename POL ,typename Q, typename MTH>
  void Solver<C,Sleeve<K> >::solve(output& sol, const POL& r, const Q& u, const Q& v, const MTH& mth)
  {
    unsigned sz= r.size();
    std::vector<C> ubp(sz), dbp(sz);
    init_pol(ubp,dbp,r,sz,u,v);
    run(ubp,dbp,this->eps,mth.isolate);
    solutions<typename output::value_type>::get(this,sol,u,v);
  }
//====================================================================
} //namespace mmx
#endif  //synaps_solver1d_sleeve_hpp
