#ifndef subdivix_subresultant_hpp
#define subdivix_subresultant_hpp
/********************************************************************/
#include <subdivix/euclidean.hpp>
#include <subdivix/Seq.hpp>

namespace mmx {

template<class PREM>
struct subresultant
{
  template<class Pol> 
  static Seq<Pol> sequence(const Pol & p, const Pol & q)
  {
    typedef typename Pol::value_type coeff_type;
    assert(degree(p) >=0 && degree(q) >= 0);
    int n1 = degree(p), n2= degree(q);
    coeff_type lp = p[n1], lq = q[n2];
    
    Seq<Pol> tab;
    unsigned N = min(n1,n2);
    tab.resize(N);
    for(unsigned i=0;i<N;i++) tab[i]=Pol(0);
    
    int d,da,db;
    coeff_type la;
    Pol a, b, Sr;
    
    if (n1>=n2) {
	a=q;
	b=PREM::prem(p,q);
	d=n1-n2-1;
	la=pow(lq,n1-n2);
    } else {
      a=p;
      b=PREM::prem(q,p);
      d=n2-n1-1;
      la=pow(lp,n2-n1);
    }
    
    if (d!=0) {
      a*=pow<int>(-1,d*(d+1)/2);
      b*=pow<int>(-1,(d+2)*(d+3)/2);
    }
    da=degree(a);
    db=degree(b);
    if(db>=0) tab[db]=b;
    
    while (db>=0) {
      Sr=PREM::prem(a,b)/pow(la,da-db)/a[da];
      Sr*=pow<int>(-1,((da-db+1)*(da-db+2))/2);
      a=b*pow(b[db],da-db-1)/pow(la,da-db-1); 
      da=degree(a); 
      b=Sr; 
      la=a[da]; 
      db=degree(b);
      if (db>=0) tab[db]=b;
    }
    return tab;
  }

  template<class Pol> 
  static typename Pol::value_type resultant(const Pol & p, const Pol & q)
  {
    Seq<Pol> s= sequence(p,q);
    if(s.size()>0) 
      return s[0][0];
    else return 0;
  }
};

}// namespace mmx
#endif //subdivix_subresultant_hpp
