#ifndef SYNAPS_ARITHM_TRAITS_ISEXACT_H
#define SYNAPS_ARITHM_TRAITS_ISEXACT_H

#include <subdivix/texp_kernelof.hpp>

namespace mmx {

namespace meta {

  template<class X> struct isexact_ { typedef false_t T; };
  
  template<class K, class X> struct kernel_isexact  { typedef true_t T; };
  template<class X> struct kernel_isexact<null_t,X> { typedef typename isexact_<X>::T T; };

  template<class K> struct kernel_isexact< K, typename K::integer >  { typedef true_t T; };
  template<class K> struct kernel_isexact< K, typename K::rational > { typedef true_t T; };
  template<class K> struct kernel_isexact< K, typename K::floating > { typedef false_t T; };
  template<class K> struct kernel_isexact< K, typename K::ieee >     { typedef false_t T; };

  /** Traits class which specifies if the type \c X is exact.
      For a type \c X in a kernel \c K, we have the following type values fo:
      \li \c isexact<K::integer>:  \c true_t
      \li \c isexact<K::rational>: \c true_t
      \li \c isexact<K::floating>: \c false_t
      \li \c isexact<K::ieee>:     \c false_t
      If not specified, the default value \c isexact<X>::T is \c false_t.
  **/
  template<typename X>
  struct isexact 
  {
    typedef  typename ucstref<X>::T   _X;
    typedef  typename kernel_isexact< typename kernelof<X>::T, _X >::T T;
  };
  
  /// Traits class which specifies if the type \c X is not exact.
  template<typename X>
  struct isapproximate { typedef typename tnot< typename isexact< X >::T >::T T; };

}

} //namespace mmx

#endif


