#ifndef SYNAPS_ARITHM_TEXP_OPERATORS_H
#define SYNAPS_ARITHM_TEXP_OPERATORS_H

#include <subdivix/texp_double.hpp>
#include <subdivix/texp_expression.hpp>

namespace mmx {

#define define_binary_operator_structure(opname,function,call)      \
                                                          \
struct opname                                             \
{                                                         \
  inline static const char * name() { return #call; };    \
  template < class A, class B, class C > inline           \
  void operator()( A & a, const B & b, const C & c) const \
  {                                                       \
    call (a, b, c);                                       \
  };                                                      \
  template < class A, class B > inline                    \
  void operator()( A & a, const B & b )             const \
  {                                                       \
    call (a, b);                                          \
  };                                                      \
};                                                        \

#define define_binary_operator(opname,function,call)\
define_binary_operator_structure(opname,function,call)\
template<class X, class Y>  inline                                                 \
meta::template_expression<                                                       \
meta::binary_operator<opname,                                                    \
	meta::template_expression<X>,                                            \
	meta::template_expression<Y> > >                                         \
function( const meta::template_expression<X>& x,                                 \
	  const meta::template_expression<Y>& y )                                \
{                                                                                  \
  return meta::template_expression< meta::binary_operator< opname,             \
                                      meta::template_expression<X>,              \
                                      meta::template_expression<Y> > > ( x, y ); \
};                                                                                 \
\
template<class X, class Y> inline                                                  \
meta::template_expression<                                                       \
meta::binary_operator<opname,                                                    \
	meta::template_expression<X>,                                            \
	meta::template_expression<Y> > >                                         \
function( const meta::template_expression<X>& x, const Y& y )                    \
{                                                                                  \
  return meta::template_expression< meta::binary_operator< opname,             \
                                      meta::template_expression<X>,              \
                                      meta::template_expression<Y> > >           \
                                   ( x, meta::template_expression<Y>(y));        \
};                                                                                 \
\
template<class X, class Y> inline                                                  \
meta::template_expression<                                                       \
meta::binary_operator<opname,                                                    \
			meta::template_expression<X>,                            \
			meta::template_expression<Y> > >                         \
function( const X& x, const meta::template_expression<Y>& y )                    \
{                                                                                  \
  return                                                                           \
    meta::template_expression< meta::binary_operator<opname                    \
    ,                            meta::template_expression<X>                    \
    ,                            meta::template_expression<Y> > >                \
    (meta::template_expression<X>(x),y);                                         \
};                                                                                 \
\
template<class X,class Y>                                                          \
X & call( X & r, const meta::template_expression<Y> & exp )                      \
{                                                                                  \
  typename template_expression<Y>::E tmp;                                          \
  exp.eval(tmp);                                                                   \
  call(r,tmp);                                                                     \
  return r;                                                                        \
}                                                                                  \

#define define_unary_operator(opname,function,call)\
\
struct opname                                      \
{                                                  \
  inline static const char * name()                \
  {                                                \
    return #call;                                  \
  };                                               \
                                                   \
  template < class A, class B > inline             \
  void operator()(A & a, const B & b) const        \
  {                                                \
    call (a, b);                                   \
  };                                               \
  template < class A> inline                       \
  void operator()(A & a) const                     \
  {                                                \
    call (a);                                      \
  };                                               \
};                                                 \
\
template<class X> inline                                            \
meta::template_expression<                                        \
  unary_operator< opname,                                           \
		  meta::template_expression<X> > >                \
function( const meta::template_expression<X>& x )                 \
{                                                                   \
  return                                                            \
    meta::template_expression<                                    \
    unary_operator<opname, meta::template_expression<X> > >( x ); \
};                                                                  \
\
template<class X> inline                                            \
meta::template_expression<                                        \
  unary_operator< opname,                                           \
		  meta::template_expression<X> > >                \
function( const X& x )                                              \
{                                                                   \
  return                                                            \
    meta::template_expression<                                    \
    unary_operator<opname, meta::template_expression<X> > >       \
    ( meta::template_expression<X>(x) );                          \
};                                                                  \

#define declare_binary_operator(tp_arg_list,parm0,parm1,opname,function)\
tp_arg_list inline \
meta::template_expression< meta::binary_operator< opname,\
meta::template_expression< parm0 >, \
meta::template_expression< parm1 > > > \
function( const parm0 & x, const parm1 & y )\
{ return \
    meta::template_expression< \
    meta::binary_operator< \
    opname, meta::template_expression< parm0 >, meta::template_expression< parm1 > > >\
    (meta::template_expression< parm0 >(x),meta::template_expression< parm1 >(y));\
};\
tp_arg_list inline \
meta::template_expression< meta::binary_operator< opname , \
meta::template_expression< parm0 >,\
meta::template_expression< parm1 > > >\
function( const meta::template_expression<parm0> & x, const parm1 & y )\
{ return meta::template_expression< meta::binary_operator< opname, meta::template_expression< parm0 >, meta::template_expression< parm1 > > >\
 (x,meta::template_expression<parm1>(y)); };\
tp_arg_list inline \
meta::template_expression< meta::binary_operator< opname , \
meta::template_expression< parm0 >, \
meta::template_expression< parm1 > > >\
function( const parm0 & x, const meta::template_expression<parm1> & y )\
{ return meta::template_expression< meta::binary_operator< opname, meta::template_expression< parm0 >, meta::template_expression< parm1 > > >\
(meta::template_expression< parm0 >(x),y); }
  
#define declare_unary_operator(tp_arg_list,parm,opname,function)\
tp_arg_list inline \
meta::template_expression< meta::unary_operator< opname, \
						     meta::template_expression<parm> > >\
function( const parm& x ) \
{ \
  return meta::template_expression<\
    meta::unary_operator< opname, \
    meta::template_expression<parm> > >( meta::template_expression<parm>(x) ); \
};\


}// end namespace mmx

#endif
