#ifndef SYNAPS_ARITHM_TRAITS_STRUCTUREOF_H
#define SYNAPS_ARITHM_TRAITS_STRUCTUREOF_H

#include <subdivix/type.hpp>
#include <subdivix/texp_kernelof.hpp>

namespace mmx {

namespace meta {
  
  namespace structure {
    
    struct scalar
    {
      struct floating;
      struct integer;
      struct rational;
      struct complex;
      struct interval;
    };
    
    struct vector;
    struct matrix;
    struct polynomial;

    struct dense;
    struct sparse;
    struct structured;
    struct univariate;
    struct multivariate;
    struct monomial;
    struct bernstein;
  };
  
  template<class X>  struct scalar_type {};
  
  template<class X> struct structureof_
  { 
    typedef meta::null_t T; 
  };

  template<class K, class X>
  struct kernel_structureof  { typedef structure::scalar T; };

  template<class X>
  struct kernel_structureof<null_t,X> { typedef typename structureof_<X>::T T; };

  template<class K>
  struct kernel_structureof< K, typename K::integer >  { typedef structure::scalar T; };
  template<class K>
  struct kernel_structureof< K, typename K::rational > { typedef structure::scalar T; };
  template<class K>
  struct kernel_structureof< K, typename K::floating > { typedef structure::scalar T; };
  template<class K>
  struct kernel_structureof< K, typename K::ieee >     { typedef structure::scalar T; };



  
  template<class X> struct structureof { 
    typedef typename ucstref<X>::T   _X;
    typedef typename kernelof<X>::T  K;
    typedef typename kernel_structureof<K, _X >::T T;
};



  //  template<> struct structureof_<double> { typedef structure::scalar T; };

  
  
  
};

}// end namespace mmx

#endif
