/*********************************************************************
*    This file is part of the source code of SYNAPS library.         *
*    Author(s): B. Mourrain, GALAAD, INRIA                           *
**********************************************************************/
#ifndef synaps_upol_xml_H_
#define synaps_upol_xml_H_
#include <string.h>
#include <string>
//--------------------------------------------------------------------
namespace mmx {

//--------------------------------------------------------------------
/// Namespace for parsing and printing in @xml@  format.
namespace xml {
  /** Skip the input stream until s is found. It returns true if @s@ is found.
   **/
  template<class ISTREAM>
  bool skip(ISTREAM& is, const char* s) 
  {
    char c=(*s); const char* t=s;
    unsigned n= strlen(s),i=0;    
    while(c != EOF)
      {
	for(i=0; i<n && (c !=EOF) && (c=is.get()) == (*t); i++,t++);
	if(i==n) 
	  return true;
	else 
	  t=s;
      }
    return false;
  }

  template<class OSTREAM, typename X>
  OSTREAM& print(OSTREAM& os, const X& x, const char *tag)
  {
    os<<"<"<<tag<<">"<<x<<"</"<<tag<<">";
    return os;
  }


  #define begin_polynomial "<polynomial>"
  #define end_polynomial "</polynomial>"

 
  /// Read between @tag@ and "<" and build p from this string;
  template<class ISTREAM>
  std::string read_tag(ISTREAM& is, const char* tag)
  {
    if(skip(is,tag))
      {
	std::string s; char c;
	while((c=is.get()) != EOF && c != '<') 
	  {
	    if(c != '\n') s+=c;
	  }
	//	s+='\0';
	return s;
      }
    else
      {
	return std::string("");
      }
  }
 /// Read a polynomial, enclosed in @polynomial@ tags.
  template<class ISTREAM, typename T>
  bool read_pol(ISTREAM& is, T& p)
  {
    if(skip(is,begin_polynomial))
      {
        ::std::string s; char c;
	while((c=is.get()) != EOF && c != '<') 
	  {
	    if(c != ' ' && c != '\n') s+=c;
	  }
	s+='\0';
	//PRINT_DEBUG(s.data());
	p = T(s.data());
	return true;
      }
    else 
      {
	p=T(0); return false;
      }
  }

  /// Read a polynomial, enclosed in @polynomial@ tags.
  template<class ISTREAM, typename T, class VAR>
  bool read_pol(ISTREAM& is, T& p, const VAR& V)
  {
    if(skip(is,begin_polynomial))
      {
        std::string s; char c;
	while((c=is.get()) != EOF && c != '<') 
	  {
	    if(c != ' ' && c != '\n') s+=c;
	  }
	s+='\0';
	p = T(s.data(),V);
	return true;
      }
    else 
      {
	p=T(0); return false;
      }
  }
}

//--------------------------------------------------------------------
template<typename C,typename R> struct upoldse;
//--------------------------------------------------------------------
namespace xml 
{
  //  using namespace SYNAPS;

  /// Read a polynomial, enclosed in @polynomial@ tags.
  template<class ISTREAM, typename T>
  bool read_upol(ISTREAM& is, T& p)
  {
    if(skip(is,begin_polynomial))
      {
        std::string s; char c;
	while((c=is.get()) != EOF && c != '<') 
	  {
	    if(c != ' ' && c != '\n') s+=c;
	  }
	s+='\0';
	p = T(s.data());
	return true;
      }
    else 
      {
	p=T(0); return false;
      }
  }

  /// Read a univariate polynomial, enclosed in @polynomial@ tags.
  template<class ISTREAM, typename C,typename R>
  bool read(ISTREAM& is, upoldse<C,R>& p)
  {
    return read_upol(is,p);
  }

  /// Print a univariate polynomial, with its @polynomial@ tags.
  template<class OSTREAM, typename C,typename R>
  OSTREAM& print(OSTREAM& os, const upoldse<C,R>& p)
  {
    xml::print(os,p,"polynomial");
    return os;
  }

}
//----------------------------------------------------------------------
} //namespace mmx
//----------------------------------------------------------------------
#endif //synaps_upol_xml_H_
