
/******************************************************************************
* MODULE     : symbolic.hpp
* DESCRIPTION: Symbolic expressions
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC__HPP
#define __SYMBOLIC__HPP
#include <basix/vector.hpp>
#include <basix/table.hpp>
#include <basix/literal.hpp>
#include <basix/compound.hpp>
namespace mmx {
template<typename R> struct std_symbolic;
#define TMPL_DEF template<typename W=std_symbolic<generic> >
#define TMPL template<typename W>
#define TMPLV template<typename W, typename Var>
#define XTMPL template<typename X, typename W>
#define Symbolic symbolic<W>
#define Vector vector<Symbolic >
#define Set table<bool,Symbolic >
#define Table table<Symbolic,Symbolic >
TMPL class symbolic_rep;
TMPL class symbolic;

/******************************************************************************
* Symbolic class
******************************************************************************/

TMPL_DEF
class symbolic {
MMX_ALLOCATORS
public:
  typedef typename W::R R;
private:
  R rep;
public:
  inline symbolic (): rep (W::construct ()) {}
  template<typename C> inline symbolic (const C& c): rep (W::construct (c)) {}
  inline symbolic (const char* s): rep (W::construct (s)) {}
  inline symbolic (const R& g): rep (g) {}
  inline R operator * () const { return rep; }
  inline friend nat N (const Symbolic& x) { return W::arity (x); }
  inline Symbolic operator [] (nat i) const { return W::access (*this, i); }
  inline Symbolic operator () () const { return W::apply (rep); }
  inline Symbolic operator () (const Symbolic& x1) const {
    return W::apply (*this, x1); }
  Symbolic operator () (const Symbolic& x1, const Symbolic& x2) const {
    return W::apply (*this, x1, x2); }
  Symbolic operator () (const Symbolic& x1, const Symbolic& x2,
			const Symbolic& x3) const {
    return W::apply (*this, x1, x2, x3); }
};

/******************************************************************************
* Basic operations
******************************************************************************/

TMPL inline nat hash (const Symbolic& x) { return hash (*x); }
TMPL inline nat soft_hash (const Symbolic& x) { return soft_hash (*x); }
TMPL inline nat hard_hash (const Symbolic& x) { return hard_hash (*x); }
TMPL inline bool eq (const Symbolic& x, const Symbolic& y) {
  return eq (*x, *y); }
TMPL inline bool neq (const Symbolic& x, const Symbolic& y) {
  return neq (*x, *y); }
TMPL inline bool hard_eq (const Symbolic& x, const Symbolic& y) {
  return hard_eq (*x, *y); }
TMPL inline bool hard_neq (const Symbolic& x, const Symbolic& y) {
  return hard_neq (*x, *y); }

TMPL inline syntactic flatten (const Symbolic& x) {
  return W::expression (x); }
PRINT_SUGAR(TMPL,Symbolic)

TMPL inline nat symbolic_type (const Symbolic& x) {
  return symbolic_type (*x); }
XTMPL inline bool is (const Symbolic& x) {
  return is<X> (*x); }
TMPL inline bool is_scalar (const Symbolic& x) {
  return W::is_scalar (*x); }
TMPL inline bool is_atomic (const Symbolic& x) {
  return W::is_atomic (*x); }
TMPL inline bool is_composite (const Symbolic& x) {
  return W::is_composite (*x); }
TMPL inline bool is_literal (const Symbolic& x) {
  return is<typename W::literal> (x); }
TMPL inline bool is_compound (const Symbolic& x) {
  return is<typename W::compound> (x); }
TMPL inline bool is_sum (const Symbolic& x) {
  return is<typename W::sum> (x); }
TMPL inline bool is_product (const Symbolic& x) {
  return is<typename W::product> (x); }
TMPL inline bool is_derivative (const Symbolic& x) {
  return is<typename W::derivative> (x); }
TMPL inline bool is_function (const Symbolic& x) {
  return is<typename W::function> (x); }
TMPL inline bool is_term (const Symbolic& x) {
  return W::is_term (x); }
TMPL inline bool is_power (const Symbolic& x) {
  return W::is_power (x); }
TMPL inline bool is_func (const Symbolic& x, const Symbolic& f) {
  return is<typename W::compound> (*x) && N(x) > 0 && x[0] == f; }
TMPL inline bool is_func (const Symbolic& x, const Symbolic& f, nat n) {
  return is<typename W::compound> (*x) && N(x) == n+1 && x[0] == f; }

TMPL inline Symbolic apply (const vector<Symbolic >& v) {
  return W::apply (v); }
TMPL inline Symbolic apply (const Symbolic& s, const vector<Symbolic >& v) {
  return W::apply (s, v); }
TMPL inline Symbolic unapply (const Symbolic& s, const vector<Symbolic >& v) {
  return W::unapply (s, v); }
TMPL inline Symbolic access (const Symbolic& s, const vector<Symbolic >& v) {
  return W::apply (Symbolic (".[]"), cons (s, v)); }
TMPL inline vector<Symbolic > components (const Symbolic& x) {
  return W::components (x); }
TMPL inline vector<Symbolic > arguments (const Symbolic& x) {
  return cdr (components (x)); }
TMPL inline Symbolic component (const Symbolic& x, nat i) {
  return W::access (x, i); }
TMPL inline Symbolic argument (const Symbolic& x, nat i) {
  return W::access (x, i + 1); }
TMPL inline Symbolic get_monomial (const Symbolic& x) {
  return W::get_monomial (x); }
TMPL inline Symbolic get_coefficient (const Symbolic& x) {
  return W::get_coefficient (x); }
TMPL inline Symbolic get_base (const Symbolic& x) {
  return W::get_base (x); }
TMPL inline Symbolic get_exponent (const Symbolic& x) {
  return W::get_exponent (x); }

XTMPL
struct as_helper<X,Symbolic > {
  static inline X cv (const Symbolic& x) {
    return as<X> (*x); }
};

TMPL
struct as_helper<generic,Symbolic > {
  static inline generic cv (const Symbolic& x) {
    return new generic_concrete_rep<Symbolic > (x); }
};

TMPL
struct as_helper<literal,Symbolic > {
  static inline literal cv (const Symbolic& x) {
    return literal (as<symbol<string> > (*x)); }
};

TMPL
struct as_helper<compound,Symbolic > {
  static inline compound cv (const Symbolic& x) {
    vector<Symbolic > v= *as<symbol<vector<Symbolic > > > (*x);
    return compound (as<vector<generic> > (v)); }
};

/******************************************************************************
* Operators
******************************************************************************/

TMPL inline Symbolic operator - (const Symbolic& x1) {
  return Symbolic (W::neg (x1)); }
TMPL inline Symbolic operator + (const Symbolic& x1, const Symbolic& x2) {
  return Symbolic (W::add (x1, x2)); }
TMPL inline Symbolic operator - (const Symbolic& x1, const Symbolic& x2) {
  return Symbolic (W::sub (x1, x2)); }
TMPL inline Symbolic operator * (const Symbolic& x1, const Symbolic& x2) {
  return Symbolic (W::mul (x1, x2)); }
TMPL inline Symbolic operator / (const Symbolic& x1, const Symbolic& x2) {
  return Symbolic (W::div (x1, x2)); }
TMPL inline Symbolic big_add (const vector<Symbolic >& v) {
  return Symbolic (W::add (v)); }
TMPL inline Symbolic big_mul (const vector<Symbolic >& v) {
  return Symbolic (W::mul (v)); }

TMPL inline Symbolic sqrt (const Symbolic& x1) {
  return Symbolic (W::sqrt (x1)); }
TMPL inline Symbolic exp (const Symbolic& x1) {
  return Symbolic (W::exp (x1)); }
TMPL inline Symbolic log (const Symbolic& x1) {
  return Symbolic (W::log (x1)); }
TMPL inline Symbolic pow (const Symbolic& x1, const Symbolic& x2) {
  return Symbolic (W::pow (x1, x2)); }
TMPL inline Symbolic pow (const Symbolic& x1, const int& x2) {
  return pow (x1, Symbolic (x2)); }
TMPL inline Symbolic pow (const int& x1, const Symbolic& x2) {
  return pow (Symbolic (x1), x2); }
TMPL inline Symbolic cos (const Symbolic& x1) {
  return Symbolic (W::cos (x1)); }
TMPL inline Symbolic sin (const Symbolic& x1) {
  return Symbolic (W::sin (x1)); }
TMPL inline Symbolic tan (const Symbolic& x1) {
  return Symbolic (W::tan (x1)); }
TMPL inline Symbolic acos (const Symbolic& x1) {
  return Symbolic (W::acos (x1)); }
TMPL inline Symbolic asin (const Symbolic& x1) {
  return Symbolic (W::asin (x1)); }
TMPL inline Symbolic atan (const Symbolic& x1) {
  return Symbolic (W::atan (x1)); }
//TMPL inline Symbolic trig (const Symbolic& x1);

TMPL inline bool operator == (const Symbolic& x1, const Symbolic& x2) {
  return W::equal (x1, x2); }
TMPL inline bool operator != (const Symbolic& x1, const Symbolic& x2) {
  return W::unequal (x1, x2); }
TMPL inline bool operator <= (const Symbolic& x1, const Symbolic& x2) {
  return W::lesseq (x1, x2); }
TMPL inline bool operator < (const Symbolic& x1, const Symbolic& x2) {
  return W::less  (x1, x2); }
TMPL inline bool operator >= (const Symbolic& x1, const Symbolic& x2) {
  return W::gtreq (x1, x2); }
TMPL inline bool operator > (const Symbolic& x1, const Symbolic& x2) {
  return W::gtr  (x1, x2); }

ARITH_INT_SUGAR(TMPL,Symbolic,Symbolic);
EQUAL_INT_SUGAR(TMPL,Symbolic,Symbolic);
COMPARE_INT_SUGAR(TMPL,Symbolic,Symbolic);

TMPLV inline Symbolic derive (const Symbolic& x1, const Var& v) {
  return W::derive (x1, v); }
TMPL inline Symbolic derive (const Symbolic& x1, const vector<Symbolic >& v) {
  return W::derive (x1, v); }
TMPL inline nat size (const Symbolic& x1) {
  return W::size (x1); }
TMPL inline nat occurrences (const Symbolic& x1, const Symbolic& x2) {
  return W::occurrences (x1, x2); }
TMPL inline nat occurrences (const Symbolic& x1, const Vector& v) {
  return W::occurrences (x1, v); }
TMPL inline nat occurrences (const Symbolic& x1, const Set& s) {
  return W::occurrences (x1, s); }
TMPL inline bool depends (const Symbolic& x1, const Symbolic& x2) {
  return W::depends (x1, x2); }
TMPL inline bool depends (const Symbolic& x1, const Vector& v) {
  return W::depends (x1, v); }
TMPL inline bool depends (const Symbolic& x1, const Set& s) {
  return W::depends (x1, s); }
TMPL inline Symbolic
replace (const Symbolic& x1, const Symbolic& w, const Symbolic& b) {
  return W::replace (x1, w, b); }
TMPL inline Symbolic
replace (const Symbolic& x1, const Vector& w, const Vector& b) {
  return W::replace (x1, w, b); }
TMPL inline Symbolic replace (const Symbolic& x1, const Table& t) {
  return W::replace (x1, t); }
TMPL inline Symbolic solve (const Symbolic& x1, const Symbolic& x2) {
  return W::solve (x1, x2); }
TMPL inline vector<Symbolic >
solve (const vector<Symbolic >& x1, const vector<Symbolic >& x2) {
  return W::solve (x1, x2); }

//TMPL inline Symbolic integrate (const Symbolic& x1) {
//TMPL inline Symbolic integrate (const Symbolic& x1, const Symbolic& v) {
//TMPL inline Symbolic compose (const Symbolic& x1, const Symbolic& x2) {
//TMPL inline Symbolic dilate (const Symbolic& x1, const Symbolic& x2) {

#undef TMPL_DEF
#undef TMPL
#undef TMPLV
#undef XTMPL
#undef Symbolic
#undef Vector
#undef Set
#undef Table
} // namespace mmx
#endif // __SYMBOLIC__HPP
