
/******************************************************************************
* MODULE     : symbolic_elementary.hpp
* DESCRIPTION: Elementary functions on symbolic expressions
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_ELEMENTARY__HPP
#define __SYMBOLIC_ELEMENTARY__HPP
#include <symbolix/symbolic.hpp>
namespace mmx {

template<typename V>
struct symbolic_elementary: public V {
  typedef typename V::S S;
  typedef typename V::sum sum;
  typedef typename V::product product;
  using V::ZERO;
  using V::ONE;
  using V::MATHE;
  using V::COS;
  using V::SIN;
  using V::TAN;
  using V::ARCCOS;
  using V::ARCSIN;
  using V::ARCTAN;

  static S
  exp (const S& x) {
    if (x == ZERO) return ONE;
    return pow (MATHE, x);
  }

  static S
  log (const S& x) {
    if (x == ONE) return ZERO;
    if (x == MATHE) return ONE;
    if (is<sum > (x) && N (*as<sum > (x)) == 1) {
      sum s= as<sum > (x);
      return log (co (s, 0)) + log (mo (s, 0));
    }
    if (is<product> (x)) {
      product p= as<product> (x);
      vector<S> v;
      for (nat i=0; i<N(p); i++)
	v << (co (p, i) * log (mo (p, i)));
      return big_add (v);
    }
    return V::log (x);
  }

  static S
  cos (const S& x) {
    if (x == ZERO) return ONE;
    if (is_func (x, ARCCOS, 1)) return x[1];
    return V::cos (x);
  }

  static S
  sin (const S& x) {
    if (x == ZERO) return ZERO;
    if (is_func (x, ARCSIN, 1)) return x[1];
    return V::sin (x);
  }
  
  static S
  tan (const S& x) {
    if (x == ZERO) return ZERO;
    if (is_func (x, ARCTAN, 1)) return x[1];
    return V::tan (x);
  }
  
  static S
  acos (const S& x) {
    if (x == ONE) return ZERO;
    if (is_func (x, COS, 1)) return x[1];
    return V::acos (x);
  }

  static S
  asin (const S& x) {
    if (x == ZERO) return ZERO;
    if (is_func (x, SIN, 1)) return x[1];
    return V::asin (x);
  }

  static S
  atan (const S& x) {
    if (x == ZERO) return ZERO;
    if (is_func (x, TAN, 1)) return x[1];
    return V::atan (x);
  }
};

} // namespace mmx
#endif // __SYMBOLIC_ELEMENTARY__HPP
