
/******************************************************************************
* MODULE     : wrap_derivative.hpp
* DESCRIPTION: Symbolic lambda expressions
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __WRAP_DERIVATIVE__HPP
#define __WRAP_DERIVATIVE__HPP
#include <symbolix/symbolic.hpp>
#include <basix/symbol.hpp>
#include <basix/wrap.hpp>
namespace mmx {
#define TMPL template<typename W>
#define Symbolic symbolic<W>
#define Derivative wrap_derivative<W>
#define Derivative_rep symbol<pair<Symbolic,vector<nat> > >

TMPL
struct wrap_derivative {
MMX_ALLOCATORS
  Derivative_rep rep;
public:
  inline Derivative_rep operator * () const { return rep; }
  inline wrap_derivative (): rep () {}
  inline wrap_derivative (const Derivative_rep& rep2): rep (rep2) {}
  inline wrap_derivative (const Symbolic& fun):
    rep (Derivative_rep (pair<Symbolic,vector<nat> > (fun, vec<nat> ()))) {}
  inline wrap_derivative (const Symbolic& fun, const vector<nat>& by):
    rep (Derivative_rep (pair<Symbolic,vector<nat> > (fun, by))) {}
  inline friend syntactic flatten (const Derivative& df) {
    return W::expression (df); }
  inline friend Symbolic root (const Derivative& df) {
    return (*df.rep).x1; }
  inline friend vector<nat> orders (const Derivative& df) {
    return (*df.rep).x2; }
};

WRAP_WRAPPED_IMPL(TMPL inline,Derivative)
PRINT_SUGAR(TMPL,Derivative)

TMPL struct symbolic_type_information<Derivative > {
  static const nat id= SYMBOLIC_DERIVATIVE; };

TMPL syntactic dump (const Derivative& df) {
  return apply ("derivative", dump (root (df)), dump (orders (df))); }

TMPL inline Derivative derive (const Derivative& df, const nat& var) {
  vector<nat> by= copy (orders (df));
  if (N(by) <= var) by << fill<nat> (var + 1 - N(by));
  by[var]++;
  return Derivative (root (df), by); }

#undef TMPL
#undef Symbolic
#undef Derivative
#undef Derivative_rep
} // namespace mmx
#endif // __WRAP_DERIVATIVE__HPP
