
/******************************************************************************
* MODULE     : wrap_function.hpp
* DESCRIPTION: Symbolic lambda expressions
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __WRAP_FUNCTION__HPP
#define __WRAP_FUNCTION__HPP
#include <symbolix/symbolic.hpp>
#include <basix/symbol.hpp>
#include <basix/wrap.hpp>
namespace mmx {
#define TMPL template<typename W>
#define Symbolic symbolic<W>
#define Function wrap_function<W>
#define Function_rep symbol<vector<Symbolic > >

TMPL
struct wrap_function {
MMX_ALLOCATORS
  Function_rep rep;
public:
  inline Function_rep operator * () const { return rep; }
  inline wrap_function (): rep () {}
  inline wrap_function (const Function_rep& rep2): rep (rep2) {}
  inline wrap_function (const vector<Symbolic >& args, const Symbolic& body):
    rep (Function_rep (cons (body, args))) {}
  inline friend syntactic flatten (const Function& s) {
    return W::expression (s); }
  inline friend vector<Symbolic > arguments (const Function& s) {
    return cdr (*s.rep); }
  inline friend Symbolic body (const Function& s) {
    return car (*s.rep); }
};

WRAP_WRAPPED_IMPL(TMPL inline,Function)
PRINT_SUGAR(TMPL,Function)

TMPL struct symbolic_type_information<Function > {
  static const nat id= SYMBOLIC_FUNCTION; };

TMPL syntactic dump (const Function& f) {
  return apply ("lambda", dump (arguments (f)), dump (body (f))); }

#undef TMPL
#undef Symbolic
#undef Function
#undef Function_rep
} // namespace mmx
#endif // __WRAP_FUNCTION__HPP
