
/******************************************************************************
* MODULE     : wrap_partial.hpp
* DESCRIPTION: Partial derivatives
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __WRAP_PARTIAL__HPP
#define __WRAP_PARTIAL__HPP
#include <symbolix/symbolic.hpp>
#include <basix/symbol.hpp>
#include <basix/wrap.hpp>
namespace mmx {
#define TMPL template<typename W>
#define Symbolic symbolic<W>
#define Partial wrap_partial<W>
#define Partial_rep pair<Symbolic,Symbolic >

TMPL
struct wrap_partial {
MMX_ALLOCATORS
  Partial_rep rep;
public:
  inline Partial_rep operator * () const { return rep; }
  inline wrap_partial (): rep () {}
  inline wrap_partial (const Partial_rep& rep2): rep (rep2) {}
  inline wrap_partial (const Symbolic& var, const Symbolic& mult= 1):
    rep (Partial_rep (var, mult)) {}
  inline friend syntactic flatten (const Partial& p) {
    return flatten (p.rep.x2) * apply ("partial", flatten (p.rep.x1)); }
  inline friend Symbolic variable (const Partial& p) {
    return p.rep.x1; }
  inline friend nat number (const Partial& p) {
    return (nat) as_int (as<integer> (p.rep.x1)); }
  inline friend Symbolic multiplier (const Partial& p) {
    return p.rep.x2; }
};

WRAP_WRAPPED_IMPL(TMPL inline,Partial)
PRINT_SUGAR(TMPL,Partial)

TMPL syntactic dump (const Partial& p) {
  return apply ("partial", dump (variable (p)), dump (multiplier (p))); }

#undef TMPL
#undef Symbolic
#undef Partial
#undef Partial_rep
} // namespace mmx
#endif // __WRAP_PARTIAL__HPP
