
/******************************************************************************
* MODULE     : wrap_product.hpp
* DESCRIPTION: Symbolic power products with scalar exponents
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __WRAP_PRODUCT__HPP
#define __WRAP_PRODUCT__HPP
#include <symbolix/symbolic.hpp>
#include <basix/sparse_vector.hpp>
#include <basix/wrap.hpp>
namespace mmx {
#define TMPL template<typename W>
#define Symbolic symbolic<W>
#define Product wrap_product<W>
#define Product_rep sparse_vector<Symbolic,symbol<Symbolic >,hard_eq_sparse_vector>
#define Pair pair<symbol<Symbolic >,Symbolic>

TMPL
struct wrap_product {
MMX_ALLOCATORS
  Product_rep rep;
public:
  inline Product_rep operator * () const { return rep; }
  inline wrap_product (): rep () {}
  inline wrap_product (const Product_rep& rep2): rep (rep2) {}
  inline wrap_product (const wrap_product& p): rep (p.rep) {}
  inline wrap_product (const Symbolic& m, const Symbolic& c):
    rep (Pair (m, c)) {}
  inline friend syntactic flatten (const Product& p) {
    return W::expression (p); }
  inline friend nat N (const Product& p) { return N (p.rep); }
  inline friend Symbolic mo (const Product& p, nat i) {
    return *(p.rep[i].x1); }
  inline friend Symbolic co (const Product& p, nat i) {
    return p.rep[i].x2; }
  inline Symbolic operator [] (nat i) const {
    return W::power (*(rep[i].x1), rep[i].x2); }
};

WRAP_WRAPPED_IMPL(TMPL inline,Product)
PRINT_SUGAR(TMPL,Product)

TMPL struct symbolic_type_information<Product > {
  static const nat id= SYMBOLIC_PRODUCT; };

TMPL syntactic dump (const Product& p) {
  return apply ("product", arguments (dump (*p))); }

TMPL inline Product operator * (const Product& p1, const Product& p2) {
  return Product ((*p1) + (*p2)); }
TMPL inline Product operator / (const Product& p1, const Product& p2) {
  return Product ((*p1) - (*p2)); }
TMPL inline Product pow (const Product& p1, const Symbolic& x) {
  return Product ((*p1) * x); }
TMPL inline Product sqrtn (const Product& p1, const Symbolic& x) {
  return Product ((*p1) / x); }

#undef TMPL
#undef Symbolic
#undef Product
#undef Product_rep
#undef Pair
} // namespace mmx
#endif // __WRAP_PRODUCT__HPP
