
/******************************************************************************
* MODULE     : wrap_sum.hpp
* DESCRIPTION: Linear combinations of symbolic expressions
* COPYRIGHT  : (C) 2008  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __WRAP_SUM__HPP
#define __WRAP_SUM__HPP
#include <symbolix/symbolic.hpp>
#include <basix/sparse_vector.hpp>
#include <basix/wrap.hpp>
namespace mmx {
#define TMPL template<typename W>
#define Symbolic symbolic<W>
#define Sum wrap_sum<W>
#define Sum_rep sparse_vector<Symbolic,symbol<Symbolic >,hard_eq_sparse_vector>
#define Pair pair<symbol<Symbolic >,Symbolic>

TMPL
struct wrap_sum {
MMX_ALLOCATORS
  Sum_rep rep;
public:
  inline Sum_rep operator * () const { return rep; }
  inline wrap_sum (): rep () {}
  inline wrap_sum (const Sum_rep& rep2): rep (rep2) {}
  inline wrap_sum (const wrap_sum& s): rep (s.rep) {}
  inline wrap_sum (const Symbolic& m, const Symbolic& c): rep (Pair (m, c)) {}
  inline friend syntactic flatten (const Sum& s) { return W::expression (s); }
  inline friend nat N (const Sum& s) { return N (s.rep); }
  inline friend Symbolic mo (const Sum& s, nat i) { return *(s.rep[i].x1); }
  inline friend Symbolic co (const Sum& s, nat i) { return s.rep[i].x2; }
  inline Symbolic operator [] (nat i) const {
    return W::term (*(rep[i].x1), rep[i].x2); }
};

WRAP_WRAPPED_IMPL(TMPL inline,Sum)
PRINT_SUGAR(TMPL,Sum)

TMPL struct symbolic_type_information<Sum > {
  static const nat id= SYMBOLIC_SUM; };

TMPL syntactic dump (const Sum& s) {
  return apply ("sum", arguments (dump (*s))); }

TMPL inline Sum operator - (const Sum& s1) {
  return Sum (-(*s1)); }
TMPL inline Sum operator + (const Sum& s1, const Sum& s2) {
  return Sum ((*s1) + (*s2)); }
TMPL inline Sum operator - (const Sum& s1, const Sum& s2) {
  return Sum ((*s1) - (*s2)); }
TMPL inline Sum operator * (const Sum& s1, const Symbolic& x) {
  return Sum ((*s1) * x); }
TMPL inline Sum operator / (const Sum& s1, const Symbolic& x) {
  return Sum ((*s1) / x); }

#undef TMPL
#undef Symbolic
#undef Sum
#undef Sum_rep
#undef Pair
} // namespace mmx
#endif // __WRAP_SUM__HPP
